/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream>
#include <string>
#include <vector>
#include <pargal.hpp>
#include <moab/Core.hpp>
#include "mpi.h"

using namespace pargal;
using namespace moab;

std::string var_name("T"); 
std::string var_avg_name("Tavg");
void get_file_names(std::vector<std::string>& file_names); 

int main(int argc, char *argv[])
{  
  std::vector<std::string> file_names;
  MPI_Init(&argc, &argv);
  try 
    {
      get_file_names(file_names);
      
      // create moab instance 
      Core moab;
      Interface& mb = moab;
      moab::ParallelComm mb_pcomm(&mb, MPI_COMM_WORLD);
      
      // create meshset
      EntityHandle file_set;
      ErrorCode rval = mb.create_meshset(MESHSET_SET, file_set);
      if (rval != MB_SUCCESS) 
	throw pargal_except("Error: " + mb.get_error_string(rval), 
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);
    
      // load file metadata  
      std::string part_method("SQIJ");
      fileinfo finfo(file_names, mb, mb_pcomm, file_set, part_method);

      // load mesh
      std::stringstream opts;  
      opts << "PARALLEL=READ_PART;PARTITION;PARALLEL_RESOLVE_SHARED_ENTS;PARALLEL_GHOSTS=2.0.1;PARTITION_METHOD=";
      opts << part_method;
      opts << ";VARIABLE=";
      rval = mb.load_file(file_names.front().c_str(), &file_set, opts.str().c_str());
      if (moab::MB_SUCCESS != rval) 
	throw pargal_except("Error: " + mb.get_error_string(rval), 
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);

      // compute time steps    
      std::size_t num_tsteps = finfo.get_dim("time").size() * file_names.size();      
      std::vector<int> tsteps(num_tsteps);
      for (std::size_t i = 0; i != num_tsteps; ++i)
	tsteps[i] = i;
    
      // create variable
      pcvar var(var_name, moab::MB_TYPE_DOUBLE, finfo.get_var_dims(var_name), 
		finfo, tsteps, false);
    
      // compute dimensions for average variable
      std::vector<pcdim> var_avg_dims = finfo.get_var_dims(var_name);
      var_avg_dims[0].set_size(1);

      // create average variable
      std::vector<int> avg_tsteps(1, 0);
      pcvar var_avg(var_avg_name, moab::MB_TYPE_DOUBLE, 
		    var_avg_dims, finfo, avg_tsteps, true);
    
      // comupte average
      std::vector<std::size_t> dims(1,0);
      dim_avg_n(var, var_avg, dims);
      
      // gather average
      std::vector<double> avg_result;
      gather(var_avg, avg_tsteps, avg_result);
    }
  catch (pargal::pargal_except& e) 
    {
      e.get_info();
    }
  catch (std::exception& e) 
    {
      std::cout << e.what() << std::endl;
    }
  catch (...) 
    {
      std::cerr << "Unrecoganized Exception!" << std::endl;
    }
  
  MPI_Finalize();  

  return 0;
}


void get_file_names(std::vector<std::string>& file_names) 
{
  //std::string head("/fusion/group/climate/Parvis/atmos/b40.1850.track1.1deg.006/b40.1850.track1.1deg.006.cam2.h0.01");
  //size_t y_start = 0;
  //size_t y_end = 11;

  std::string head("/fusion/group/climate/Parvis/atmos/b40.1850_ramp.track1.1deg/b40.1850_ramp.track1.1deg.001.cam2.h0.19");
  std::size_t y_start = 9;
  std::size_t y_end = 29;

  std::size_t m_start = 1;
  std::size_t m_end = 13;

  for (std::size_t i = y_start; i != y_end; ++i)
    for (std::size_t j = m_start; j != m_end; ++j) 
      {
	std::ostringstream s;
	s << head;
	if (i < 10)
	  s << 0 << i;
	else
	  s << i;
	s << "-";
	if (j < 10)
	  s << 0 << j;
	else
	  s << j;
	s << ".nc";
	file_names.push_back(s.str());
      }
}
