/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream>
#include <string>
#include <vector>
#include <moab/Core.hpp>
#include <pargal.hpp>
#include "mpi.h"

using namespace pargal;
using namespace moab;

std::string var_name("T"); 
std::string file_name("sample_file");

int main(int argc, char *argv[])
{
  MPI_Init(&argc, &argv);
  try 
    {
      // create moab instance
      Core moab;
      Interface& mb = moab;
      moab::ParallelComm mb_pcomm(&mb, MPI_COMM_WORLD);
      
      // create meshset
      EntityHandle file_set;
      ErrorCode rval = mb.create_meshset(MESHSET_SET, file_set);
      if (rval != MB_SUCCESS) 
	throw pargal_except("Error: " + mb.get_error_string(rval),
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);
      
      // load file metadata
      std::vector<std::string> file_names;
      file_names.push_back(file_name);  
      std::string part_method("SQIJ");
      fileinfo finfo(file_names, mb, mb_pcomm, file_set, part_method);
      
      // load mesh
      std::stringstream opts;  
      opts << "PARALLEL=READ_PART;PARTITION;PARALLEL_RESOLVE_SHARED_ENTS;PARALLEL_GHOSTS=2.0.1;PARTITION_METHOD=";
      opts << part_method;
      opts << ";VARIABLE=";
      rval = mb.load_file(file_names.front().c_str(), &file_set, opts.str().c_str());
      if (rval != moab::MB_SUCCESS) 
	throw pargal_except("Error: " + mb.get_error_string(rval),
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);
  
      std::size_t num_tsteps = finfo.get_dim("time").size();
      std::vector<int> tsteps(num_tsteps);
      for (std::size_t i = 0; i != num_tsteps; ++i)
	tsteps[i] = i;
      
      // create variable
      pcvar var(var_name, moab::MB_TYPE_DOUBLE, finfo.get_var_dims(var_name),
		finfo, tsteps, false);
      
      double res_max = max(var);
      double res_min = min(var);
      std::cout << "max=" << res_max << std::endl
		<< "min=" << res_min << std::endl;
    }
  catch (pargal::pargal_except& e) 
    {
      e.get_info();
    }
  catch (std::exception& e) 
    {
      std::cout << e.what() << std::endl;
    }
  catch (...) 
    {
      std::cerr << "Unrecoganized Exception!" << std::endl;
    }
  
  MPI_Finalize();

  return 0;
}
