/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <pargal/algo_util.hpp>

namespace pargal 
{  
  bool 
  max_compare::operator()(double lhs, double rhs) {
    if (m_has_missing_val) {
      if (lhs==m_missing_val && rhs==m_missing_val) return false;
      else if (lhs==m_missing_val)
	return true;
      else if (rhs==m_missing_val)
	return false;
    }
    return lhs < rhs;
  }    
      
  bool 
  min_compare::operator()(double lhs, double rhs) {
    if (m_has_missing_val) {
      if (lhs==m_missing_val && rhs==m_missing_val) return false;
      else if (lhs==m_missing_val)
	return false;
      else if (rhs==m_missing_val)
	return true;
    }
    return lhs < rhs;
  }    

  double 
  fn_plus::operator()(const double& lhs, const double& rhs) {
    if (m_has_missing_val && rhs==m_missing_val)
      return lhs;
    ++counter;
    return lhs+rhs;
  }

  void dim_max_min_n_core(pcvar& var, pcvar& var_avg, const std::vector<std::size_t>& dims,
			  MPI_Op mpi_op, bool keep)
  {
    std::size_t sz = dims.size();
    if (sz == 1) {
      dim_max_min_n_single_dim_core(var, var_avg, dims[0], mpi_op, keep);
      return;
    } 
    if (sz > 4)
      throw pargal_except("Error: dim_max_n/dim_min_n dimension size can not exceed 4",
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);	    
    std::string var_tmp_name("_PARGAL_"+var_avg.name());
    pcvar var_tmp(var_tmp_name, var_avg.type(), 
		  var_avg.get_dims(), var_avg.get_finfo(), 
		  var_avg.get_tsteps(), true);
    if (sz == 2) {
      dim_max_min_n_single_dim_core(var, var_tmp, dims[0], mpi_op, keep);
      dim_max_min_n_single_dim_core(var_tmp, var_avg, dims[1], mpi_op, keep);
    }
    else if (sz == 3) {
      dim_max_min_n_single_dim_core(var, var_avg, dims[0], mpi_op, keep);
      dim_max_min_n_single_dim_core(var_avg, var_tmp, dims[1], mpi_op, keep);
      dim_max_min_n_single_dim_core(var_tmp, var_avg, dims[2], mpi_op, keep);
    }
    else if (sz == 4) {
      dim_max_min_n_single_dim_core(var, var_tmp, dims[0], mpi_op, keep);
      dim_max_min_n_single_dim_core(var_tmp, var_avg, dims[1], mpi_op, keep);
      dim_max_min_n_single_dim_core(var_avg, var_tmp, dims[2], mpi_op, keep);
      dim_max_min_n_single_dim_core(var_tmp, var_avg, dims[3], mpi_op, keep);
    }
  }
  
  //! computes the maximum of a variable's given dimension at all other dimensions
  void dim_max_min_n_single_dim_core(pcvar& var, pcvar& var_max, const std::size_t dim,
				     MPI_Op mpi_op, bool keep)
  {    
    max_compare maxcomp(var.has_missing_val(), var.get_missing_val());
    min_compare mincomp(var.has_missing_val(), var.get_missing_val());
    fn_max fnmax(maxcomp);
    fn_min fnmin(mincomp); 
    bool is_file_var = !var.is_usr_created();

    //! time dimension
    if (dim == 0) 
      {
	const std::vector<int>& tsteps = var.get_tsteps();
	double* var_max_ptr = var_max.get_storage();
	std::size_t var_max_sz = var_max.size();
	double* var_ptr = var.get_storage(0);
	std::size_t var_tstep_sz = var.tstep_size();
	if (var_max_sz != var_tstep_sz)
	  throw pargal_except("Error: var and var_max have different size",
			      __FILE__, __LINE__, __PRETTY_FUNCTION__);
	
	std::copy(var_ptr, var_ptr+var_tstep_sz, var_max_ptr);    
	if (!keep && is_file_var) var.delete_storage(0);
	for (std::size_t i = 1; i != tsteps.size(); ++i) 
	  {
	    var_ptr = var.get_storage(i);
	    double* var_ptr_end = var_ptr+var_tstep_sz;
	    if (mpi_op == MPI_MAX)
	      std::transform(var_ptr, var_ptr_end,
			     var_max_ptr, var_max_ptr,
			     fnmax);
	    else
	      std::transform(var_ptr, var_ptr_end,
			     var_max_ptr, var_max_ptr,
			     fnmin);
	    if (!keep && is_file_var) var.delete_storage(i);
	  }
      }   
    //! level dimension
    else if (dim == 1)
      {
	const std::vector<int>& tsteps = var.get_tsteps();	
	const std::size_t num_lev = var.num_levels();
	double* var_max_ptr=NULL;
	double* var_ptr=NULL;
	const std::size_t tstep_sz = var.tstep_size();
	for (std::size_t i = 0; i != tsteps.size(); ++i) 
	  {
	    var_max_ptr = var_max.get_storage(i);
	    var_ptr = var.get_storage(i);
	    if (mpi_op == MPI_MAX)
	      for (std::size_t j = 0; j != tstep_sz/num_lev; ++j)
		var_max_ptr[j*num_lev] = *std::max_element(var_ptr+j*num_lev, var_ptr+(j+1)*num_lev, maxcomp);
	    else
	      for (std::size_t j = 0; j != tstep_sz/num_lev; ++j)
		var_max_ptr[j*num_lev] = *std::min_element(var_ptr+j*num_lev, var_ptr+(j+1)*num_lev, mincomp);
	    if (!keep && is_file_var) var.delete_storage(i);
	  }
      } 
    //! lat dimension
    else if (dim == 2)
      {
	const std::size_t num_lev = var.num_levels();
	std::vector<int> ijminmax = var.get_dim_loc_minmax_owned();
	const std::size_t num_lon = static_cast<std::size_t>(ijminmax[1] - ijminmax[0] + 1);
	const std::size_t num_lat = static_cast<std::size_t>(ijminmax[3] - ijminmax[2] + 1);
	const std::size_t num_levlon = num_lev*num_lon; 
	const std::vector<int>& tsteps = var.get_tsteps();
	for (std::size_t t = 0; t != tsteps.size(); ++t)
	  {
	    double* var_ptr = var.get_storage(t);
	    double* var_max_ptr = var_max.get_storage(t);
	    std::copy(var_ptr, var_ptr+num_levlon, var_max_ptr);
	    if (mpi_op == MPI_MAX)
	      {
		for (std::size_t j = 1; j != num_lat; ++j)
		  std::transform(var_ptr+j*num_levlon, var_ptr+(j+1)*num_levlon, 
				 var_max_ptr, var_max_ptr,
				 fnmax);
	      }
	    else
	      {
		for (std::size_t j = 1; j != num_lat; ++j)
		  std::transform(var_ptr+j*num_levlon, var_ptr+(j+1)*num_levlon, 
				 var_max_ptr, var_max_ptr,
				 fnmin);
	      }
	    const moab::ParallelComm& mb_pcomm = var.get_finfo().get_pcomm();
	    int rank = mb_pcomm.proc_config().proc_rank();
	    int np = mb_pcomm.proc_config().proc_size();
	    std::vector<int> vec_col(var.get_column_ranks());
	    //! np columns
	    if (vec_col.size() == 1)
	      {		
	      }
	    //! 1 column
	    else if (vec_col.size() == static_cast<std::size_t>(np))
	      {
		std::vector<double> recvbuf(num_levlon, 0);
		MPI_Reduce(var_max_ptr, &recvbuf[0], static_cast<int>(num_levlon),
			   MPI_DOUBLE, mpi_op, 0, MPI_COMM_WORLD);
		if (rank == 0)
		  std::copy(recvbuf.begin(), recvbuf.end(), var_max_ptr);
	      }
	    //! multiple columns
	    else
	      {
		MPI_Group orig_group, new_group;
		MPI_Comm new_comm;
		//! extract the original group handle
		MPI_Comm_group(MPI_COMM_WORLD, &orig_group);
		//! create groups
		MPI_Group_incl(orig_group, vec_col.size(), &vec_col[0], &new_group);
		//! create new communicator
		MPI_Comm_create(MPI_COMM_WORLD, new_group, &new_comm);
		//! perform collective communications for each group
		int new_rank = 0;
		MPI_Group_rank(new_group, &new_rank);
		std::vector<double> recvbuf(num_levlon, 0);
		MPI_Reduce(var_max_ptr, &recvbuf[0], static_cast<int>(num_levlon),
			   MPI_DOUBLE, mpi_op, 0, new_comm);
		if (new_rank == 0)
		  std::copy(recvbuf.begin(), recvbuf.end(), var_max_ptr);
		//! free up new communicator and group
		MPI_Comm_free(&new_comm);
		MPI_Group_free(&new_group);
	      } 	
	    if (!keep && is_file_var) var.delete_storage(t);
   	  } //! end for each time step
      }
    //! lon dimension
    else if (dim == 3)
      {
	const std::size_t num_lev = var.num_levels();
	std::vector<int> ijminmax = var.get_dim_loc_minmax_owned();
	const std::size_t num_lon = static_cast<std::size_t>(ijminmax[1] - ijminmax[0] + 1);
	const std::size_t num_lat = static_cast<std::size_t>(ijminmax[3] - ijminmax[2] + 1);
	const std::size_t num_levlon = num_lev*num_lon; 
	const std::size_t num_levlat = num_lev*num_lat;
	const std::vector<int>& tsteps = var.get_tsteps();
	for (std::size_t t = 0; t != tsteps.size(); ++t)
	  {
	    double* var_ptr = var.get_storage(t);
	    double* var_max_ptr = var_max.get_storage(t);
	    for (std::size_t j = 0; j != num_lat; ++j)
	      std::copy(var_ptr+j*num_levlon, var_ptr+j*num_levlon+num_lev,
			var_max_ptr+j*num_levlon);
	    if (mpi_op == MPI_MAX) 
	      {
		for (std::size_t j = 0; j != num_lat; ++j)
		  for (std::size_t i = 1; i != num_lon; ++i)
		    std::transform(var_ptr+j*num_levlon +i*num_lev, var_ptr+j*num_levlon+(i+1)*num_lev,
				   var_max_ptr+j*num_levlon, var_max_ptr+j*num_levlon,
				   fnmax);
	      }
	    else
	      {
		for (std::size_t j = 0; j != num_lat; ++j)
		  for (std::size_t i = 1; i != num_lon; ++i)
		    std::transform(var_ptr+j*num_levlon +i*num_lev, var_ptr+j*num_levlon+(i+1)*num_lev,
				   var_max_ptr+j*num_levlon, var_max_ptr+j*num_levlon,
				   //static_cast<const double& (*) (const double&, const double&)>(std::min<double>));
				   fnmin);
	      }
	    const moab::ParallelComm& mb_pcomm = var.get_finfo().get_pcomm();
	    int rank = mb_pcomm.proc_config().proc_rank();
	    int np = mb_pcomm.proc_config().proc_size();
	    std::vector<int> vec_row(var.get_row_ranks());
	    //! np rows
	    if (vec_row.size() == 1)
	      {		
	      }
	    //! 1 row
	    else if (vec_row.size() == static_cast<std::size_t>(np))
	      {
		std::vector<double> max_res(num_levlat, 0);	    
		for (std::size_t j = 0; j != num_lat; ++j)
		  std::copy(var_max_ptr+j*num_levlon, 
			    var_max_ptr+j*num_levlon+num_lev,
			    &max_res[j*num_lev]);	       
		std::vector<double> recvbuf(max_res.size(), 0);
		MPI_Reduce(&max_res[0], &recvbuf[0], static_cast<int>(max_res.size()),
			   MPI_DOUBLE, mpi_op, 0, MPI_COMM_WORLD);
		if (rank == 0)
		  for (std::size_t j = 0; j != num_lat; ++j)
		    std::copy(&recvbuf[j*num_lev], &recvbuf[j*num_lev+num_lev],
			      var_max_ptr+j*num_levlon);
 	      }
	    //! multiple rows
	    else 
	      {
		MPI_Group orig_group, new_group;
		MPI_Comm new_comm;
		//! extract the original group handle
		MPI_Comm_group(MPI_COMM_WORLD, &orig_group);
		//! create groups
		MPI_Group_incl(orig_group, vec_row.size(), &vec_row[0], &new_group);
		//! create new communicator
		MPI_Comm_create(MPI_COMM_WORLD, new_group, &new_comm);
		//! perform collective communications for each group
		int new_rank = 0;
		MPI_Group_rank(new_group, &new_rank);
		std::vector<double> max_res(num_levlat, 0);	    
		for (std::size_t j = 0; j != num_lat; ++j)
		  std::copy(var_max_ptr+j*num_levlon, 
			    var_max_ptr+j*num_levlon+num_lev,
			    &max_res[j*num_lev]);	       
		std::vector<double>recvbuf(max_res.size(), 0);
		MPI_Reduce(&max_res[0], &recvbuf[0], static_cast<int>(max_res.size()),
			   MPI_DOUBLE, mpi_op, 0, new_comm);
		if (new_rank == 0)
		  for (std::size_t j = 0; j != num_lat; ++j)
		    std::copy(&recvbuf[j*num_lev], &recvbuf[j*num_lev+num_lev],
			      var_max_ptr+j*num_levlon);
		//! free up new communicator and group
		MPI_Comm_free(&new_comm);
		MPI_Group_free(&new_group);
	      }	
	    if (!keep && is_file_var) var.delete_storage(t);
   	  } //! end for each time step
      } 
    //! unrecognized dimension
    else
      {
	throw pargal_except("Error: dim_max/min_n dimension has to be 0/1/2/3",
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);
      }
  } 
} // namespace pargal

