/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_MOAB_REDUCE_COMM_HPP_INCLUDED
#define PARGAL_MOAB_REDUCE_COMM_HPP_INCLUDED

namespace pargal{
namespace algorithm{
namespace util{

class dim_reduce_comm{
	public:
		dim_reduce_comm(pargal::pcvar &var, const size_t odim);
		MPI_Comm get_mpi_comm(void );
		~dim_reduce_comm();
	private:
		MPI_Comm comm_;
		bool free_comm_;
};

inline dim_reduce_comm::dim_reduce_comm(pargal::pcvar &var, const size_t odim):comm_(MPI_COMM_NULL), free_comm_(false)
{
    switch(var.get_dim_type(odim)){
        case pargal::util::MOAB_TAG_SPARSE_DIM:
        case pargal::util::MOAB_TAG_GEN_DIM:
        case pargal::util::MOAB_TAG_LEV_DIM:
                break;
        case pargal::util::MOAB_TAG_SQIJ_LAT_DIM:
				{
					const moab::ParallelComm& mb_pcomm = var.get_finfo().get_pcomm();
					int np = mb_pcomm.proc_config().proc_size();
					std::vector<int> var_sqij_column_ranks(var.get_column_ranks());
					if(var_sqij_column_ranks.size() == 1){
						// One element per column => a row of np columns
						// We don't need to gather any info
					}
					else if(var_sqij_column_ranks.size() ==
								static_cast<std::size_t>(np)){
						// np elements per column => n columns of np rows
						// All procs need to be part of the same coll
						comm_ = MPI_COMM_WORLD;
					}
					else{
						// Each column needs to perform an independent coll
						// We create a separate group for each column
						MPI_Group worldg, columng;

						MPI_Comm_group(MPI_COMM_WORLD, &worldg);
						MPI_Group_incl(worldg, var_sqij_column_ranks.size(),
							&var_sqij_column_ranks[0], &columng);
						MPI_Comm_create(MPI_COMM_WORLD, columng, &comm_);
						free_comm_ = true;
					}
				}
				break;
        case pargal::util::MOAB_TAG_SQIJ_LON_DIM:
				{
					const moab::ParallelComm& mb_pcomm = var.get_finfo().get_pcomm();
					int np = mb_pcomm.proc_config().proc_size();
					std::vector<int> var_sqij_row_ranks(var.get_row_ranks());
					if(var_sqij_row_ranks.size() == 1){
						// One element per row => a column of np rows
						// We don't need to gather any info
					}
					else if(var_sqij_row_ranks.size() ==
								static_cast<std::size_t>(np)){
						// np elements per row => n rows of np columns
						// All procs need to be part of the same coll
						comm_ = MPI_COMM_WORLD;
					}
					else{
						// Each row needs to perform an independent coll
						// We create a separate group for each row
						MPI_Group worldg, rowg;

						MPI_Comm_group(MPI_COMM_WORLD, &worldg);
						MPI_Group_incl(worldg, var_sqij_row_ranks.size(),
							&var_sqij_row_ranks[0], &rowg);
						comm_ = MPI_COMM_WORLD;
						MPI_Comm_create(MPI_COMM_WORLD, rowg, &comm_);
						free_comm_ = true;
					}
				}
				break;
        default:
				throw pargal::pargal_except("Error initing reduce comm - invalid dim", __FILE__, __LINE__, "Unknown");
    }
}

inline MPI_Comm dim_reduce_comm::get_mpi_comm(void )
{
	return comm_;
}

inline dim_reduce_comm::~dim_reduce_comm()
{
	if(free_comm_){
		MPI_Comm_free(&comm_);
	}
}

} // namespace util
} // namespace algorithm
} // namespace pargal

#endif // PARGAL_MOAB_REDUCE_COMM_HPP_INCLUDED

