/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <pargal/algo_util.hpp>
#include <pargal/moab_tag_iterator.hpp>
#include <pargal/moab_tag_chunk.hpp>
#include <memory>
#include "moab_tag_sum.hpp"
#include "moab_tag_init.hpp"
#include "moab_tag_reduce.hpp"

namespace pargal 
{
//! computes the average of a variable's given dimension at all other dimensions

template<typename T>
void dim_sum_n_single_dim(pcvar& var, pcvar& var_sum, const size_t odim)
{
    bool has_missing_val = var.has_missing_val();
    double missing_val = var.get_missing_val();
    bool is_file_var = !var.is_usr_created();

	std::map<util::moab_tag_chunk::key_type, util::moab_tag_chunk, util::moab_tag_chunk::key_compare_functor> red_chunks;

	pcvar::iterator var_citer = var.get_chunk_iterator(odim);
	pcvar::red_iterator var_sum_citer = var_sum.get_red_chunk_iterator(odim);
	pargal::algorithm::util::dim_reduce_comm red_comm(var, odim);
	MPI_Comm red_mpi_comm = red_comm.get_mpi_comm();

	for(var_citer.chbegin(), var_sum_citer.chbegin();
		(!var_citer.chend()) && (!var_sum_citer.chend());
		++var_citer, ++var_sum_citer){

		util::moab_tag_chunk var_tc = *var_citer;
		util::moab_tag_chunk var_sum_tc = *var_sum_citer;

		red_chunks.insert(std::pair<util::moab_tag_chunk::key_type, util::moab_tag_chunk>(
			var_sum_tc.get_key(), var_sum_tc));

		for(util::moab_tag_chunk::iterator var_seq_iter = var_tc.begin(),
				var_sum_seq_iter = var_sum_tc.begin();
				(var_seq_iter != var_tc.end()) &&
				(var_sum_seq_iter != var_sum_tc.end());
				++var_seq_iter, ++var_sum_seq_iter){
			util::moab_tag_chunk_seq& var_seq = *var_seq_iter;
			util::moab_tag_chunk_seq& var_sum_seq = *var_sum_seq_iter;	

            if(!var_sum_tc.is_init()){
                algorithm::util::distributed_red_init(var_sum_seq,
                    (var_seq.has_missing_val())?
                    var_seq.get_missing_val(): 0.0,
                    0.0,
                    red_mpi_comm,
                    typename util::dim_traits<T>::dim_iterator_category(),
                    typename util::dim_traits<T>::dim_red_iterator_category());
            }

			algorithm::util::distributed_sum(var_seq, var_sum_seq,
				typename util::dim_traits<T>::dim_iterator_category(),
				typename util::dim_traits<T>::dim_red_iterator_category());
		}
		if(var_tc.needs_reduce()){
			// Average = Sum/count
			algorithm::util::dim_reduce<T>(red_chunks, var_sum, algorithm::util::PARGAL_RED_OP_DIM_SUM, odim);
			red_chunks.clear();
		}
	}
}

void dim_sum_n(pcvar& var, pcvar& var_sum, const std::vector<std::size_t>& dims, bool keep)
{
	std::size_t sz = dims.size();
	pcvar *in_var = &var;
	std::vector<pcvar *> out_var(2);
	out_var[0] = &var_sum;
	out_var[1] = NULL;

	assert(!dims.empty());
	if(sz > 1){
		// Use a single temp variable for calcs with > 1 dims
		// FIXME: Is the naming of the temp var good enough?
		std::string var_tmp_name("_PARGAL_"+var_sum.name());
		out_var[1] = new pcvar(var_tmp_name, var_sum.type(),
							var_sum.get_dims(), var_sum.get_finfo(),
							var_sum.get_tsteps(), true,
							var_sum.get_loc(), false, var_sum.get_missing_val());
	}

	size_t out_var_idx = ((sz % 2) == 0) ? 1 : 0;
	for(size_t i=0; i<sz; i++){
		PARGAL_VAR_DIM_VOID_FUNC_DISPATCH(in_var->get_dim_type(dims[i]), dim_sum_n_single_dim, *in_var, *(out_var[out_var_idx]), dims[i]);
		// Use the result as input for next calc
		in_var = out_var[out_var_idx];
		out_var_idx = (out_var_idx + 1) % 2;
	}
}

} // namespace pargal

