/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <pargal/algo_util.hpp>

namespace pargal 
{  

const double lowest_numeric_double = -1 * std::numeric_limits<double>::max();

// Get max local to this proc
template<typename iterator>
inline double get_local_max(iterator begin, iterator end, pcvar &var)
{
	double local_max = 0.0; 
	if(!var.has_missing_val()){
		local_max = *std::max_element(begin, end);
	}
	else{
		double mval = var.get_missing_val();
		pargal::algorithm::util::missing_val_comp comp(mval);
		local_max = *std::max_element(begin, end, comp);
		if(local_max == mval){
			// We perform MPI_Reduce(MPI_MAX) on these local maxes
			// Unfortunately there is no way (without using global vars)
			// to handle missing values even using user-defined MPI
			// reduction ops
			local_max = lowest_numeric_double;
		}
	}
	return local_max;
}

// Get max on all proc
inline double get_global_max(double loc_max, pcvar &var)
{
	double global_max = 0.0;
	MPI_Comm comm = var.get_finfo().get_mpi_comm();
	MPI_Reduce(&loc_max, &global_max, 1, MPI_DOUBLE, MPI_MAX, 0, comm);
	if(var.has_missing_val()){
		if(global_max == lowest_numeric_double){
			// There are only missing values here, max is also a
			// missing value
			global_max = var.get_missing_val();
		}
	}
	return global_max;
}

double max(pcvar &var, bool keep)
{
	const std::vector<int>& tsteps = var.get_tsteps();

	std::vector<double> seq_max;

	for(std::vector<int>::const_iterator iter = tsteps.begin();
		iter != tsteps.end();
		++iter){
		int tstep = *iter;
		bool cont = false;
		pcvar::cache_access_state rd_state;
		do{
			double *seq_buf = NULL;
			std::size_t seq_buf_len = 0;
			cont = var.cache_read(tstep, seq_buf, seq_buf_len, cont, rd_state); 
			double loc_seq_max = 0.0;
			// Get local max within the sequence
			loc_seq_max = get_local_max(seq_buf, seq_buf+seq_buf_len, var);
			seq_max.push_back(loc_seq_max);
		}while(cont == true);
		if(!keep && !var.is_usr_created()){
			var.cache_delete(tstep);
		}
	}
	// Get local max among on all the sequences
	double loc_max = get_local_max(seq_max.begin(), seq_max.end(), var);
	double global_max = get_global_max(loc_max, var);
	return global_max;
}

} // namespace pargal

