/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <pargal/algo_util.hpp>

namespace pargal 
{

const double max_numeric_double = std::numeric_limits<double>::max();

// Get min local to this proc
template<typename iterator>
inline double get_local_min(iterator begin, iterator end, pcvar &var)
{
	double local_min = 0.0;
	if(!var.has_missing_val()){
		local_min = *std::min_element(begin, end);
	}
	else{
		double mval = var.get_missing_val();
		pargal::algorithm::util::missing_val_comp comp(mval);
		local_min = *std::min_element(begin, end, comp);
		if(local_min == mval){
			// We perform MPI_Reduce(MPI_MIN) on these local mines
			// Unfortunately there is no way (without using global vars)
			// to handle missing values even using user-defined MPI
			// reduction ops
			local_min = max_numeric_double;
		}
	}
	return local_min;
}

// Get min on all proc
inline double get_global_min(double loc_min, pcvar &var)
{
	double global_min = 0.0;
	MPI_Comm comm = var.get_finfo().get_mpi_comm();
	MPI_Reduce(&loc_min, &global_min, 1, MPI_DOUBLE, MPI_MIN, 0, comm);
	if(var.has_missing_val()){
		if(global_min == max_numeric_double){
			// There are only missing values here, min is also a
			// missing value
			global_min = var.get_missing_val();
		}
	}
	return global_min;
}

double min(pcvar &var, bool keep)
{
	const std::vector<int>& tsteps = var.get_tsteps();

	std::vector<double> seq_min;

	for(std::vector<int>::const_iterator iter = tsteps.begin();
		iter != tsteps.end();
		++iter){
		int tstep = *iter;
		bool cont = false;
		pcvar::cache_access_state rd_state;
		do{
			double *seq_buf = NULL;
			std::size_t seq_buf_len = 0;
			cont = var.cache_read(tstep, seq_buf, seq_buf_len, cont, rd_state); 
			double loc_seq_min = 0.0;
			// Get local min within the sequence
			loc_seq_min = get_local_min(seq_buf, seq_buf+seq_buf_len, var);
			seq_min.push_back(loc_seq_min);
		}while(cont == true);
		if(!keep && !var.is_usr_created()){
			var.cache_delete(tstep);
		}
	}
	// Get local min among on all the sequences
	double loc_min = get_local_min(seq_min.begin(), seq_min.end(), var);
	double global_min = get_global_min(loc_min, var);
	return global_min;
}

} // namespace pargal

