/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a init of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_MOAB_TAG_INIT_HPP_INCLUDED
#define PARGAL_MOAB_TAG_INIT_HPP_INCLUDED

#include <pargal/pcvar.hpp>
#include <pargal/moab_tag_iterator.hpp>
#include <pargal/moab_tag_chunk.hpp>
#include <pargal/moab_tag_counter.hpp>
#include <iterator>
#include <pargal/algo_util.hpp>

namespace pargal{
namespace algorithm{
namespace util{

inline void distributed_init(pargal::util::moab_tag_chunk_seq &var_seq,
	double init_val,
	pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator)
{
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_seq.get_static_count();

	// FIXME: Use std::fill
	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin();
		iter_var != var_seq.end();){
		
		*iter_var = init_val;
		iter_var++;
	}
}

inline void distributed_init(pargal::util::moab_tag_chunk_seq &var_seq,
	double init_val,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator)
{
	size_t reset_cnt = var_seq.get_reset_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin();
		iter_var != var_seq.end();){
		std::size_t dist = pargal::util::distance(iter_var, var_seq.end());
		std::size_t fill_cnt = (reset_cnt <= dist) ? reset_cnt : dist;

		std::fill(iter_var, iter_var + fill_cnt, init_val);
		iter_var += reset_cnt;
		
		// Since we reset, we only init the first block 
		break;
	}
}

inline void distributed_init(pargal::util::moab_tag_chunk_seq &var_seq,
	double init_val,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator)
{
	size_t static_cnt = var_seq.get_static_count();
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin();
		iter_var != var_seq.end();){

		*iter_var = init_val;
		// We ignore the static behavior since we need to init only once,
		// instead of re-initing, in a static range
		//skip
		std::advance(iter_var, skip_dist);
	}
}

inline void distributed_init(pargal::util::moab_tag_chunk_seq &var_seq,
	double init_val,
    pargal::util::moab_tag_gen_dim_iterator,
    pargal::util::moab_tag_reset_skip_dim_iterator)
{
    size_t reset_cnt = var_seq.get_reset_count();
    size_t skip_cnt = var_seq.get_skip_count();
    size_t skip_dist = var_seq.get_skip_dist();

    for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin();
        iter_var != var_seq.end();){
        for(size_t i = 0;
            (i < skip_cnt) &&
            (iter_var != var_seq.end());){

            pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var;
            for(size_t j = 0;
                (j < reset_cnt) &&
                (iter_var != var_seq.end());
                ++iter_var, i++, j++){
                *iter_var = init_val;
            }
            //reset
            iter_var = reset_iter;
			// Since we reset, we only init the first block 
			// Manually move the var iter by skip count since we break here
			//std::advance(iter_var, skip_cnt);
			break;
        }
        // skip
        std::advance(iter_var, skip_dist);
    }
}

inline double get_distributed_red_init_val(double mval, double alt_init_val,
	MPI_Comm red_mpi_comm)
{
	double init_val;
	if(red_mpi_comm == MPI_COMM_NULL){
		// No reducn - so init with mval
		init_val = mval;
	}
	else{
		int rank;
		// Only init the root with mval since we don't redn operns on mvals
		MPI_Comm_rank(red_mpi_comm, &rank);
		if(rank == 0){
			init_val = mval;
		}
		else{
			init_val = alt_init_val;
		}
	}
	return init_val;
}

inline void distributed_red_init(pargal::util::moab_tag_chunk_seq &var_seq,
	double mval,
	double alt_init_val,
	MPI_Comm red_mpi_comm,
	pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator)
{
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_seq.get_static_count();

	double init_val = get_distributed_red_init_val(mval, alt_init_val, red_mpi_comm);

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin();
		iter_var != var_seq.end();){
		
		*iter_var = init_val;
		iter_var++;
	}
}

inline void distributed_red_init(pargal::util::moab_tag_chunk_seq &var_seq,
	double mval,
	double alt_init_val,
	MPI_Comm red_mpi_comm,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator)
{
	size_t reset_cnt = var_seq.get_reset_count();

	double init_val = get_distributed_red_init_val(mval, alt_init_val, red_mpi_comm);
	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin();
		iter_var != var_seq.end();){
		std::size_t dist = pargal::util::distance(iter_var, var_seq.end());
		std::size_t fill_cnt = (reset_cnt <= dist) ? reset_cnt : dist;

		std::fill(iter_var, iter_var + fill_cnt, init_val);
		iter_var += reset_cnt;
		
		// Since we reset, we only init the first block 
		break;
	}
}

inline void distributed_red_init(pargal::util::moab_tag_chunk_seq &var_seq,
	double mval,
	double alt_init_val,
	MPI_Comm red_mpi_comm,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator)
{
	size_t static_cnt = var_seq.get_static_count();
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();

	double init_val = get_distributed_red_init_val(mval, alt_init_val, red_mpi_comm);
	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin();
		iter_var != var_seq.end();){

		*iter_var = init_val;
		// We ignore the static behavior since we need to init only once,
		// instead of re-initing, in a static range
		//skip
		std::advance(iter_var, skip_dist);
	}
}

inline void distributed_red_init(pargal::util::moab_tag_chunk_seq &var_seq,
	double mval,
	double alt_init_val,
	MPI_Comm red_mpi_comm,
    pargal::util::moab_tag_gen_dim_iterator,
    pargal::util::moab_tag_reset_skip_dim_iterator)
{
    size_t reset_cnt = var_seq.get_reset_count();
    size_t skip_cnt = var_seq.get_skip_count();
    size_t skip_dist = var_seq.get_skip_dist();

	double init_val = get_distributed_red_init_val(mval, alt_init_val, red_mpi_comm);
    for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin();
        iter_var != var_seq.end();){
        for(size_t i = 0;
            (i < skip_cnt) &&
            (iter_var != var_seq.end());){

            pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var;
            for(size_t j = 0;
                (j < reset_cnt) &&
                (iter_var != var_seq.end());
                ++iter_var, i++, j++){
                *iter_var = init_val;
            }
            //reset
            iter_var = reset_iter;
			// Since we reset, we only init the first block 
			// Manually move the var iter by skip count since we break here
			//std::advance(iter_var, skip_cnt);
			break;
        }
        // skip
        std::advance(iter_var, skip_dist);
    }
}
} //namespace util
} //namespace algorithm
} // namespace pargal

#endif // PARGAL_MOAB_TAG_INIT_HPP_INCLUDED
