/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_MOAB_TAG_MINMAX_HPP_INCLUDED
#define PARGAL_MOAB_TAG_MINMAX_HPP_INCLUDED

#include <pargal/pcvar.hpp>
#include <pargal/moab_tag_iterator.hpp>
#include <pargal/moab_tag_chunk.hpp>
#include <pargal/moab_tag_counter.hpp>
#include <functional>
#include <pargal/algo_util.hpp>

namespace pargal{
namespace algorithm{
namespace util{

struct minmax_comp : public std::binary_function<double, double, bool>{
    virtual bool operator() (double arg1, double arg2) = 0;
};

struct max_comp_m : public minmax_comp{
    max_comp_m(double missing_val):missing_val_(missing_val){}
    // Returns true if arg2 > arg1
    bool operator() (double arg1, double arg2);
    double missing_val_;
};

inline bool max_comp_m::operator() (double arg1, double arg2)
{
	if(arg1 == -1 * std::numeric_limits<double>::max()){
		return true;
	}
	else if((arg1 != missing_val_) && (arg2 != missing_val_)){
		return (arg2 > arg1);
	}
	else if(arg1 == missing_val_){
		return true;
	}
	else{
		return false;
	}
}

struct min_comp_m : public minmax_comp{
    min_comp_m(double missing_val):missing_val_(missing_val){}
    /// Returns true if arg2 < arg1
    bool operator() (double arg1, double arg2);
    double missing_val_;
};

inline bool min_comp_m::operator() (double arg1, double arg2)
{
	if(arg1 == std::numeric_limits<double>::max()){
		return true;
	}
	else if((arg1 != missing_val_) && (arg2 != missing_val_)){
		return (arg2 < arg1);
	}
	else if(arg1 == missing_val_){
		return true;
	}
	else{
		return false;
	}
}

// Distributed minmax - with missing vals
template<typename dim_iter, typename red_dim_iter>
inline void distributed_minmax_m(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_minmax_seq,
	minmax_comp &comp)
{
	throw pargal::pargal_except("Error: Unsupported iterator type. Cannot perform distributed minmax", __FILE__, __LINE__, "Unknown");
}

template<>
inline void distributed_minmax_m<pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_minmax_seq,
	minmax_comp &comp)
{
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_minmax_seq.get_static_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_minmax = var_minmax_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_minmax != var_minmax_seq.end());){
		size_t j=0;

		// Unchecked advances of iter_var is dangerous - however a
		// safe advance might be more costly
		for(j=0; 
			(j<static_cnt) &&
			(iter_var != var_seq.end());
			j++){
			// FIXME: Introduce value_type to use for temp
			if(comp(*iter_var_minmax, *iter_var)){
				*iter_var_minmax = *iter_var;
			}
			std::advance(iter_var, skip_dist);
		}
		if(j != skip_cnt){
			for(;j != skip_cnt; j++){
				std::advance(iter_var, skip_dist);
			}
		}
		iter_var_minmax++;
	}
}

template<>
inline void distributed_minmax_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_minmax_seq,
	minmax_comp &comp)
{
	size_t reset_cnt = var_minmax_seq.get_reset_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_minmax = var_minmax_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_minmax != var_minmax_seq.end());){

		pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_minmax;
		for(size_t i = 0;
			(i < reset_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_minmax != var_minmax_seq.end());
			++iter_var, ++iter_var_minmax, i++){
			if(comp(*iter_var_minmax, *iter_var)){
				*iter_var_minmax = *iter_var;
			}
		}
		iter_var_minmax = reset_iter;
	}
}

template<>
inline void distributed_minmax_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_minmax_seq,
	minmax_comp &comp)
{
	size_t static_cnt = var_minmax_seq.get_static_count();
	size_t skip_cnt = var_minmax_seq.get_skip_count();
	size_t skip_dist = var_minmax_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_minmax = var_minmax_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_minmax != var_minmax_seq.end());){

		for(size_t i = 0;
			(i < static_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_minmax != var_minmax_seq.end());
			++iter_var, i++){
			if(comp(*iter_var_minmax, *iter_var)){
				*iter_var_minmax = *iter_var;
			}
		}
		//skip
		std::advance(iter_var_minmax, skip_dist);
	}
}

template<>
inline void distributed_minmax_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_skip_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_minmax_seq,
	minmax_comp &comp)
{
	size_t reset_cnt = var_minmax_seq.get_reset_count();
	size_t skip_cnt = var_minmax_seq.get_skip_count();
	size_t skip_dist = var_minmax_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_minmax = var_minmax_seq.begin();
		(iter_var != var_seq.end()) && (iter_var_minmax != var_minmax_seq.end());){
		for(size_t i = 0;
			(i < skip_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_minmax != var_minmax_seq.end());
			){

			pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_minmax;
			for(size_t j = 0;
				(j < reset_cnt) &&
				(iter_var != var_seq.end()) &&
				(iter_var_minmax != var_minmax_seq.end());
				++iter_var, ++iter_var_minmax, i++, j++){
				if(comp(*iter_var_minmax, *iter_var)){
					*iter_var_minmax = *iter_var;
				}
			}
			//reset
			iter_var_minmax = reset_iter;
		}
		// skip
		std::advance(iter_var_minmax, skip_dist);
	}
}

inline void distributed_min(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_min_seq,
	pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator)
{
	if(var_seq.has_missing_val()){
		min_comp_m comp(var_seq.get_missing_val());
		distributed_minmax_m<pargal::util::moab_tag_sparse_skip_dim_iterator,
			pargal::util::moab_tag_sparse_static_dim_iterator>(
			var_seq, var_min_seq, comp);
		return;
	}
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_min_seq.get_static_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_min = var_min_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_min != var_min_seq.end());){
		size_t j=0;

		// Unchecked advances of iter_var is dangerous - however a
		// safe advance might be more costly
		for(j=0; 
			(j<static_cnt) &&
			(iter_var != var_seq.end());
			j++){
			if(*iter_var < *iter_var_min){
				*iter_var_min = *iter_var;
			}
			std::advance(iter_var, skip_dist);
		}
		if(j != skip_cnt){
			for(;j != skip_cnt; j++){
				std::advance(iter_var, skip_dist);
			}
		}
		iter_var_min++;
	}
}

inline void distributed_min(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_min_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator)
{
	if(var_seq.has_missing_val()){
		min_comp_m comp(var_seq.get_missing_val());
		distributed_minmax_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_reset_dim_iterator>(
			var_seq, var_min_seq, comp);
		return;
	}
	size_t reset_cnt = var_min_seq.get_reset_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_min = var_min_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_min != var_min_seq.end());){

		pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_min;
		for(size_t i = 0;
			(i < reset_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_min != var_min_seq.end());
			++iter_var, ++iter_var_min, i++){
			if(*iter_var < *iter_var_min){
				*iter_var_min = *iter_var;
			}
		}
		//reset
		iter_var_min = reset_iter;
	}
}

inline void distributed_min(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_min_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator)
{
	if(var_seq.has_missing_val()){
		min_comp_m comp(var_seq.get_missing_val());
		distributed_minmax_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_static_skip_dim_iterator>(
			var_seq, var_min_seq, comp);
		return;
	}
	size_t static_cnt = var_min_seq.get_static_count();
	size_t skip_cnt = var_min_seq.get_skip_count();
	size_t skip_dist = var_min_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_min = var_min_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_min != var_min_seq.end());){

		for(size_t i = 0;
			(i < static_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_min != var_min_seq.end());
			++iter_var, i++){
			if(*iter_var < *iter_var_min){
				*iter_var_min = *iter_var;
			}
		}
		//skip
		std::advance(iter_var_min, skip_dist);
	}
}

inline void distributed_min(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_min_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_skip_dim_iterator)
{
	if(var_seq.has_missing_val()){
		min_comp_m comp(var_seq.get_missing_val());
		distributed_minmax_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_reset_skip_dim_iterator>(
			var_seq, var_min_seq, comp);
		return;
	}
	size_t reset_cnt = var_min_seq.get_reset_count();
	size_t skip_cnt = var_min_seq.get_skip_count();
	size_t skip_dist = var_min_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_min = var_min_seq.begin();
		(iter_var != var_seq.end()) && (iter_var_min != var_min_seq.end());){
		for(size_t i = 0;
			(i < skip_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_min != var_min_seq.end());){

			pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_min;
			for(size_t j = 0;
				(j < reset_cnt) &&
				(iter_var != var_seq.end()) &&
				(iter_var_min != var_min_seq.end());
				++iter_var, ++iter_var_min, i++, j++){

				if(*iter_var < *iter_var_min){
					*iter_var_min = *iter_var;
				}
			}
			//reset
			iter_var_min = reset_iter;
		}
		// skip
		std::advance(iter_var_min, skip_dist);
	}
}

inline void distributed_max(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_max_seq,
	pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator)
{
	if(var_seq.has_missing_val()){
		max_comp_m comp(var_seq.get_missing_val());
		distributed_minmax_m<pargal::util::moab_tag_sparse_skip_dim_iterator,
			pargal::util::moab_tag_sparse_static_dim_iterator>(
			var_seq, var_max_seq, comp);
		return;
	}
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_max_seq.get_static_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_max = var_max_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_max != var_max_seq.end());){
		size_t j=0;

		// Unchecked advances of iter_var is dangerous - however a
		// safe advance might be more costly
		for(j=0; 
			(j<static_cnt) &&
			(iter_var != var_seq.end());
			j++){
			if(*iter_var > *iter_var_max){
				*iter_var_max = *iter_var;
			}
			std::advance(iter_var, skip_dist);
		}
		if(j != skip_cnt){
			for(;j != skip_cnt; j++){
				std::advance(iter_var, skip_dist);
			}
		}
		iter_var_max++;
	}
}

inline void distributed_max(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_max_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator)
{
	if(var_seq.has_missing_val()){
		max_comp_m comp(var_seq.get_missing_val());
		distributed_minmax_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_reset_dim_iterator>(
			var_seq, var_max_seq, comp);
		return;
	}
	size_t reset_cnt = var_max_seq.get_reset_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_max = var_max_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_max != var_max_seq.end());){

		pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_max;
		for(size_t i = 0;
			(i < reset_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_max != var_max_seq.end());
			++iter_var, ++iter_var_max, i++){
			if(*iter_var > *iter_var_max){
				*iter_var_max = *iter_var;
			}
		}
		//reset
		iter_var_max = reset_iter;
	}
}

inline void distributed_max(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_max_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator)
{
	if(var_seq.has_missing_val()){
		max_comp_m comp(var_seq.get_missing_val());
		distributed_minmax_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_static_skip_dim_iterator>(
			var_seq, var_max_seq, comp);
		return;
	}
	size_t static_cnt = var_max_seq.get_static_count();
	size_t skip_cnt = var_max_seq.get_skip_count();
	size_t skip_dist = var_max_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_max = var_max_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_max != var_max_seq.end());){

		for(size_t i = 0;
			(i < static_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_max != var_max_seq.end());
			++iter_var, i++){
			if(*iter_var > *iter_var_max){
				*iter_var_max = *iter_var;
			}
		}
		//skip
		std::advance(iter_var_max, skip_dist);
	}
}

inline void distributed_max(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_max_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_skip_dim_iterator)
{
	if(var_seq.has_missing_val()){
		max_comp_m comp(var_seq.get_missing_val());
		distributed_minmax_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_reset_skip_dim_iterator>(
			var_seq, var_max_seq, comp);
		return;
	}
	size_t reset_cnt = var_max_seq.get_reset_count();
	size_t skip_cnt = var_max_seq.get_skip_count();
	size_t skip_dist = var_max_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_max = var_max_seq.begin();
		(iter_var != var_seq.end()) && (iter_var_max != var_max_seq.end());){
		for(size_t i = 0;
			(i < skip_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_max != var_max_seq.end());){

			pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_max;
			for(size_t j = 0;
				(j < reset_cnt) &&
				(iter_var != var_seq.end()) &&
				(iter_var_max != var_max_seq.end());
				++iter_var, ++iter_var_max, i++, j++){

				if(*iter_var > *iter_var_max){
					*iter_var_max = *iter_var;
				}
			}
			//reset
			iter_var_max = reset_iter;
		}
		// skip
		std::advance(iter_var_max, skip_dist);
	}
}

} //namespace util
} //namespace algorithm
} // namespace pargal

#endif // PARGAL_MOAB_TAG_MINMAX_HPP_INCLUDED
