/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_MOAB_TAG_PROD_HPP_INCLUDED
#define PARGAL_MOAB_TAG_PROD_HPP_INCLUDED

#include <pargal/pcvar.hpp>
#include <pargal/moab_tag_iterator.hpp>
#include <pargal/moab_tag_chunk.hpp>
#include <pargal/moab_tag_counter.hpp>
#include <functional>
#include <pargal/algo_util.hpp>

namespace pargal{
namespace algorithm{
namespace util{

// Distributed product - with missing vals
template<typename dim_iter, typename red_dim_iter>
inline void distributed_product_m(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_product_seq)
{
	throw pargal::pargal_except("Error: Unsupported iterator type. Cannot perform distributed product", __FILE__, __LINE__, "Unknown");
}

template<>
inline void distributed_product_m<pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_product_seq)
{
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_product_seq.get_static_count();
	double mval = var_seq.get_missing_val();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_product = var_product_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_product != var_product_seq.end());){
		size_t j=0;

		// Unchecked advances of iter_var is dangerous - however a
		// safe advance might be more costly
		for(j=0; 
			(j<static_cnt) &&
			(iter_var != var_seq.end());
			j++){
			// FIXME: Introduce value_type to use for temp
			if(*iter_var_product == mval){
				*iter_var_product = *iter_var;
			}
			else if(*iter_var != mval){
				*iter_var_product *= *iter_var;
			}
			std::advance(iter_var, skip_dist);
		}
		if(j != skip_cnt){
			for(;j != skip_cnt; j++){
				std::advance(iter_var, skip_dist);
			}
		}
		iter_var_product++;
	}
}

template<>
inline void distributed_product_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_product_seq)
{
	size_t reset_cnt = var_product_seq.get_reset_count();
	double mval = var_seq.get_missing_val();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_product = var_product_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_product != var_product_seq.end());){

		pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_product;
		for(size_t i = 0;
			(i < reset_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_product != var_product_seq.end());
			++iter_var, ++iter_var_product, i++){
			if(*iter_var_product == mval){
				*iter_var_product = *iter_var;
			}
			else if(*iter_var != mval){
				*iter_var_product *= *iter_var;
			}
		}
		iter_var_product = reset_iter;
	}
}

template<>
inline void distributed_product_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_product_seq)
{
	size_t static_cnt = var_product_seq.get_static_count();
	size_t skip_cnt = var_product_seq.get_skip_count();
	size_t skip_dist = var_product_seq.get_skip_dist();
	double mval = var_seq.get_missing_val();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_product = var_product_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_product != var_product_seq.end());){

		for(size_t i = 0;
			(i < static_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_product != var_product_seq.end());
			++iter_var, i++){
			if(*iter_var_product == mval){
				*iter_var_product = *iter_var;
			}
			else if(*iter_var != mval){
				*iter_var_product *= *iter_var;
			}
		}
		//skip
		std::advance(iter_var_product, skip_dist);
	}
}

template<>
inline void distributed_product_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_skip_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_product_seq)
{
	size_t reset_cnt = var_product_seq.get_reset_count();
	size_t skip_cnt = var_product_seq.get_skip_count();
	size_t skip_dist = var_product_seq.get_skip_dist();
	double mval = var_seq.get_missing_val();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_product = var_product_seq.begin();
		(iter_var != var_seq.end()) && (iter_var_product != var_product_seq.end());){
		for(size_t i = 0;
			(i < skip_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_product != var_product_seq.end());
			){

			pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_product;
			for(size_t j = 0;
				(j < reset_cnt) &&
				(iter_var != var_seq.end()) &&
				(iter_var_product != var_product_seq.end());
				++iter_var, ++iter_var_product, i++, j++){
				if(*iter_var_product == mval){
					*iter_var_product = *iter_var;
				}
				else if(*iter_var != mval){
					*iter_var_product *= *iter_var;
				}
			}
			//reset
			iter_var_product = reset_iter;
		}
		// skip
		std::advance(iter_var_product, skip_dist);
	}
}

inline void distributed_product(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_product_seq,
	pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_product_m<pargal::util::moab_tag_sparse_skip_dim_iterator,
			pargal::util::moab_tag_sparse_static_dim_iterator>(
			var_seq, var_product_seq);
		return;
	}
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_product_seq.get_static_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_product = var_product_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_product != var_product_seq.end());){
		size_t j=0;

		// Unchecked advances of iter_var is dangerous - however a
		// safe advance might be more costly
		for(j=0; 
			(j<static_cnt) &&
			(iter_var != var_seq.end());
			j++){
			*iter_var_product *= *iter_var;
			std::advance(iter_var, skip_dist);
		}
		if(j != skip_cnt){
			for(;j != skip_cnt; j++){
				std::advance(iter_var, skip_dist);
			}
		}
		iter_var_product++;
	}
}

inline void distributed_product(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_product_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_product_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_reset_dim_iterator>(
			var_seq, var_product_seq);
		return;
	}
	size_t reset_cnt = var_product_seq.get_reset_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_product = var_product_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_product != var_product_seq.end());){

		pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_product;
		for(size_t i = 0;
			(i < reset_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_product != var_product_seq.end());
			++iter_var, ++iter_var_product, i++){
			*iter_var_product *= *iter_var;
		}
		//reset
		iter_var_product = reset_iter;
	}
}

inline void distributed_product(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_product_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_product_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_static_skip_dim_iterator>(
			var_seq, var_product_seq);
		return;
	}
	size_t static_cnt = var_product_seq.get_static_count();
	size_t skip_cnt = var_product_seq.get_skip_count();
	size_t skip_dist = var_product_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_product = var_product_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_product != var_product_seq.end());){

		for(size_t i = 0;
			(i < static_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_product != var_product_seq.end());
			++iter_var, i++){
			*iter_var_product *= *iter_var;
		}
		//skip
		std::advance(iter_var_product, skip_dist);
	}
}

inline void distributed_product(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_product_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_skip_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_product_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_reset_skip_dim_iterator>(
			var_seq, var_product_seq);
		return;
	}
	size_t reset_cnt = var_product_seq.get_reset_count();
	size_t skip_cnt = var_product_seq.get_skip_count();
	size_t skip_dist = var_product_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_product = var_product_seq.begin();
		(iter_var != var_seq.end()) && (iter_var_product != var_product_seq.end());){
		for(size_t i = 0;
			(i < skip_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_product != var_product_seq.end());){

			pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_product;
			for(size_t j = 0;
				(j < reset_cnt) &&
				(iter_var != var_seq.end()) &&
				(iter_var_product != var_product_seq.end());
				++iter_var, ++iter_var_product, i++, j++){

				*iter_var_product *= *iter_var;
			}
			//reset
			iter_var_product = reset_iter;
		}
		// skip
		std::advance(iter_var_product, skip_dist);
	}
}

} //namespace util
} //namespace algorithm
} // namespace pargal

#endif // PARGAL_MOAB_TAG_PROD_HPP_INCLUDED
