/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_MOAB_TAG_REDUCE_HPP_INCLUDED
#define PARGAL_MOAB_TAG_REDUCE_HPP_INCLUDED

#include <cstddef>
#include <pargal/moab_tag_chunk.hpp>
#include <pargal/moab_tag_counter.hpp>
#include "moab_tag_divide.hpp"
#include "dim_reduce_comm.hpp"

namespace pargal{
namespace algorithm{
namespace util{

enum pargal_red_op{
	PARGAL_RED_OP_DIM_AVG=1,
	PARGAL_RED_OP_DIM_SUM,
	PARGAL_RED_OP_DIM_PRODUCT,
	PARGAL_RED_OP_DIM_MAX,
	PARGAL_RED_OP_DIM_MIN,
	PARGAL_INVALID_RED_OP	
};

// FIXME: Overloading funcs might be a better choice here - since we don;t
// need this functionality for some iterators
inline void moab_tag_reduce(pargal::util::moab_tag_chunk_seq &var_seq,
		std::size_t red_sz,
		MPI_Op op, MPI_Comm comm,
		pargal::util::moab_tag_sparse_skip_dim_iterator,
		pargal::util::moab_tag_sparse_static_dim_iterator)
{
	throw pargal::pargal_except("Error: Invalid iterator, cannot perform redn op",
		__FILE__, __LINE__, "Unknown");
}

inline void moab_tag_reduce(pargal::util::moab_tag_chunk_seq &var_seq,
		std::size_t red_sz,
		MPI_Op op, MPI_Comm comm,
		pargal::util::moab_tag_gen_dim_iterator,
		pargal::util::moab_tag_static_skip_dim_iterator)
{
	throw pargal::pargal_except("Error: Invalid iterator, cannot perform redn op",
		__FILE__, __LINE__, "Unknown");
}

inline void moab_tag_reduce(
	pargal::util::moab_tag_chunk_seq &var_seq,
	std::size_t red_sz,
	MPI_Op op, MPI_Comm comm,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator)
{
	int rank;

	MPI_Comm_rank(comm, &rank);

	if(rank == 0){
		MPI_Reduce(MPI_IN_PLACE, var_seq.begin(), red_sz,
			MPI_DOUBLE, op, 0, comm);
	}
	else{
		MPI_Reduce(var_seq.begin(), NULL, red_sz,
			MPI_DOUBLE, op, 0, comm);
	}
}

inline void moab_tag_reduce(
	pargal::util::moab_tag_chunk_seq &var_seq,
	std::size_t red_sz,
	MPI_Op op, MPI_Comm comm,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_skip_dim_iterator)
{
	int rank;
	size_t reset_cnt = var_seq.get_reset_count();
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();

	MPI_Comm_rank(comm, &rank);

	std::vector<double> red_buf(red_sz);
	
	// FIXME: Is this copy-redn-copy approach better than just
	// redn on the whole chunk?

	// Copy redn data from var_seq to tmp redn buf
	pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin();
	for(std::vector<double>::iterator red_buf_iter = red_buf.begin();
		(iter_var != var_seq.end()) &&
		(red_buf_iter != red_buf.end());
		){
		std::copy(iter_var, iter_var + reset_cnt, red_buf_iter);
		std::advance(iter_var, skip_dist);
		std::advance(red_buf_iter, reset_cnt);
	}

	if(rank == 0){
		MPI_Reduce(MPI_IN_PLACE, &red_buf[0], red_sz,
			MPI_DOUBLE, op, 0, comm);
	}
	else{
		MPI_Reduce(&red_buf[0], NULL, red_sz,
			MPI_DOUBLE, op, 0, comm);
	}

	// Copy redn result back to var_seq
	if(rank == 0){
		pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin();
		for(std::vector<double>::iterator red_buf_iter = red_buf.begin();
			(iter_var != var_seq.end()) &&
			(red_buf_iter != red_buf.end());
			){
			std::copy(red_buf_iter, red_buf_iter + reset_cnt, iter_var);
			std::advance(iter_var, skip_dist);
			std::advance(red_buf_iter, reset_cnt);
		}
	}
}

// FIXME: Investigate why the redn size has to be different from seq size
inline std::size_t get_var_red_sz(pargal::pcvar &var, const std::size_t odim)
{
	pargal::util::moab_tag_dim_type odim_type = var.get_dim_type(odim);
	std::size_t red_sz = 0;
	std::vector<std::size_t> var_ldims;
	switch(odim_type){
		case pargal::util::MOAB_TAG_SQIJ_LAT_DIM:
				var.get_local_dims(var_ldims);
				red_sz = var_ldims[var_ldims.size() - 1] * var.num_levels();
				break;
		case pargal::util::MOAB_TAG_SQIJ_LON_DIM:
				var.get_local_dims(var_ldims);
				red_sz = var_ldims[var_ldims.size() - 2] * var.num_levels();
				break;
		default:
			// Assuming a zero redn size would flag an error if one
			// tries to use it
			red_sz = 0;
	}
	return red_sz;
}

typedef std::map<pargal::util::moab_tag_chunk::key_type,
					pargal::util::moab_tag_chunk,
					pargal::util::moab_tag_chunk::key_compare_functor>
		moab_tag_chunk_map_t;

template<typename T>
void dim_reduce(
	moab_tag_chunk_map_t &red_chunks,
	pargal::pcvar &var, pargal_red_op op,
	const std::size_t odim,
	pargal::util::var_red_counter &counter)
{
	dim_reduce_comm red_comm(var, odim);
	MPI_Comm red_mpi_comm = red_comm.get_mpi_comm();
	std::size_t ncounters = counter.size();
	if(red_mpi_comm != MPI_COMM_NULL){
		// We need to add up the counters
		pargal::util::var_red_counter::iterator counter_iter = counter.begin();
		int rank;
		MPI_Comm_rank(red_mpi_comm, &rank);
		for(size_t i=0; i < ncounters; i++){
			// Note: For all cases that we cover right now there is only 
			// one counter to reduce so we won't have a perf issue
			// with multiple reduces
			pargal::util::moab_tag_chunk_counter &val_counter = counter[i];
			int red_sz = static_cast<int>(val_counter.size());
			if(rank == 0){
				MPI_Reduce(MPI_IN_PLACE, &(val_counter[0]), red_sz,
					MPI_INT, MPI_SUM, 0, red_mpi_comm);	
			}
			else{
				MPI_Reduce(&(val_counter[0]), NULL, red_sz,
					MPI_INT, MPI_SUM, 0, red_mpi_comm);	
			}
		}

		MPI_Op mpi_op = MPI_OP_NULL;
		// FIXME: Shouldn't this info be in a var seq?
		std::size_t tag_red_sz = get_var_red_sz(var, odim);
		switch(op){
			case PARGAL_RED_OP_DIM_AVG:
					mpi_op = MPI_SUM;
					break;
			default:
					break;
		}
		switch(op){
			case PARGAL_RED_OP_DIM_AVG:
				// Sum up the data - if needed
				counter_iter = counter.begin();
				for(moab_tag_chunk_map_t::iterator iter = red_chunks.begin();
						(iter != red_chunks.end()) &&
						(counter_iter != counter.end());
						++iter, ++counter_iter){
					pargal::util::moab_tag_chunk &chunk = iter->second;
					int red_sz = static_cast<int>(chunk.size());

					pargal::util::moab_tag_chunk_counter tc_counter = *counter_iter;
					pargal::util::moab_tag_chunk_counter::iterator seq_citer = 
						tc_counter.begin();

					for(pargal::util::moab_tag_chunk::iterator seq_iter = chunk.begin();
							(seq_iter != chunk.end()) &&
							(seq_citer != tc_counter.end());
							++seq_iter, ++seq_citer){
						moab_tag_reduce(*seq_iter, tag_red_sz, mpi_op, red_mpi_comm,
							typename pargal::util::dim_traits<T>::dim_iterator_category(),
							typename pargal::util::dim_traits<T>::dim_red_iterator_category());
					}
				}
				break;
			// FIXME: Implement the ops below
			case PARGAL_RED_OP_DIM_MAX:
			case PARGAL_RED_OP_DIM_MIN:
			default:
				throw pargal::pargal_except("Error during reduce : The op is not implemented yet",
						__FILE__, __LINE__, "Unknown");	
		}
	}

	switch(op){
		case PARGAL_RED_OP_DIM_AVG:
			{
				// Now that we have the counter and data summed up
				// - perform dim avg = sum/count
				moab_tag_chunk_map_t::iterator var_iter = red_chunks.begin();
				pargal::util::var_red_counter::iterator counter_iter = counter.begin();

				for(;
					(var_iter != red_chunks.end()) && (counter_iter != counter.end());
					++var_iter, ++counter_iter){
					pargal::util::moab_tag_chunk& chunk = var_iter->second;
					pargal::util::moab_tag_chunk_counter& tc_counter = *counter_iter;
					pargal::util::moab_tag_chunk_counter::iterator seq_citer =
						tc_counter.begin();
					for(pargal::util::moab_tag_chunk::iterator seq_iter = chunk.begin();
							(seq_iter != chunk.end()) &&
							(seq_citer != tc_counter.end());
							++seq_iter, ++seq_citer){
						algorithm::util::distributed_divide(
							*seq_iter, *seq_citer,
							typename pargal::util::dim_traits<T>::dim_iterator_category(),
							typename pargal::util::dim_traits<T>::dim_red_iterator_category());
					}
				}
			}
			break;
		// FIXME: Implement the ops below
		case PARGAL_RED_OP_DIM_MAX:
		case PARGAL_RED_OP_DIM_MIN:
		default:
			throw pargal::pargal_except("Error during reduce : The op is not implemented yet",
					__FILE__, __LINE__, "Unknown");	
	}
}

template<typename T>
void dim_reduce(
	moab_tag_chunk_map_t &red_chunks,
	pargal::pcvar &var, pargal_red_op op,
	const std::size_t odim)
{
	dim_reduce_comm red_comm(var, odim);
	MPI_Comm red_mpi_comm = red_comm.get_mpi_comm();
	if(red_mpi_comm != MPI_COMM_NULL){
		int rank;
		MPI_Comm_rank(red_mpi_comm, &rank);

		MPI_Op mpi_op = MPI_OP_NULL;
		// FIXME: Shouldn't this info be in a var seq?
		std::size_t tag_red_sz = get_var_red_sz(var, odim);
		switch(op){
			case PARGAL_RED_OP_DIM_MIN:
					mpi_op = MPI_MIN;
					break;
			case PARGAL_RED_OP_DIM_MAX:
					mpi_op = MPI_MAX;
					break;
			case PARGAL_RED_OP_DIM_SUM:
					mpi_op = MPI_SUM;
					break;
			case PARGAL_RED_OP_DIM_PRODUCT:
					mpi_op = MPI_PROD;
					break;
			default:
					break;
		}
		switch(op){
			case PARGAL_RED_OP_DIM_MIN:
			case PARGAL_RED_OP_DIM_MAX:
			case PARGAL_RED_OP_DIM_SUM:
			case PARGAL_RED_OP_DIM_PRODUCT:
				// Sum up the data - if needed
				for(moab_tag_chunk_map_t::iterator iter = red_chunks.begin();
						(iter != red_chunks.end());
						++iter){
					pargal::util::moab_tag_chunk &chunk = iter->second;
					int red_sz = static_cast<int>(chunk.size());

					for(pargal::util::moab_tag_chunk::iterator seq_iter = chunk.begin();
							(seq_iter != chunk.end());
							++seq_iter){
						moab_tag_reduce(*seq_iter, tag_red_sz, mpi_op, red_mpi_comm,
							typename pargal::util::dim_traits<T>::dim_iterator_category(),
							typename pargal::util::dim_traits<T>::dim_red_iterator_category());
					}
				}
				break;
			default:
				throw pargal::pargal_except("Error during reduce : The op is not implemented yet",
						__FILE__, __LINE__, "Unknown");	
		}
	}
}

} //util
}//algorithm
} // pargal

#endif // PARGAL_MOAB_TAG_REDUCE_HPP_INCLUDED
