/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_MOAB_TAG_SUM_HPP_INCLUDED
#define PARGAL_MOAB_TAG_SUM_HPP_INCLUDED

#include <pargal/pcvar.hpp>
#include <pargal/moab_tag_iterator.hpp>
#include <pargal/moab_tag_chunk.hpp>
#include <pargal/moab_tag_counter.hpp>

namespace pargal{
namespace algorithm{
namespace util{

// ============ DISTRIBUTED SUM with counting ==================
// Distributed sum - with missing vals
template<typename dim_iter, typename red_dim_iter>
inline void distributed_sum_m(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_seq_counter &counter)
{
	throw pargal::pargal_except("Error: Unsupported iterator type. Cannot perform distributed sum", __FILE__, __LINE__, "Unknown");
}

template<>
inline void distributed_sum_m<pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_seq_counter &counter)
{
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_sum_seq.get_static_count();
	double mval = var_seq.get_missing_val();

	assert(counter.size() == 1);
	pargal::util::moab_tag_seq_counter::iterator citer = counter.begin();
	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end());){
		size_t j=0;

		// Unchecked advances of iter_var is dangerous - however a
		// safe advance might be more costly
		for(j=0; 
			(j<static_cnt) &&
			(iter_var != var_seq.end());
			j++){
			// FIXME: Introduce value_type to use for temp
			if(*iter_var_sum == mval){
				if(*iter_var != mval){
					*iter_var_sum = *iter_var;
					*citer += 1;
				}
			}
			else if(*iter_var != mval){	
				*iter_var_sum += *iter_var;
				*citer++;
			}
			std::advance(iter_var, skip_dist);
		}
		if(j != skip_cnt){
			for(;j != skip_cnt; j++){
				std::advance(iter_var, skip_dist);
			}
		}
		iter_var_sum++;
	}
}

template<>
inline void distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_seq_counter &counter)
{
	size_t reset_cnt = var_sum_seq.get_reset_count();
	pargal::util::moab_tag_seq_counter::iterator citer = counter.begin();
	double mval = var_seq.get_missing_val();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end()) &&
		(citer != counter.end());){

		pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_sum;
		pargal::util::moab_tag_seq_counter::iterator reset_citer = citer;
		for(size_t i = 0;
			(i < reset_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end()) &&
			(citer != counter.end());
			++iter_var, ++iter_var_sum, ++citer, i++){
			if(*iter_var_sum == mval){
				if(*iter_var != mval){
					*iter_var_sum = *iter_var;
					*citer += 1;
				}
			}
			else if(*iter_var != mval){
				*iter_var_sum += *iter_var;	
				*citer += 1;
			}
		}
		iter_var_sum = reset_iter;
		citer = reset_citer;
	}
}

template<>
inline void distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_seq_counter &counter)
{
	size_t static_cnt = var_sum_seq.get_static_count();
	size_t skip_cnt = var_sum_seq.get_skip_count();
	size_t skip_dist = var_sum_seq.get_skip_dist();
	double mval = var_seq.get_missing_val();
	pargal::util::moab_tag_seq_counter::iterator citer = counter.begin();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end()) &&
		(citer != counter.end());){

		for(size_t i = 0;
			(i < static_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end()) &&
			(citer != counter.end());
			++iter_var, i++){
			if(*iter_var_sum == mval){
				if(*iter_var != mval){
					*iter_var_sum = *iter_var;
					*citer += 1;
				}
			}
			else if(*iter_var != mval){
				*iter_var_sum += *iter_var;	
				*citer += 1;
			}
		}
		//skip
		std::advance(iter_var_sum, skip_dist);
		++citer;
	}
}

template<>
inline void distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_skip_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_seq_counter &counter)
{
	size_t reset_cnt = var_sum_seq.get_reset_count();
	size_t skip_cnt = var_sum_seq.get_skip_count();
	size_t skip_dist = var_sum_seq.get_skip_dist();
	double mval = var_seq.get_missing_val();
	pargal::util::moab_tag_seq_counter::iterator citer = counter.begin();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) && (iter_var_sum != var_sum_seq.end());){
		for(size_t i = 0;
			(i < skip_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end()) &&
			(citer != counter.end());
			){

			pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_sum;
			pargal::util::moab_tag_seq_counter::iterator reset_citer = citer;
			for(size_t j = 0;
				(j < reset_cnt) &&
				(iter_var != var_seq.end()) &&
				(iter_var_sum != var_sum_seq.end()) &&
				(citer != counter.end());
				++iter_var, ++iter_var_sum, ++citer, i++, j++){
				if(*iter_var_sum == mval){
					if(*iter_var != mval){
						*iter_var_sum = *iter_var;
						*citer += 1;
					}
				}
				else if(*iter_var != mval){
					*iter_var_sum += *iter_var;
					*citer += 1;
				}
			}
			//reset
			iter_var_sum = reset_iter;
			if(i < skip_cnt){
				citer = reset_citer;
			}
		}
		// skip
		std::advance(iter_var_sum, skip_dist);
	}
}

inline void distributed_sum(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_seq_counter &counter,
	pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_sum_m<pargal::util::moab_tag_sparse_skip_dim_iterator,
			pargal::util::moab_tag_sparse_static_dim_iterator>(
			var_seq, var_sum_seq, counter);
		return;
	}
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_sum_seq.get_static_count();

	assert(counter.size() == 1);
	pargal::util::moab_tag_seq_counter::iterator citer = counter.begin();
	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end());){
		size_t j=0;

		// Unchecked advances of iter_var is dangerous - however a
		// safe advance might be more costly
		for(j=0; 
			(j<static_cnt) &&
			(iter_var != var_seq.end());
			j++){
			*iter_var_sum += *iter_var;
			*citer++;
			std::advance(iter_var, skip_dist);
		}
		if(j != skip_cnt){
			for(;j != skip_cnt; j++){
				std::advance(iter_var, skip_dist);
			}
		}
		iter_var_sum++;
	}
}

inline void distributed_sum(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_seq_counter &counter,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_reset_dim_iterator>(
			var_seq, var_sum_seq, counter);
		return;
	}
	size_t reset_cnt = var_sum_seq.get_reset_count();
	pargal::util::moab_tag_seq_counter::iterator citer = counter.begin();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end()) &&
		(citer != counter.end());){

		pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_sum;
		pargal::util::moab_tag_seq_counter::iterator reset_citer = citer;
		for(size_t i = 0;
			(i < reset_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end()) &&
			(citer != counter.end());
			++iter_var, ++iter_var_sum, ++citer, i++){
			*iter_var_sum += *iter_var;	
			*citer += 1;
		}
		//reset
		iter_var_sum = reset_iter;
		citer = reset_citer;
	}
}

inline void distributed_sum(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_seq_counter &counter,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_static_skip_dim_iterator>(
			var_seq, var_sum_seq, counter);
		return;
	}
	size_t static_cnt = var_sum_seq.get_static_count();
	size_t skip_cnt = var_sum_seq.get_skip_count();
	size_t skip_dist = var_sum_seq.get_skip_dist();
	pargal::util::moab_tag_seq_counter::iterator citer = counter.begin();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end()) &&
		(citer != counter.end());){

		for(size_t i = 0;
			(i < static_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end()) &&
			(citer != counter.end());
			++iter_var, i++){
			*iter_var_sum += *iter_var;	
			*citer += 1;
		}
		//skip
		std::advance(iter_var_sum, skip_dist);
		++citer;
	}
}

inline void distributed_sum(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_seq_counter &counter,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_skip_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_reset_skip_dim_iterator>(
			var_seq, var_sum_seq, counter);
		return;
	}
	size_t reset_cnt = var_sum_seq.get_reset_count();
	size_t skip_cnt = var_sum_seq.get_skip_count();
	size_t skip_dist = var_sum_seq.get_skip_dist();
	pargal::util::moab_tag_seq_counter::iterator citer = counter.begin();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) && (iter_var_sum != var_sum_seq.end());){
		for(size_t i = 0;
			(i < skip_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end()) &&
			(citer != counter.end());
			){

			pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_sum;
			pargal::util::moab_tag_seq_counter::iterator reset_citer = citer;
			for(size_t j = 0;
				(j < reset_cnt) &&
				(iter_var != var_seq.end()) &&
				(iter_var_sum != var_sum_seq.end()) &&
				(citer != counter.end());
				++iter_var, ++iter_var_sum, ++citer, i++, j++){

				*iter_var_sum += *iter_var;
				*citer += 1;
			}
			//reset
			iter_var_sum = reset_iter;
			// FIXME: Get rid of this counter check
			if(i < skip_cnt){
				citer = reset_citer;
			}
		}
		// skip
		std::advance(iter_var_sum, skip_dist);
	}
}

// FIXME: There is very little difference btw sum with and without
// counting. These functions below should be autogenerated

// ============ DISTRIBUTED SUM without counting ==================
// Distributed sum - with missing vals
template<typename dim_iter, typename red_dim_iter>
inline void distributed_sum_m(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq)
{
	throw pargal::pargal_except("Error: Unsupported iterator type. Cannot perform distributed sum", __FILE__, __LINE__, "Unknown");
}

template<>
inline void distributed_sum_m<pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq)
{
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_sum_seq.get_static_count();
	double mval = var_seq.get_missing_val();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end());){
		size_t j=0;

		// Unchecked advances of iter_var is dangerous - however a
		// safe advance might be more costly
		for(j=0; 
			(j<static_cnt) &&
			(iter_var != var_seq.end());
			j++){
			// FIXME: Introduce value_type to use for temp
			if(*iter_var_sum == mval){
				*iter_var_sum = *iter_var;
			}
			else if(*iter_var != mval){	
				*iter_var_sum += *iter_var;
			}
			std::advance(iter_var, skip_dist);
		}
		if(j != skip_cnt){
			for(;j != skip_cnt; j++){
				std::advance(iter_var, skip_dist);
			}
		}
		iter_var_sum++;
	}
}

template<>
inline void distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq)
{
	size_t reset_cnt = var_sum_seq.get_reset_count();
	double mval = var_seq.get_missing_val();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end()) ;){

		pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_sum;
		for(size_t i = 0;
			(i < reset_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end());
			++iter_var, ++iter_var_sum, i++){
			if(*iter_var_sum == mval){
				*iter_var_sum = *iter_var;
			}
			else if(*iter_var != mval){	
				*iter_var_sum += *iter_var;
			}
		}
		iter_var_sum = reset_iter;
	}
}

template<>
inline void distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq)
{
	size_t static_cnt = var_sum_seq.get_static_count();
	size_t skip_cnt = var_sum_seq.get_skip_count();
	size_t skip_dist = var_sum_seq.get_skip_dist();
	double mval = var_seq.get_missing_val();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end());){

		for(size_t i = 0;
			(i < static_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end());
			++iter_var, i++){
			if(*iter_var_sum == mval){
				*iter_var_sum = *iter_var;
			}
			else if(*iter_var != mval){	
				*iter_var_sum += *iter_var;
			}
		}
		//skip
		std::advance(iter_var_sum, skip_dist);
	}
}

template<>
inline void distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_skip_dim_iterator>(
	pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq)
{
	size_t reset_cnt = var_sum_seq.get_reset_count();
	size_t skip_cnt = var_sum_seq.get_skip_count();
	size_t skip_dist = var_sum_seq.get_skip_dist();
	double mval = var_seq.get_missing_val();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) && (iter_var_sum != var_sum_seq.end());){
		for(size_t i = 0;
			(i < skip_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end());){

			pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_sum;
			for(size_t j = 0;
				(j < reset_cnt) &&
				(iter_var != var_seq.end()) &&
				(iter_var_sum != var_sum_seq.end());
				++iter_var, ++iter_var_sum, i++, j++){
				if(*iter_var_sum == mval){
					*iter_var_sum = *iter_var;
				}
				else if(*iter_var != mval){	
					*iter_var_sum += *iter_var;
				}
			}
			//reset
			iter_var_sum = reset_iter;
		}
		// skip
		std::advance(iter_var_sum, skip_dist);
	}
}

inline void distributed_sum(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_sparse_skip_dim_iterator,
	pargal::util::moab_tag_sparse_static_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_sum_m<pargal::util::moab_tag_sparse_skip_dim_iterator,
			pargal::util::moab_tag_sparse_static_dim_iterator>(
			var_seq, var_sum_seq);
		return;
	}
	size_t skip_cnt = var_seq.get_skip_count();
	size_t skip_dist = var_seq.get_skip_dist();
	size_t static_cnt = var_sum_seq.get_static_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end());){
		size_t j=0;

		// Unchecked advances of iter_var is dangerous - however a
		// safe advance might be more costly
		for(j=0; 
			(j<static_cnt) &&
			(iter_var != var_seq.end());
			j++){
			*iter_var_sum += *iter_var;
			std::advance(iter_var, skip_dist);
		}
		if(j != skip_cnt){
			for(;j != skip_cnt; j++){
				std::advance(iter_var, skip_dist);
			}
		}
		iter_var_sum++;
	}
}

inline void distributed_sum(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_reset_dim_iterator>(
			var_seq, var_sum_seq);
		return;
	}
	size_t reset_cnt = var_sum_seq.get_reset_count();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end());){

		pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_sum;
		for(size_t i = 0;
			(i < reset_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end());
			++iter_var, ++iter_var_sum, i++){
			*iter_var_sum += *iter_var;	
		}
		//reset
		iter_var_sum = reset_iter;
	}
}

inline void distributed_sum(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_static_skip_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_static_skip_dim_iterator>(
			var_seq, var_sum_seq);
		return;
	}
	size_t static_cnt = var_sum_seq.get_static_count();
	size_t skip_cnt = var_sum_seq.get_skip_count();
	size_t skip_dist = var_sum_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) &&
		(iter_var_sum != var_sum_seq.end());){

		for(size_t i = 0;
			(i < static_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end());
			++iter_var, i++){
			*iter_var_sum += *iter_var;	
		}
		//skip
		std::advance(iter_var_sum, skip_dist);
	}
}

inline void distributed_sum(pargal::util::moab_tag_chunk_seq &var_seq,
	pargal::util::moab_tag_chunk_seq &var_sum_seq,
	pargal::util::moab_tag_gen_dim_iterator,
	pargal::util::moab_tag_reset_skip_dim_iterator)
{
	if(var_seq.has_missing_val()){
		distributed_sum_m<pargal::util::moab_tag_gen_dim_iterator,
			pargal::util::moab_tag_reset_skip_dim_iterator>(
			var_seq, var_sum_seq);
		return;
	}
	size_t reset_cnt = var_sum_seq.get_reset_count();
	size_t skip_cnt = var_sum_seq.get_skip_count();
	size_t skip_dist = var_sum_seq.get_skip_dist();

	for(pargal::util::moab_tag_chunk_seq::iterator iter_var = var_seq.begin(),
		iter_var_sum = var_sum_seq.begin();
		(iter_var != var_seq.end()) && (iter_var_sum != var_sum_seq.end());){
		for(size_t i = 0;
			(i < skip_cnt) &&
			(iter_var != var_seq.end()) &&
			(iter_var_sum != var_sum_seq.end());){

			pargal::util::moab_tag_chunk_seq::iterator reset_iter = iter_var_sum;
			for(size_t j = 0;
				(j < reset_cnt) &&
				(iter_var != var_seq.end()) &&
				(iter_var_sum != var_sum_seq.end());
				++iter_var, ++iter_var_sum, i++, j++){

				*iter_var_sum += *iter_var;
			}
			//reset
			iter_var_sum = reset_iter;
		}
		// skip
		std::advance(iter_var_sum, skip_dist);
	}
}

} //namespace util
} //namespace algorithm
} // namespace pargal

#endif // PARGAL_MOAB_TAG_SUM_HPP_INCLUDED
