/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <pargal/pcalgo.hpp>

namespace pargal 
{
  int pcalgo::numCubPoints;
  shards::CellTopology pcalgo::cellType;
  Intrepid::FieldContainer<double> pcalgo::basisValues;
  Intrepid::FieldContainer<double> pcalgo::basisGrads;
  Intrepid::FieldContainer<double> pcalgo::cubPoints;
  Intrepid::FieldContainer<double> pcalgo::cubWeights;
  int pcalgo::numFields;
  double pcalgo::radius = 6.37122e6;

  void 
  pcalgo::init_bilinear_quad()
  {    
    // Quad cell topology
    cellType = shards::CellTopology(shards::getCellTopologyData<shards::Quadrilateral<4> >());  

    // Integration points and weights
    int cubDegree = 2;
    Intrepid::DefaultCubatureFactory<double> cubFactory;
    Teuchos::RCP<Intrepid::Cubature<double> > cellCubature = cubFactory.create(cellType, cubDegree);
    numCubPoints = cellCubature -> getNumPoints();
    cubPoints.resize(numCubPoints, cellCubature -> getDimension());
    cubWeights.resize(numCubPoints);
    cellCubature -> getCubature(cubPoints, cubWeights);

    // Basis function
    Intrepid::Basis_HGRAD_QUAD_C1_FEM<double, Intrepid::FieldContainer<double> > bilinearBasisQuad;
    numFields = bilinearBasisQuad.getCardinality();

    // Basis values at cubature points
    basisValues.resize(bilinearBasisQuad.getCardinality(), numCubPoints);
    bilinearBasisQuad.getValues(basisValues, cubPoints, Intrepid::OPERATOR_VALUE);    

    // Basis gradients at cubature points
    basisGrads.resize(bilinearBasisQuad.getCardinality(), 
		      numCubPoints, cellType.getDimension());
    bilinearBasisQuad.getValues(basisGrads, cubPoints, Intrepid::OPERATOR_GRAD);    

  }

  void 
  pcalgo::vorticity(pcvar& var_u, pcvar& var_v, pcvar& var_vort)
  {    
    std::string grid_type = var_u.get_finfo().get_grid_type();
    if (grid_type == "CAM_EUL" || grid_type == "CAM_FV") {
       vort_div_cell_center(var_u, var_v, var_vort, true);
    }
    else if (grid_type == "CAM_SE") {
       vort_div_nodes_spectral(var_u, var_v, var_vort, true);
    }
  }
  
  void 
  pcalgo::divergence(pcvar& var_u, pcvar& var_v, pcvar& var_div)
  {    
    std::string grid_type = var_u.get_finfo().get_grid_type();
    if (grid_type == "CAM_EUL" || grid_type == "CAM_FV") {
       vort_div_cell_center(var_u, var_v, var_div, false);
    }
    else if (grid_type == "CAM_SE") {
       vort_div_nodes_spectral(var_u, var_v, var_div, false);
    }
  }
    
  void 
  pcalgo::vort_div_cell_center(pcvar& var_u, pcvar& var_v, pcvar& var_vort_div, 
			       bool is_algo_vort)
  {    

    // Check for missing values
    if (var_u.has_missing_val() || var_v.has_missing_val()) {
      throw pargal_except("Error: voriticity/divergence algorithm does not accept missing values", 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    }

    init_bilinear_quad();

    // Get cell and basis dimensions 
    int numNodesPerElem = cellType.getNodeCount();
    int spaceDim = cellType.getDimension();

    // Get local mesh dimensions 
    std::vector<int> lCDims = var_u.get_dim_loc_minmax_owned();
    std::size_t num_lon = (lCDims[1] - lCDims[0] + 1);
    std::size_t num_lat = (lCDims[3] - lCDims[2] + 1);
    std::size_t num_lev = var_u.num_levels();
    std::size_t num_tsteps = var_u.get_finfo().get_dim("time").size();


    // Get moab
    moab::Interface& mb = var_u.get_finfo().get_moab();
    moab::EntityHandle& file_set = var_u.get_finfo().get_file_set();
    const moab::ParallelComm& mb_pcomm = var_u.get_finfo().get_pcomm();

    Epetra_MpiComm Comm(const_cast<moab::ParallelComm&>(mb_pcomm).comm() );
    
    // Get local elements
    moab::Range elems; 
    moab::ErrorCode rval;
    rval = mb.get_entities_by_type(file_set, moab::MBQUAD, elems);
    if (rval != moab::MB_SUCCESS)
      throw pargal_except("Error: " + mb.get_error_string(rval), 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);

    if (elems.empty()) {
         throw pargal_except("Error: file contained no quadrilaterals",
        		      __FILE__, __LINE__, __PRETTY_FUNCTION__);
    }

    // Get owned elements
    moab::Range elems_owned;
    rval = const_cast<moab::ParallelComm&>(mb_pcomm).filter_pstatus(elems, PSTATUS_NOT_OWNED, PSTATUS_NOT, -1, &elems_owned);

    // Get local vertices 
    moab::Range verts;
    rval = mb.get_entities_by_type(file_set, moab::MBVERTEX, verts);

    // Get owned vertices
    moab::Range verts_owned;
    rval = const_cast<moab::ParallelComm&>(mb_pcomm).filter_pstatus(verts, PSTATUS_NOT_OWNED, PSTATUS_NOT, -1, &verts_owned);

    // Get tags for global id and set global ids 
    moab::Tag globalIdTag;
    rval=mb.tag_get_handle("GLOBAL_ID",1,moab::MB_TYPE_INTEGER,globalIdTag);
    rval = const_cast<moab::ParallelComm&>(mb_pcomm).assign_global_ids(file_set,2,1,true,true,false);

    // Tags for cell center coords
    moab::Tag coordTag;
    rval=mb.tag_get_handle("COORDS", 3, moab::MB_TYPE_DOUBLE, coordTag);

    // Get shared entities and exchange tags
    moab::Range shared_ents;
    rval = const_cast<moab::ParallelComm&>(mb_pcomm).get_shared_entities(-1,shared_ents,2,false,false);
    rval = const_cast<moab::ParallelComm&>(mb_pcomm).exchange_tags(globalIdTag, shared_ents);
    rval = const_cast<moab::ParallelComm&>(mb_pcomm).exchange_tags(coordTag, shared_ents);


    ////////////////////////////////////////////////////
    // Set up dual mesh
    ////////////////////////////////////////////////////

    // Count dual elements
    int idualelem = 0;
    for (moab::Range::iterator itNode = verts_owned.begin(); itNode != verts_owned.end(); ++itNode) {
        // get vertex handle
          moab::EntityHandle nodeHandle = *itNode;
          std::vector<moab::EntityHandle> dualNodes;
          rval = mb.get_adjacencies(&nodeHandle, 1, 2, true, dualNodes);
          if (dualNodes.size() == 4) {
             idualelem++;
          }
     }
    int numDualElems = idualelem;

    // Set up arrays for dual mesh
    Intrepid::FieldContainer<double> dualElemCoords(numDualElems,4,2);
    Intrepid::FieldContainer<double> dualElemToNode(numDualElems,4);
    Intrepid::FieldContainer<int> dualElemSortMap(numDualElems,4);
    
    idualelem = 0;
    for (moab::Range::iterator itNode = verts_owned.begin(); itNode != verts_owned.end(); ++itNode) {

          moab::EntityHandle nodeHandle = *itNode;
          std::vector<moab::EntityHandle> dualNodes;
          rval = mb.get_adjacencies(&nodeHandle, 1, 2, true, dualNodes);

          if (dualNodes.size() == 4) {
             Intrepid::FieldContainer<double> coord(dualNodes.size(), spaceDim);
             Intrepid::FieldContainer<int> cid(dualNodes.size());
             for (std::size_t i=0; i<dualNodes.size(); i++) {

                double cellCenter[3];
                rval = mb.tag_get_data(coordTag, &dualNodes[i], 1, &cellCenter);
                coord(i,0) = M_PI*cellCenter[0]/180.0;
                coord(i,1) = M_PI*cellCenter[1]/180.0;

                int cellId;
                rval = mb.tag_get_data(globalIdTag, &dualNodes[i], 1, &cellId);
                cid(i) = cellId-1;
             }

             // Sort dual cell vertices assuming structured mesh
             int next = -1;
             int last = -1;
             int mid = -1;
             int start = 0;
             for (int i=1; i < 4; i++){
                 if (coord(i,0) <= coord(start,0) && coord(i,1) <= coord(start,1)) start = i;
             }
             for (int i=0; i < 4; i++){
                 if (start != i) {
                   if (coord(i,1) - coord(start,1) < 1e-8) next = i;
                   if (coord(i,0) - coord(start,0) < 1e-8) last = i;
                   if ((coord(i,0) - coord(start,0) > 1e-8) && (coord(i,1) - coord(start,1)> 1e-8)) mid = i;
                 }
              }
              dualElemSortMap(idualelem,0) = start;
              dualElemSortMap(idualelem,1) = next;
              dualElemSortMap(idualelem,2) = mid;
              dualElemSortMap(idualelem,3) = last;

              for (std::size_t i=0; i<dualNodes.size(); i++) {
                 double lambda = coord(dualElemSortMap(idualelem,i),0);
                 double theta = coord(dualElemSortMap(idualelem,i),1);
                 dualElemCoords(idualelem,i,0) = lambda;
                 dualElemCoords(idualelem,i,1) = theta;
                 dualElemToNode(idualelem,i) = cid(dualElemSortMap(idualelem,i));
              }

              // adjust coordinates for periodic boundary cells
               if ((dualElemCoords(idualelem,0,0) - dualElemCoords(idualelem,1,0)) > M_PI){
                    dualElemCoords(idualelem,1,0) = dualElemCoords(idualelem,1,0)+2.0*M_PI;
                    dualElemCoords(idualelem,2,0) = dualElemCoords(idualelem,2,0)+2.0*M_PI;
               }
               else if ((dualElemCoords(idualelem,0,0) - dualElemCoords(idualelem,1,0)) < -M_PI) {
                    dualElemCoords(idualelem,1,0) = dualElemCoords(idualelem,1,0)-2.0*M_PI;
                    dualElemCoords(idualelem,2,0) = dualElemCoords(idualelem,2,0)-2.0*M_PI;
               }

               idualelem++;

          } // if dualNodes.size() = 4
     } // end loop over primary mesh owned nodes


    ////////////////////////////////////////////////////
    // Build mass matrix
    ////////////////////////////////////////////////////

    // Owned elements are dual mesh nodes
    Epetra_Map globalMapDualNode(-1,elems_owned.size(),0,Comm);

    // Mass matrix and rhs
    Epetra_FECrsMatrix MassMatrix(Copy, globalMapDualNode, numNodesPerElem*10);
    Epetra_FEVector rhsVector(globalMapDualNode);

    // Cell Jacobian, inverse, and determinant
    Intrepid::FieldContainer<double> cellJacobian (numDualElems, numCubPoints, spaceDim, spaceDim);
    Intrepid::FieldContainer<double> cellJacobInv (numDualElems, numCubPoints, spaceDim, spaceDim);
    Intrepid::FieldContainer<double> cellJacobDet (numDualElems, numCubPoints);
    Intrepid::CellTools<double>::setJacobian(cellJacobian, cubPoints, dualElemCoords, cellType);
    Intrepid::CellTools<double>::setJacobianInv(cellJacobInv, cellJacobian );
    Intrepid::CellTools<double>::setJacobianDet(cellJacobDet, cellJacobian );

    // Map integration points to physical space
    Intrepid::FieldContainer<double> physPoints (numDualElems, numCubPoints, spaceDim);
    Intrepid::CellTools<double>::mapToPhysicalFrame (physPoints, cubPoints, dualElemCoords, cellType);

    // Transform basis values and gradients to physical frame
    Intrepid::FieldContainer<double> physBasisGrads (numDualElems, numFields, numCubPoints, spaceDim);
    Intrepid::FieldContainer<double> physBasisVals (numDualElems, numFields, numCubPoints);
    Intrepid::FunctionSpaceTools::HGRADtransformGRAD<double>(physBasisGrads, cellJacobInv, basisGrads);
    Intrepid::FunctionSpaceTools::HGRADtransformVALUE<double>(physBasisVals, basisValues);

    // Compute cell measure
    Intrepid::FieldContainer<double> cellMeasure (numDualElems, numCubPoints);
    Intrepid::FunctionSpaceTools::computeCellMeasure<double>(cellMeasure, cellJacobDet, cubWeights);

    // Compute cell measure with cosphi
    Intrepid::FieldContainer<double> cellMeasureWithCosphi (numDualElems, numCubPoints);
    for (int ielem=0; ielem < numDualElems; ielem++) {
       for (int ipt=0; ipt < numCubPoints; ipt++ ) {
         double cosphi = cos(physPoints(ielem,ipt,1));
         cellMeasureWithCosphi(ielem,ipt) = cellMeasure(ielem,ipt)*cosphi;
        }
     }

    // Multiply transformed values with weighted measure
    Intrepid::FieldContainer<double> physBasisValsWeighted (numDualElems, numFields, numCubPoints);
    Intrepid::FunctionSpaceTools::multiplyMeasure<double>(physBasisValsWeighted, cellMeasure, physBasisVals);

    // Multiply transformed values with weighted measure times cosphi
    Intrepid::FieldContainer<double> physBasisValsWeightedCosphi (numDualElems, numFields, numCubPoints);
    Intrepid::FunctionSpaceTools::multiplyMeasure<double>(physBasisValsWeightedCosphi,
                                                          cellMeasureWithCosphi, physBasisVals);

    // Integrate to compute element mass matrix
    Intrepid::FieldContainer<double> elemMassMatrix (numDualElems, numFields, numFields);
    Intrepid::FunctionSpaceTools::integrate<double>(elemMassMatrix, physBasisValsWeightedCosphi,
                                                    physBasisVals, Intrepid::COMP_BLAS);

    // Loop over cells to assemble global matrix/vector
    for (int ielem = 0; ielem < numDualElems; ielem++) {

       std::vector<int> nodeIndex(numFields);
       for (int cellCol = 0; cellCol < numFields; cellCol++){
          nodeIndex[cellCol] = dualElemToNode(ielem,cellCol);
       }
       for (int cellRow = 0; cellRow < numFields; cellRow++){
	 //int globalRow = dualElemToNode(ielem,cellRow);
          MassMatrix.InsertGlobalValues(1, &nodeIndex[cellRow], numFields, &nodeIndex[0], &elemMassMatrix(ielem,cellRow,0));
       }
    }  // end loop over cells
      
    // Assemble over multiple processors
    MassMatrix.GlobalAssemble();
    MassMatrix.FillComplete();

    // Build ML preconditioner for use in computing vorticity/divergence
    Teuchos::ParameterList MLList;
    ML_Epetra::SetDefaults("SA",MLList);
    MLList.set("ML output",1);
    MLList.set("max levels",10);
    MLList.set("coarse: type","Amesos-KLU");
    ML_Epetra::MultiLevelPreconditioner *MLPrec = new ML_Epetra::MultiLevelPreconditioner(MassMatrix, MLList, true);


    ////////////////////////////////////////////////////
    // Compute vorticity or divergence
    ////////////////////////////////////////////////////

    // Loop over time
    for (std::size_t itime = 0; itime < num_tsteps; itime++) {

       // Get tags for velocity data and exchange data
       moab::Tag uvelTag = var_u.get_moab_tag(itime);
       moab::Tag vvelTag = var_v.get_moab_tag(itime);
       rval = const_cast<moab::ParallelComm&>(mb_pcomm).exchange_tags(uvelTag, shared_ents);
       rval = const_cast<moab::ParallelComm&>(mb_pcomm).exchange_tags(vvelTag, shared_ents);

       // Get pointer to vort_div data
       double* var_vort_div_ptr = var_vort_div.get_storage(itime);

       // Loop over levels
       for (std::size_t ilev = 0; ilev < num_lev; ilev++) {

          rhsVector.PutScalar(0.0);

          // loop over dual cells to get velocity and compute vorticity or divergence
          Intrepid::FieldContainer<double> vort_div(numDualElems, numCubPoints);
          Intrepid::FieldContainer<double> elemRHSVector(numDualElems, numFields);
          idualelem = 0;
          for (moab::Range::iterator itNode = verts_owned.begin(); itNode != verts_owned.end(); ++itNode) {

             moab::EntityHandle nodeHandle = *itNode;
             std::vector<moab::EntityHandle> dualNodes;
             rval = mb.get_adjacencies(&nodeHandle, 1, 2, true, dualNodes);

             if (dualNodes.size() == 4) {

               Intrepid::FieldContainer<double> vel(dualNodes.size(), spaceDim);
               for (std::size_t i=0; i<dualNodes.size(); i++) {
                   std::vector<double> uvel(num_lev);
                   std::vector<double> vvel(num_lev);
                   rval = mb.tag_get_data(uvelTag, &dualNodes[i], 1, &uvel[0]);
                   rval = mb.tag_get_data(vvelTag, &dualNodes[i], 1, &vvel[0]);
                   vel(i,0) = uvel[ilev];
                   vel(i,1) = vvel[ilev];
               }

               for (int ipt=0; ipt < numCubPoints; ipt++ ) {

                  double sinphi = sin(physPoints(idualelem,ipt,1));
                  double cosphi = cos(physPoints(idualelem,ipt,1));

                  if (is_algo_vort) {
                    double uVelocity = 0.0;
                    double uVelGrady = 0.0;
                    double vVelGradx = 0.0;
                    for (int inode=0; inode < numNodesPerElem; inode++) {
                        uVelocity += vel(dualElemSortMap(idualelem,inode),0)*physBasisVals(idualelem,inode,ipt);
                        uVelGrady += vel(dualElemSortMap(idualelem,inode),0)*physBasisGrads(idualelem,inode,ipt,1);
                        vVelGradx += vel(dualElemSortMap(idualelem,inode),1)*physBasisGrads(idualelem,inode,ipt,0);
                    }
                    vort_div(idualelem,ipt) = (vVelGradx + uVelocity*sinphi - uVelGrady*cosphi)/radius;
                  }
                  else
                  {
                    double vVelocity = 0.0;
                    double uVelGradx = 0.0;
                    double vVelGrady = 0.0;
                    for (int inode=0; inode < numNodesPerElem; inode++) {
                        vVelocity += vel(dualElemSortMap(idualelem,inode),1)*physBasisVals(idualelem,inode,ipt);
                        uVelGradx += vel(dualElemSortMap(idualelem,inode),0)*physBasisGrads(idualelem,inode,ipt,0);
                        vVelGrady += vel(dualElemSortMap(idualelem,inode),1)*physBasisGrads(idualelem,inode,ipt,1);
                    }
                    vort_div(idualelem,ipt) = (uVelGradx - vVelocity*sinphi + vVelGrady*cosphi)/radius;
                  }

               } // end loop over cubature points

               idualelem++;

             } // end if dualNodes.size()=4

          } // end loop over owned nodes

          Intrepid::FunctionSpaceTools::integrate<double>(elemRHSVector, vort_div, physBasisValsWeighted,
                                                           Intrepid::COMP_BLAS);

          // Loop over cells to assemble global vector
          for (int ielem = 0; ielem < numDualElems; ielem++) {

            // Loop over rows in RHS vector
            for (int cellRow = 0; cellRow < numFields; cellRow++){
                int globalRow = dualElemToNode(ielem,cellRow);
                double rhsContribution = elemRHSVector(ielem,cellRow);
                rhsVector.SumIntoGlobalValues(1, &globalRow, &rhsContribution);
             }
          } // end loop over cells

          // Assemble over multiple processors
          rhsVector.GlobalAssemble();

          // Solve
          Epetra_MultiVector x(rhsVector);
          x.PutScalar(0.0);

          Epetra_LinearProblem Problem(&MassMatrix,&x,&rhsVector);
          Epetra_MultiVector* lhs = Problem.GetLHS();

          AztecOO solver(Problem);
          solver.SetPrecOperator(MLPrec);
          solver.SetAztecOption(AZ_solver, AZ_cg);
          solver.SetAztecOption(AZ_output, 0);

          solver.Iterate(100, 1e-10);

          // Put results in array
          for (std::size_t ilat=0; ilat< num_lat; ilat++){
            for (std::size_t ilon=0; ilon< num_lon; ilon++){
               int ielem = ilat*(num_lon)+ilon;
               int ind = ilat*(num_lon*num_lev)+ilon*num_lev+ilev;
               var_vort_div_ptr[ind] = (*lhs)[0][ielem];
             }
          }

       } // end loop over levels
    } // end loop over time

    delete MLPrec;

  } //end compute vorticity/divergence at cell center 

  void 
  pcalgo::vort_div_nodes_spectral(pcvar& var_u, pcvar& var_v, pcvar& var_vort_div, 
			       bool is_algo_vort)
  {    
    // Check for missing values
    if (var_u.has_missing_val() || var_v.has_missing_val()) {
      throw pargal_except("Error: voriticity/divergence algorithm does not accept missing values", 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    }

    // MOAB currently reads in CAM-SE mesh as bilinear quads, need to change to spectral quad later
    init_bilinear_quad();

    // Get cell and basis dimensions 
    int numNodesPerElem = cellType.getNodeCount();
    int spaceDim = cellType.getDimension();

   // Get moab
    moab::Interface& mb = var_u.get_finfo().get_moab();
    moab::EntityHandle& file_set = var_u.get_finfo().get_file_set();
    const moab::ParallelComm& mb_pcomm = var_u.get_finfo().get_pcomm();

    Epetra_MpiComm Comm(const_cast<moab::ParallelComm&>(mb_pcomm).comm() );

    // Get local elements
    moab::Range elems;
    moab::ErrorCode rval = mb.get_entities_by_type(file_set, moab::MBQUAD, elems);
    if (rval != moab::MB_SUCCESS)
      throw pargal_except("Error: " + mb.get_error_string(rval), 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);

    // Get owned elements
    moab::Range elems_owned;
    rval = const_cast<moab::ParallelComm&>(mb_pcomm).filter_pstatus(elems, PSTATUS_NOT_OWNED, PSTATUS_NOT, -1, &elems_owned);
    std::size_t numElems = elems_owned.size();

   // Get local vertices (includes shared)
    moab::Range verts;
    rval = mb.get_entities_by_type(file_set, moab::MBVERTEX, verts);

   // Get owned vertices
    moab::Range verts_owned;
    rval = const_cast<moab::ParallelComm&>(mb_pcomm).filter_pstatus(verts, PSTATUS_NOT_OWNED, PSTATUS_NOT, -1, &verts_owned);

    // get num time steps
    std::size_t num_tsteps = var_u.get_finfo().get_dim("time").size();
    std::vector<int> tsteps(num_tsteps);
    for (std::size_t i = 0; i != num_tsteps; ++i)
       tsteps[i] = i;

    // get num levels
     std::size_t num_lev = var_u.num_levels();

    // Get tags for global id
     moab::Tag globalIdTag;
     rval=mb.tag_get_handle("GLOBAL_ID",1,moab::MB_TYPE_INTEGER,globalIdTag);
     rval = const_cast<moab::ParallelComm&>(mb_pcomm).assign_global_ids(file_set,2,1,false,true,false);

    // Get shared entities and exchange tags
     moab::Range temp_ents;
     rval = const_cast<moab::ParallelComm&>(mb_pcomm).get_shared_entities(-1,temp_ents,-1,false,false);
     rval = const_cast<moab::ParallelComm&>(mb_pcomm).exchange_tags(globalIdTag, temp_ents);

    // global map
     Epetra_Map globalMapNode(-1,verts_owned.size(),0,Comm);

    // Mass matrix
     Epetra_FECrsMatrix MassMatrix(Copy, globalMapNode, numNodesPerElem*10);
  
     Intrepid::FieldContainer<double> elemCoordsCart3(numElems,numNodesPerElem,3);
     Intrepid::FieldContainer<double> elemToNode(numElems,numNodesPerElem);
 
    // loop over cells and get nodal coords and cell to node coords
     int ielem = 0;
     for (moab::Range::iterator itElem = elems_owned.begin(); itElem != elems_owned.end(); ++itElem) {

      // get element handle
       moab::EntityHandle cellHandle = *itElem;
 
      // get vertex connectivity for element
       const moab::EntityHandle *nodeConn;
       int nNodes = 0;
       std::vector<moab::EntityHandle> sm_storage;
       rval = mb.get_connectivity(cellHandle, nodeConn, nNodes, true, &sm_storage);

      // loop over nodes
       double coords[3];
       for (int node=0; node < nNodes; node++) {
          rval = mb.get_coords(&nodeConn[node], 1, coords);

          int nodeId;
          rval = mb.tag_get_data(globalIdTag, &nodeConn[node],1, &nodeId);

          double radmoab = std::sqrt(coords[0]*coords[0] + coords[1]*coords[1] + coords[2]*coords[2]);
          elemCoordsCart3(ielem,node,0) = coords[0]/radmoab;
          elemCoordsCart3(ielem,node,1) = coords[1]/radmoab;
          elemCoordsCart3(ielem,node,2) = coords[2]/radmoab;

          elemToNode(ielem,node) = nodeId-1;

       } // end loop over cell nodes

       ielem++;

    } // end loop over cells
 
    Intrepid::FieldContainer<double> elemCoords(numElems,numNodesPerElem,spaceDim);
    Intrepid::FieldContainer<int> elemFaceNum(numElems);
    Intrepid::FieldContainer<double> Dmatrix(numElems,numNodesPerElem,spaceDim,spaceDim);
 
    // get face number
    getElementFaceNo(elemFaceNum, elemCoordsCart3);
 
    // get coordinates on cubed-sphere
    getCoordsCube(elemCoords, elemFaceNum, elemCoordsCart3);
 
    // get mapping matrix
    getDMatrix(Dmatrix, elemCoords, elemFaceNum);
 
   // initialize multi-dimensional arrays for mapping between reference and physical element
    Intrepid::FieldContainer<double> cellJacobian (numElems, numCubPoints, spaceDim, spaceDim);
    Intrepid::FieldContainer<double> cellJacobInv (numElems, numCubPoints, spaceDim, spaceDim);
    Intrepid::FieldContainer<double> cellJacobDet (numElems, numCubPoints);

   // calculate Jacobians
    Intrepid::CellTools<double>::setJacobian(cellJacobian, cubPoints, elemCoords, cellType);
    Intrepid::CellTools<double>::setJacobianInv(cellJacobInv, cellJacobian );
    Intrepid::CellTools<double>::setJacobianDet(cellJacobDet, cellJacobian );

   // map reference points to physical space
    Intrepid::FieldContainer<double> physPoints (numElems, numCubPoints, spaceDim);
    Intrepid::CellTools<double>::mapToPhysicalFrame (physPoints, cubPoints, elemCoords, cellType);

   // Containers for basis grads and values transformed to physical cell
    Intrepid::FieldContainer<double> physBasisGrads (numElems, numFields, numCubPoints, spaceDim);
    Intrepid::FieldContainer<double> physBasisVals (numElems, numFields, numCubPoints);

   // Transform basis gradients to physical frame:
    Intrepid::FunctionSpaceTools::HGRADtransformGRAD<double>(physBasisGrads,
                                               cellJacobInv,   basisGrads);

   // Transform basis values to physical frame:
    Intrepid::FunctionSpaceTools::HGRADtransformVALUE<double>(physBasisVals,
                                                               basisValues);

   // get metric tensor jacobian at integration points
    Intrepid::FieldContainer<double> jacobG(numElems, numCubPoints);
    getJacobianG(jacobG, physPoints);

     Intrepid::FieldContainer<double> jacobGnode;
     if (!is_algo_vort){
       // for divergence also get metric tensor jacobian at nodes
       jacobGnode.resize(numElems, numNodesPerElem);
       getJacobianG(jacobGnode, elemCoords);
     }

   // Compute cell measure
    Intrepid::FieldContainer<double> cellMeasure0 (numElems, numCubPoints);
    Intrepid::FieldContainer<double> cellMeasure (numElems, numCubPoints);
    Intrepid::FunctionSpaceTools::computeCellMeasure<double>(cellMeasure0,
                                                cellJacobDet, cubWeights);
    Intrepid::FunctionSpaceTools::computeCellMeasure<double>(cellMeasure,
                                                cellMeasure0, jacobG);

   // Multiply transformed values with weighted measure
    Intrepid::FieldContainer<double> physBasisValsWeighted (numElems, numFields, numCubPoints);
    Intrepid::FunctionSpaceTools::multiplyMeasure<double>(physBasisValsWeighted,
                                             cellMeasure, physBasisVals);

   // Multiply transformed grads with weighted measure
    Intrepid::FieldContainer<double> physBasisGradsWeighted (numElems, numFields, numCubPoints, spaceDim);
    Intrepid::FunctionSpaceTools::multiplyMeasure<double>(physBasisGradsWeighted,
                                             cellMeasure, physBasisGrads);

   // Integrate to compute element mass matrix
    Intrepid::FieldContainer<double> elemMassMatrix (numElems, numFields, numFields);

    Intrepid::FunctionSpaceTools::integrate<double>(elemMassMatrix,
                                           physBasisValsWeighted,
                                           physBasisVals, Intrepid::COMP_BLAS);
 

   // Loop over cells to assemble global matrix/vector
    for (std::size_t ielem = 0; ielem < numElems; ielem++) {

       std::vector<int> nodeIndex(numFields);
       for (int cellCol = 0; cellCol < numFields; cellCol++){
          nodeIndex[cellCol] = elemToNode(ielem,cellCol);
       }
       for (int cellRow = 0; cellRow < numFields; cellRow++){
          MassMatrix.InsertGlobalValues(1, &nodeIndex[cellRow], numFields, &nodeIndex[0], &elemMassMatrix(ielem,cellRow,0));
       }
    }  // end loop over cells

  // Assemble over multiple processors
    MassMatrix.GlobalAssemble();
    MassMatrix.FillComplete();

  // Get lumped mass matrix 
  // (Spectral elems have purely diagonal mass matrix 
  //   - so using lumped mass matrix with bilinear elements
  //     is a similar procedure)
   Epetra_MultiVector lumpedMass(globalMapNode,1);
   Epetra_MultiVector ones(globalMapNode,1);
   ones.PutScalar(1.0);

   MassMatrix.Multiply(false,ones,lumpedMass);

  // Loop over time
    for (std::size_t itime = 0; itime < num_tsteps; itime++) {

     // Get tags for velocity data and exchange data
      moab::Tag uvelTag = var_u.get_moab_tag(itime);
      moab::Tag vvelTag = var_v.get_moab_tag(itime);

      rval = const_cast<moab::ParallelComm&>(mb_pcomm).exchange_tags(uvelTag, temp_ents);
      rval = const_cast<moab::ParallelComm&>(mb_pcomm).exchange_tags(vvelTag, temp_ents);

     // Get pointer to vort_div data
      double* var_vort_div_ptr = var_vort_div.get_storage(itime);

      Epetra_FEVector rhsVector(globalMapNode,num_lev);
      rhsVector.PutScalar(0.0);

        // loop over cells and get velocity data
         Intrepid::FieldContainer<double> elemVelSphere(numElems,numNodesPerElem,num_lev,spaceDim);
         int ielem = 0;
         for (moab::Range::iterator itElem = elems_owned.begin(); itElem != elems_owned.end(); ++itElem) {

             // get element handle
             moab::EntityHandle cellHandle = *itElem;

             // get vertex connectivity for element
             const moab::EntityHandle *nodeConn;
             int nNodes = 0;
             std::vector<moab::EntityHandle> sm_storage;
             rval = mb.get_connectivity(cellHandle, nodeConn, nNodes, true, &sm_storage);
 
             // loop over nodes
             for (int node=0; node < nNodes; node++) {

                std::vector<double> uvel(num_lev);
                std::vector<double> vvel(num_lev);
                rval = mb.tag_get_data(uvelTag, &nodeConn[node], 1, &uvel[0]);
                rval = mb.tag_get_data(vvelTag, &nodeConn[node], 1, &vvel[0]);
                for (std::size_t ilev = 0; ilev < num_lev; ilev++) { 
                   elemVelSphere(ielem,node,ilev,0) = uvel[ilev];
                   elemVelSphere(ielem,node,ilev,1) = vvel[ilev];
                }

              } // end loop over cell nodes

            ielem++;

       } // end loop over cells

    // Loop over levels
     for (std::size_t ilev = 0; ilev < num_lev; ilev++) {

        // Transform velocity
        Intrepid::FieldContainer<double> elemVel(numElems, numNodesPerElem, spaceDim);

        if (is_algo_vort){
           // covariant transform of velocity
             for (std::size_t icell=0; icell < numElems; icell++){
                 for (int ipt=0; ipt < numFields; ipt++){
                 elemVel(icell,ipt,0) = Dmatrix(icell,ipt,0,0)*elemVelSphere(icell,ipt,ilev,0) + Dmatrix(icell,ipt,1,0)*elemVelSphere(icell,ipt,ilev,1);
                 elemVel(icell,ipt,1) = Dmatrix(icell,ipt,0,1)*elemVelSphere(icell,ipt,ilev,0) + Dmatrix(icell,ipt,1,1)*elemVelSphere(icell,ipt,ilev,1);
                 }
             }
         }
         else {
           // contravariant transform of velocity
             for (std::size_t icell=0; icell < numElems; icell++){
                 for (int ipt=0; ipt < numFields; ipt++){
                 double Ddet = Dmatrix(icell,ipt,0,0)*Dmatrix(icell,ipt,1,1)-Dmatrix(icell,ipt,0,1)*Dmatrix(icell,ipt,1,0);
                 elemVel(icell,ipt,0) = Dmatrix(icell,ipt,1,1)*elemVelSphere(icell,ipt,ilev,0)/Ddet - Dmatrix(icell,ipt,0,1)*elemVelSphere(icell,ipt,ilev,1)/Ddet;
                 elemVel(icell,ipt,1) = -Dmatrix(icell,ipt,1,0)*elemVelSphere(icell,ipt,ilev,0)/Ddet + Dmatrix(icell,ipt,0,0)*elemVelSphere(icell,ipt,ilev,1)/Ddet;
                 }
             }
         }

       // Loop over cells to compute vorticity or divergence for RHS
         Intrepid::FieldContainer<double> vort_div(numElems, numCubPoints);
         Intrepid::FieldContainer<double> elemRHSVector (numElems, numFields);
         for (std::size_t ielem=0; ielem < numElems; ielem++) {
            for (int ipt=0; ipt < numCubPoints; ipt++ ) {

             if (is_algo_vort){

               // evaluate vorticity: 1/sqrt(G) (dv/dx - du/dy)
                double uVelGrady = 0.0;
                double vVelGradx = 0.0;
                for (int inode=0; inode < numNodesPerElem; inode++) {
                    uVelGrady += elemVel(ielem,inode,0)*physBasisGrads(ielem,inode,ipt,1);
                    vVelGradx += elemVel(ielem,inode,1)*physBasisGrads(ielem,inode,ipt,0);
                }
                vort_div(ielem,ipt) = (vVelGradx - uVelGrady)/(jacobG(ielem,ipt));
              }
              else {

               // evaluate divergence: 1/sqrt(G) (d(sqrt(G) u)/dx - d(sqrt(G) v)/dy)
                double uVelGradx = 0.0;
                double vVelGrady = 0.0;
                for (int inode=0; inode < numNodesPerElem; inode++) {
                    uVelGradx += jacobGnode(ielem,inode)*elemVel(ielem,inode,0)*physBasisGrads(ielem,inode,ipt,0);
                    vVelGrady += jacobGnode(ielem,inode)*elemVel(ielem,inode,1)*physBasisGrads(ielem,inode,ipt,1);
                }
                vort_div(ielem,ipt) = (uVelGradx + vVelGrady)/(jacobG(ielem,ipt));
              }

          } // end point loop
         } // end cell loop


          Intrepid::FunctionSpaceTools::integrate<double>(elemRHSVector, vort_div,
                                           physBasisValsWeighted, Intrepid::COMP_BLAS);

          // Loop over cells to assemble global matrix/vector
           for (std::size_t ielem = 0; ielem < numElems; ielem++) {

            // Loop over rows in stiffness matrix
             for (int cellRow = 0; cellRow < numNodesPerElem; cellRow++){

                 int globalRow = elemToNode(ielem,cellRow);
                 double rhsContribution = elemRHSVector(ielem,cellRow);
                 rhsVector.SumIntoGlobalValues(1, &globalRow, &rhsContribution, ilev);

             }// end loop over rows

          } // end loop over cells

       } // end loop over levels

       rhsVector.GlobalAssemble();

      // Solve
       Epetra_MultiVector lhsVector(globalMapNode,num_lev);
       lhsVector.ReciprocalMultiply(1.0,lumpedMass,rhsVector,0.0);

       double iradius = 1.0/radius;
       lhsVector.Scale(iradius);

      // put vorticity or divergence values into vort_div tag
       int nnode = 0;
       for (moab::Range::iterator itNode = verts_owned.begin(); itNode != verts_owned.end(); ++itNode) {
       
          for (std::size_t ilev=0; ilev<num_lev; ilev++){

             // index to storage point
             std::size_t ind = nnode*num_lev + ilev;
             var_vort_div_ptr[ind] = lhsVector[ilev][nnode];

           } // end loop over levels

           nnode++;

        } // end loop over nodes

    } // end loop over times

  } //end compute vorticity/divergence at nodes - spectral element

template<class ArrayIn, class ArrayOut>
void pcalgo::getElementFaceNo(ArrayOut &elemFaceNum, ArrayIn &elemCoordsSphere){

  // get dimension
    std::size_t numCells        = elemCoordsSphere.dimension(0);
    std::size_t numNodesPerCell = elemCoordsSphere.dimension(1);

  // loop over cells to get element face number and local face coordinates
    for (std::size_t icell=0; icell < numCells; icell++){

      // get center (average) coordinates
      double centerX = 0.0;
      double centerY = 0.0;
      double centerZ = 0.0;
      for (std::size_t inode=0; inode < numNodesPerCell; inode++){
        centerX += elemCoordsSphere(icell,inode,0);
        centerY += elemCoordsSphere(icell,inode,1);
        centerZ += elemCoordsSphere(icell,inode,2);
      }
      double X = centerX/numNodesPerCell;
      double Y = centerY/numNodesPerCell;
      double Z = centerZ/numNodesPerCell;

     // find element face number
     //
     //   5 = North Pole
     //   6 = South Pole
     //           _________
     //           |       |
     //           |   5   |
     //   ________|_______|_______________
     //   |       |       |       |       |
     //   |   4   |   1   |   2   |   3   |
     //   |_______|_______|_______|_______|
     //           |       |
     //           |   6   |
     //           |_______|
     //

      if ((Y<=X) & (Y>=-X)) {
         if (Z>X) {
           elemFaceNum(icell) = 5;
         }
         else if (Z<-X) {
           elemFaceNum(icell) = 6;
         }
         else {
           elemFaceNum(icell) = 1;
         }
      }
      else if ((Y>=X) & (Y<=-X)) {
         if (Z>-X) {
           elemFaceNum(icell) = 5;
         }
         else if (Z<X) {
           elemFaceNum(icell) = 6;
         }
         else {
           elemFaceNum(icell) = 3;
         }
      }
      else if ((Y>=X) & (Y>=-X)) {
         if (Z>Y) {
           elemFaceNum(icell) = 5;
         }
         else if (Z<-Y) {
           elemFaceNum(icell) = 6;
         }
         else {
           elemFaceNum(icell) = 2;
         }
      }
      else if ((Y<=X) & (Y<=-X)) {
         if (Z>-Y) {
           elemFaceNum(icell) = 5;
         }
         else if (Z<Y) {
           elemFaceNum(icell) = 6;
         }
         else {
           elemFaceNum(icell) = 4;
         }
      }
   else {
         if (Z>X) {
           elemFaceNum(icell) = 5;
         }
         else if (Z<Y) {
           elemFaceNum(icell) = 6;
         }
      }

    }  // end loop over cells

}

template<class ArrayIn1, class ArrayIn2, class ArrayOut>
void pcalgo::getCoordsCube(ArrayOut &elemCoordsCube, ArrayIn1 &elemFaceNum, ArrayIn2 &elemCoordsSphere) {

  // get dimension
    std::size_t numCells        = elemCoordsSphere.dimension(0);
    std::size_t numNodesPerCell = elemCoordsSphere.dimension(1);

  // loop over cells to get element face number and local face coordinates
    for (std::size_t icell=0; icell < numCells; icell++){

       switch (elemFaceNum(icell)) {

           case 1:
              for (std::size_t inode=0; inode < numNodesPerCell; inode++) {
                   elemCoordsCube(icell,inode,0) = atan(elemCoordsSphere(icell,inode,1)/elemCoordsSphere(icell,inode,0));
                   elemCoordsCube(icell,inode,1) = atan(elemCoordsSphere(icell,inode,2)/elemCoordsSphere(icell,inode,0));
              }
              break;
           case 2:
              for (std::size_t inode=0; inode < numNodesPerCell; inode++) {
                   elemCoordsCube(icell,inode,0) = atan(-elemCoordsSphere(icell,inode,0)/elemCoordsSphere(icell,inode,1));
                   elemCoordsCube(icell,inode,1) = atan(elemCoordsSphere(icell,inode,2)/elemCoordsSphere(icell,inode,1));
              }
              break;
           case 3:
           for (std::size_t inode=0; inode < numNodesPerCell; inode++) {
                   elemCoordsCube(icell,inode,0) = atan(elemCoordsSphere(icell,inode,1)/elemCoordsSphere(icell,inode,0));
                   elemCoordsCube(icell,inode,1) = atan(-elemCoordsSphere(icell,inode,2)/elemCoordsSphere(icell,inode,0));
              }
              break;
           case 4:
              for (std::size_t inode=0; inode < numNodesPerCell; inode++) {
                   elemCoordsCube(icell,inode,0) = atan(-elemCoordsSphere(icell,inode,0)/elemCoordsSphere(icell,inode,1));
                   elemCoordsCube(icell,inode,1) = atan(-elemCoordsSphere(icell,inode,2)/elemCoordsSphere(icell,inode,1));
              }
              break;
           case 5:
              for (std::size_t inode=0; inode < numNodesPerCell; inode++) {
                   elemCoordsCube(icell,inode,0) = atan(elemCoordsSphere(icell,inode,1)/elemCoordsSphere(icell,inode,2));
                   elemCoordsCube(icell,inode,1) = atan(-elemCoordsSphere(icell,inode,0)/elemCoordsSphere(icell,inode,2));
              }
              break;
           case 6:
              for (std::size_t inode=0; inode < numNodesPerCell; inode++) {
                   elemCoordsCube(icell,inode,0) = atan(-elemCoordsSphere(icell,inode,1)/elemCoordsSphere(icell,inode,2));
                   elemCoordsCube(icell,inode,1) = atan(-elemCoordsSphere(icell,inode,0)/elemCoordsSphere(icell,inode,2));
              }
              break;

       } // end switch

    } // end loop over cells

}

template<class ArrayIn, class ArrayOut>
void pcalgo::getJacobianG(ArrayOut &jacobG, ArrayIn &pointsInElem){

  // get dimension
    std::size_t numCells  = pointsInElem.dimension(0);
    std::size_t numPoints = pointsInElem.dimension(1);

  // loop over cells to get Jacobian of metric tensor
    for (std::size_t icell=0; icell < numCells; icell++){

        for (std::size_t ipt=0; ipt < numPoints; ipt++){

          // equiangular projection
           double tanx1 = tan(pointsInElem(icell,ipt,0));
           double tanx2 = tan(pointsInElem(icell,ipt,1));
           double rho2 = 1.0 + tanx1*tanx1 + tanx2*tanx2;
           double rho = sqrt(rho2);
           double cosx1 = cos(pointsInElem(icell,ipt,0));
           double cosx2 = cos(pointsInElem(icell,ipt,1));

           jacobG(icell,ipt) = 1.0/(rho2*rho*cosx1*cosx1*cosx2*cosx2);

        }

    } // end cell loop
}

template<class ArrayIn1, class ArrayIn2, class ArrayOut>
void pcalgo::getDMatrix(ArrayOut &D, ArrayIn1 &pointsInElem, ArrayIn2 &elemFaceNum){

  // get dimension
    std::size_t numCells  = pointsInElem.dimension(0);
    std::size_t numPoints = pointsInElem.dimension(1);

  // loop over cells to get metric tensor
    for (std::size_t icell=0; icell < numCells; icell++){

       if (elemFaceNum(icell) == 6) {
           for (std::size_t ipt=0; ipt < numPoints; ipt++){
              double tanx1 = tan(pointsInElem(icell,ipt,0));
              double tanx2 = tan(pointsInElem(icell,ipt,1));
              double poledist = std::sqrt(tanx1*tanx1 + tanx2*tanx2);
              if (poledist < 1e-9) {
                  D(icell,ipt,0,0) = 1.0;
                  D(icell,ipt,1,1) = 1.0;
                  D(icell,ipt,1,0) = 0.0;
                  D(icell,ipt,0,1) = 0.0;
              }
              else{
                  double cosx1 = cos(pointsInElem(icell,ipt,0));
                  double cosx2 = cos(pointsInElem(icell,ipt,1));
                  double rho2 = 1 + tanx1*tanx1 + tanx2*tanx2;
                  double rho = sqrt(rho2);
                  D(icell,ipt,0,0) = tanx2/(poledist*cosx1*cosx1*rho);
                  D(icell,ipt,1,1) = tanx2/(poledist*cosx2*cosx2*rho2);
                  D(icell,ipt,0,1) = -tanx1/(poledist*cosx2*cosx2*rho);
                  D(icell,ipt,1,0) = tanx1/(poledist*cosx1*cosx1*rho2);
              }
           }
        } // end if face 5

      else if (elemFaceNum(icell) == 5) {

           for (std::size_t ipt=0; ipt < numPoints; ipt++){

              double tanx1 = tan(pointsInElem(icell,ipt,0));
              double tanx2 = tan(pointsInElem(icell,ipt,1));
              double poledist = std::sqrt(tanx1*tanx1 + tanx2*tanx2);
              if (poledist < 1e-9) {
                  D(icell,ipt,0,0) = 1.0;
                  D(icell,ipt,1,1) = 1.0;
                  D(icell,ipt,1,0) = 0.0;
                  D(icell,ipt,0,1) = 0.0;
              }
              else{
                  double cosx1 = cos(pointsInElem(icell,ipt,0));
                  double cosx2 = cos(pointsInElem(icell,ipt,1));
                  double rho2 = 1 + tanx1*tanx1 + tanx2*tanx2;
                  double rho = sqrt(rho2);
                  D(icell,ipt,0,0) = -tanx2/(poledist*cosx1*cosx1*rho);
                  D(icell,ipt,1,1) = -tanx2/(poledist*cosx2*cosx2*rho2);
                  D(icell,ipt,0,1) = tanx1/(poledist*cosx2*cosx2*rho);
                  D(icell,ipt,1,0) = -tanx1/(poledist*cosx1*cosx1*rho2);
             }
           }
        } // end if face 6

       else  { // faces 1-4
           for (std::size_t ipt=0; ipt < numPoints; ipt++){
               double tanx1 = tan(pointsInElem(icell,ipt,0));
               double tanx2 = tan(pointsInElem(icell,ipt,1));
               double cosx1 = cos(pointsInElem(icell,ipt,0));
               double cosx2 = cos(pointsInElem(icell,ipt,1));
               double rho2 = 1 + tanx1*tanx1 + tanx2*tanx2;
               double rho = sqrt(rho2);
               D(icell,ipt,0,0) = 1.0/(rho*cosx1);
               D(icell,ipt,1,1) = 1.0/(rho2*cosx1*cosx2*cosx2);
               D(icell,ipt,0,1) = 0.0;
               D(icell,ipt,1,0) = -tanx1*tanx2/(cosx1*rho2);
           }
        } // end if faces 1-4 

    } // end cell loop

}

} // namespace pargal
