/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <pargal/fileinfo.hpp>
#include <pargal/exception.hpp>
#include <iostream>

namespace pargal 
{
  //! constructor
  fileinfo::fileinfo(const std::vector<std::string>& file_names,
		     moab::Interface& mb,
		     const moab::ParallelComm& mb_pcomm,
		     moab::EntityHandle& file_set,
		     const std::string& part_method) :
    m_file_names(file_names),
    m_mb(mb),
    m_mb_pcomm(mb_pcomm),
    m_file_set(file_set),
    m_part_method(part_method)
  {
    std::string opts = "PARALLEL=READ_PART;PARTITION;PARALLEL_GHOSTS=2.0.1;NOMESH;VARIABLE=;PARTITION_METHOD=";
    opts.append(m_part_method);
    moab::ErrorCode rval = m_mb.load_file(m_file_names[0].c_str(), 
					  &m_file_set, 
					  opts.c_str());
    if (rval != moab::MB_SUCCESS)
      throw pargal_except("Error: " + m_mb.get_error_string(rval), 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
	MPI_Comm_dup(mb_pcomm.comm(), &mpi_comm_);
    init();
  }
  
  //! destructor
  fileinfo::~fileinfo()
  {
    std::map<std::string, const fvar*>::iterator it_var = m_vars.begin();
    for (; it_var != m_vars.end(); ++it_var)
      delete it_var->second;
    
    std::map<std::string, const pcdim*>::iterator it_dim = m_dims.begin();
    for (; it_dim != m_dims.end(); ++it_dim)
      delete it_dim->second;
    
    std::map<std::string, const pcatt*>::iterator it_att = m_atts.begin();
    for (; it_att != m_atts.end(); ++it_att)
      delete it_att->second;   
  }
  
  ///////////////////////////////////////////
  // general queries
  ///////////////////////////////////////////
  
  //! return the number of dimensions in the netcdf file.
  std::size_t
  fileinfo::num_dims() const
  {
    return m_dims.size();
  }
  
  //! return the number of variables in the netcdf file
  std::size_t
  fileinfo::num_vars() const
  {
    return m_vars.size();
  }
  
  //! return the number of global attributes in the netcdf file    
  std::size_t 
  fileinfo::num_atts() const
  {
    return m_atts.find("MOAB_GLOBAL")->second->size();
  }
  
  //! return the dimension names in the netcdf file
  std::set<std::string>
  fileinfo::dim_names() const
  {
    std::set<std::string> s;
    std::map<std::string, const pcdim*>::const_iterator it_dim = m_dims.begin();
    for (;it_dim != m_dims.end(); ++it_dim)
      s.insert(it_dim->first);
    return s;
  }
  
  //! return the variable names in the netcdf file
  std::set<std::string> 
  fileinfo::var_names() const
  {
    std::set<std::string> s;
    std::map<std::string, const fvar*>::const_iterator it_var = m_vars.begin();
    for (std::size_t i = 0; it_var != m_vars.end(); ++it_var, ++i)
      s.insert(it_var->first);
    return s;
  }
  
  //! return the attribute names in the netcdf file
  std::set<std::string> 
  fileinfo::att_names() const
  {
    return const_cast<fileinfo*>(this)->m_atts.at("MOAB_GLOBAL")->get_att_names();    
  }

	//! Returns true if a dimension by name dim_name exists
	//! false otherwise
	bool
	fileinfo::has_dim(const std::string& dim_name) const
	{
		return ((m_dims.find(dim_name) != m_dims.end()) ? true : false);
	}
  
  //! return a dimension by name
  const pcdim& 
  fileinfo::get_dim(const std::string& dim_name) const
  {
    return *(const_cast<fileinfo*>(this)->m_dims.at(dim_name));
  }
  
  //! return the global attribute
  const pcatt& 
  fileinfo::get_att() const
  {
    return *(const_cast<fileinfo*>(this)->m_atts.at("MOAB_GLOBAL"));
  }
  
  ///////////////////////////////////////////
  // modifiers
  ///////////////////////////////////////////
  
  void 
  fileinfo::insert(const std::string& var_name, const fvar& var)
  {
    m_vars[var_name] = &var;
  }
  
  void 
  fileinfo::insert(const std::string& dim_name, const pcdim& dim)
  {
    m_dims[dim_name] = &dim;
  }
  
  void 
  fileinfo::insert(const std::string& att_name, const pcatt& att)
  {
    m_atts[att_name] = &att;
  }
  
  ///////////////////////////////////////////
  // initialization
  ///////////////////////////////////////////
  
  void 
  fileinfo::init()
  {
    init_grid_type();
    init_dims();
    init_vars();
    init_atts();
  }
  
  void 
  fileinfo::init_dims()
  {
    moab::Tag tag = 0;
    std::string tag_name = "__DIM_NAMES";
    const void * data = NULL;
    int num_dims = 0;
    moab::ErrorCode rval;
    rval = m_mb.tag_get_handle(tag_name.c_str(), 0, moab::MB_TYPE_OPAQUE, tag, moab::MB_TAG_ANY);
    rval = m_mb.tag_get_by_ptr(tag, &m_file_set, 1, &data, &num_dims);
    if (rval != moab::MB_SUCCESS)
      throw pargal_except("Error: " + m_mb.get_error_string(rval), 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    std::size_t start = 0;
    const char * p = static_cast<const char *>(data);
    moab::Tag dimValsTag = 0;
    tag_name = "__DIM_LENS";
    const void * valdata = NULL;
    int num_vals = 0;
    rval = m_mb.tag_get_handle(tag_name.c_str(), 0, moab::MB_TYPE_INTEGER, dimValsTag, moab::MB_TAG_ANY);
    rval = m_mb.tag_get_by_ptr(dimValsTag, &m_file_set, 1, &valdata, &num_vals);
    const int * intp = static_cast<const int*>(valdata);
    int idx = -1;

    for (std::size_t i = 0; i != static_cast<std::size_t>(num_dims); ++i)
      {
	if (p[i] == '\0') 
	  {
	    std::string dim_name(&p[start], i-start);
	    ++idx;
	    int sz = intp[idx];
	    // fixme: need info from moab to set unlimited dimension
	    // currently assume each file has the same number of time dimensions
	    if ((dim_name == "time") || (dim_name == "Time"))
	      insert(dim_name, *(new pcdim(dim_name, sz * m_file_names.size(), true)));
	    else
	      insert(dim_name, *(new pcdim(dim_name, sz)));
	    start=i+1;
	  }
      }
  }
  
  void 
  fileinfo::init_vars()
  {
    // read <__VAR_NAMES_LOCATIONS> tag
    moab::Tag varLocTag = 0;
    std::string tag_name = "__VAR_NAMES_LOCATIONS";
    const void * loc_data = NULL;
    int loc_sz = 0;
    moab::ErrorCode rval = m_mb.tag_get_handle(tag_name.c_str(), 0, moab::MB_TYPE_INTEGER, varLocTag, moab::MB_TAG_ANY);
    rval = m_mb.tag_get_by_ptr(varLocTag, &m_file_set, 1, &loc_data, &loc_sz);
    const int * loc_p = static_cast<const int *>(loc_data);
    std::vector<int> varLoc(loc_sz);
    std::copy(loc_p, loc_p+loc_sz, varLoc.begin());
	// FIXME: What is the advantage of using a different naming scheme for loc in ParGAL?
	// Why map the loc to a string instead of an enum?
    std::map<int, std::string> locmap;
    locmap[0] = "VERTEX";
    locmap[1] = "NSEDGE";
    locmap[2] = "EWEDGE";
	if(m_grid_type == "MPAS"){
		locmap[3] = "POLYGON";
	}
	else{
		locmap[3] = "QUAD";
	}
    locmap[4] = "SET";
    locmap[5] = "EDGE";
    locmap[6] = "REGION";
    int nthVar = 0;

    moab::Tag tag = 0;
    tag_name = "__VAR_NAMES";
    rval = m_mb.tag_get_handle(tag_name.c_str(), 0, moab::MB_TYPE_OPAQUE, tag, moab::MB_TAG_ANY);
    const void * data = NULL;
    int num_vars = 0;
    rval = m_mb.tag_get_by_ptr(tag, &m_file_set, 1, &data, &num_vars);
    const char * p = static_cast<const char *>(data);
    std::size_t start = 0;
    for (std::size_t i = 0; i != static_cast<std::size_t>(num_vars); ++i) 
      {
	if (p[i] == '\0') 
	  {
	    std::string var_name(&p[start], i-start);
	    int sz = 0;
	    moab::Tag dims_tag = 0;
	    std::string dim_names = "__"+var_name+"_DIMS";
	    rval = m_mb.tag_get_handle(dim_names.c_str(), 0, moab::MB_TYPE_OPAQUE, dims_tag, moab::MB_TAG_ANY);
	    //fixme: doesn't handle variables have 0 dimension 
	    if (rval != moab::MB_SUCCESS) {
	      start=i+1;
	      ++nthVar;
	      continue;
	    }
	    rval = m_mb.tag_get_length(dims_tag, sz);
	    std::vector<const pcdim*> dims(sz, NULL);
	    const void* ptr = NULL;
	    rval = m_mb.tag_get_by_ptr(dims_tag, &m_file_set, 1, &ptr);
		if(rval != moab::MB_SUCCESS){
			throw pargal_except("Error: Unable to get dim tags for variable, " +
					m_mb.get_error_string(rval),
					__FILE__, __LINE__, __PRETTY_FUNCTION__);
		}
	    const moab::Tag * ptags = static_cast<const moab::Tag*>(ptr);
	    for (std::size_t j = 0; j != static_cast<std::size_t>(sz); ++j) 
	      {
		std::string dim_name;
		rval = m_mb.tag_get_name(ptags[j], dim_name);
		dims[j] = &(get_dim(dim_name));
	      }
	    insert(var_name, *(new fvar(var_name, moab::MB_TYPE_DOUBLE, dims, locmap[varLoc[nthVar++]])));
	    start=i+1;
	  }
      }
  }
  
  void 
  fileinfo::init_atts()
  {
    std::vector<std::string> nameVec(m_vars.size()+1);
    nameVec[0] = "GLOBAL";
    std::map<std::string, const fvar*>::iterator it = m_vars.begin();
    for (std::size_t i=1; it != m_vars.end(); ++it, ++i) 
      nameVec[i] = it->first;
         
    for (std::size_t vec_counter = 0; vec_counter != nameVec.size(); ++vec_counter) 
      {
	// read __<var_name>_ATTRIBS tag
	moab::Tag tag = 0;
	std::string tag_name = "__"+nameVec[vec_counter]+"_ATTRIBS";
	const void * data = NULL;
	int sz = 0;
	moab::ErrorCode rval = m_mb.tag_get_handle(tag_name.c_str(), 0, moab::MB_TYPE_OPAQUE, tag, moab::MB_TAG_ANY);
	if (rval != moab::MB_SUCCESS)
	  throw pargal_except("Error: " + m_mb.get_error_string(rval), 
			      __FILE__, __LINE__, __PRETTY_FUNCTION__);
	
	rval = m_mb.tag_get_by_ptr(tag, &m_file_set, 1, &data, &sz);
	const char * p = static_cast<const char *>(data);
	std::string att_val(&p[0], sz);
	if (vec_counter == 0) nameVec[0]="MOAB_GLOBAL";
	const std::string& var_name = nameVec[vec_counter];

	// read __<var_name>_ATTRIBS_LEN tag
	moab::Tag attLenTag = 0;
	tag_name = tag_name + "_LEN";
	const void * len_data = NULL;
	int len_sz = 0;
	rval = m_mb.tag_get_handle(tag_name.c_str(), 0, moab::MB_TYPE_INTEGER, attLenTag, moab::MB_TAG_ANY);
	rval = m_mb.tag_get_by_ptr(attLenTag, &m_file_set, 1, &len_data, &len_sz);
	const int * len_p = static_cast<const int *>(len_data);
	std::vector<int> attLen(len_sz);
	std::copy(len_p, len_p+len_sz, attLen.begin());
	
	// create attribute
	insert(var_name, *(new pcatt(var_name, att_val, attLen)));
      }
  }

  void
  fileinfo::init_grid_type()
  {
    moab::Tag tag = 0;
    std::string tag_name = "__MESH_TYPE";
    const void * data = NULL;
    int sz = 0;
    moab::ErrorCode rval;
    rval = m_mb.tag_get_handle(tag_name.c_str(), 0, moab::MB_TYPE_OPAQUE, tag, moab::MB_TAG_ANY);
    rval = m_mb.tag_get_by_ptr(tag, &m_file_set, 1, &data, &sz);
    if (rval != moab::MB_SUCCESS)
      throw pargal_except("Error: " + m_mb.get_error_string(rval), 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    const char * p = static_cast<const char *>(data);
    m_grid_type = std::string(&p[0], sz);
  }
   
  //! return file names
  const std::vector<std::string>& 
  fileinfo::file_names() const
  {
    return m_file_names;
  }

  //! look up files containing specific tstep
  const std::string& 
  fileinfo::tstep_lookup(int tstep) const
  {
    if (m_file_names.size() == 1)
      return m_file_names[0];
    else {
      size_t tsteps_per_file;
		// FIXME: This code needs revision
		// 1. get_dim() assumes the passed in dim name is valid
		if(m_grid_type == "MPAS"){
			tsteps_per_file = get_dim("Time").size();
		}
		else{
			tsteps_per_file = get_dim("time").size();
		}
      size_t j = 0;
      if (tsteps_per_file != 1)
	j = tstep % tsteps_per_file;
      else
	j = tstep;
      return m_file_names[j];
    }
  }
  
  ///////////////////////////////////////////
  // variable related queries
  ///////////////////////////////////////////
  
  //! return the type of the variable
  moab::DataType 
  fileinfo::get_var_type(const std::string& name) const
  {
    return const_cast<fileinfo*>(this)->m_vars.at(name)->type();
  }
  
  //! return the number of dimensions of the variable
  std::size_t 
  fileinfo::get_var_num_dims(const std::string& name) const
  {
    return const_cast<fileinfo*>(this)->m_vars.at(name)->num_dims();
  }
  
  //! return a reference to the index-th dimension (starting at 0)
  const pcdim& 
  fileinfo::get_var_dim(const std::string& name, 
			const std::size_t index) const
  {
    return const_cast<fileinfo*>(this)->m_vars.at(name)->get_dim(index);
  }
  
  //! return a vector of dimensions
  std::vector<pcdim> 
  fileinfo::get_var_dims(const std::string& name) const
  {
    std::vector<pcdim> vec;
    std::vector<const pcdim*> vecp = const_cast<fileinfo*>(this)->m_vars.at(name)->get_dims();
    for (std::size_t i = 0; i != vecp.size(); ++i)
      vec.push_back(*vecp[i]);
    return vec;
  }
  
  //! return the number of attributes of the variable
  std::size_t 
  fileinfo::get_var_num_atts(const std::string& name) const
  {
    return const_cast<fileinfo*>(this)->m_vars.at(name)->num_atts();
  }
    
  //! return the attribute by name
  const pcatt& 
  fileinfo::get_var_att(const std::string& varname) const  
  {
    return *(const_cast<fileinfo*>(this)->m_atts.at(varname));
  }

  //! return the location information of a variable
  const std::string&
  fileinfo::get_var_location(const std::string& varname) const
  {
    return const_cast<fileinfo*>(this)->m_vars.at(varname)->location();
  }
  
  ///////////////////////////////////////////
  // moab queries
  ///////////////////////////////////////////
  
  //! return the moab instance
  moab::Interface& 
  fileinfo::get_moab()
  {
    return m_mb;
  }

  //! return the moab ParallelComm instance
  const moab::ParallelComm&
  fileinfo::get_pcomm() const
  {
    return m_mb_pcomm;
  }

  //! return the file_set
  moab::EntityHandle& 
  fileinfo::get_file_set()
  {
    return m_file_set;
  }
  
  //! return the partition method
  const std::string& 
  fileinfo::get_part_method() const
  {
    return m_part_method;
  }  

	//! return MPI communicator for ops associated with the files
	//! represented by this file info
	MPI_Comm fileinfo::get_mpi_comm(void ) const
	{
		return mpi_comm_;
	}

  const std::string& 
  fileinfo::get_grid_type() const
  {
    return m_grid_type;
  }

} // namespace pargal
