/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <pargal/mesh.hpp>

namespace pargal 
{
  ///////////////////////////////////////////
  // mesh info
  ///////////////////////////////////////////
  
  //! return the total number of vertices at rank 0
  std::size_t 
  get_num_vertices(const moab::Interface& mb, moab::EntityHandle& file_set)
  {
    return get_num_entities<moab::MBVERTEX>(mb, file_set);
  }

  //! return the total number of quads at rank 0
  std::size_t 
  get_num_quads(const moab::Interface& mb, moab::EntityHandle& file_set)
  {
    return get_num_entities<moab::MBQUAD>(mb, file_set);
  }
  
  //! return the total number of polygons at rank 0
  std::size_t 
  get_num_polygons(const moab::Interface& mb, moab::EntityHandle& file_set)
  {
    return get_num_entities<moab::MBPOLYGON>(mb, file_set);
  }
  
  //! return the total number of edges at rank 0
  std::size_t 
  get_num_edges(const moab::Interface& mb, moab::EntityHandle& file_set)
  {
    return get_num_entities<moab::MBEDGE>(mb, file_set);
  }
  
  //! return the number of local vertices    
  std::size_t 
  get_local_num_vertices(const moab::Interface& mb,
			 moab::EntityHandle& file_set)
  {
    return get_local_num_entities_owned<moab::MBVERTEX>(mb, file_set);
  }

  //! return the number of local quads    
  std::size_t 
  get_local_num_quads(const moab::Interface& mb,
		      moab::EntityHandle& file_set)
  {
    return get_local_num_entities_owned<moab::MBQUAD>(mb, file_set);
  }
  
  //! return the number of local polygons    
  std::size_t 
  get_local_num_polygons(const moab::Interface& mb,
		      moab::EntityHandle& file_set)
  {
    return get_local_num_entities_owned<moab::MBPOLYGON>(mb, file_set);
  }
  
  //! return the number of local edges    
  std::size_t 
  get_local_num_edges(const moab::Interface& mb,
		      moab::EntityHandle& file_set)
  {
    return get_local_num_entities_owned<moab::MBEDGE>(mb, file_set);
  }
  
  //! return the dimensions of a variable loaded in the mesh
  std::vector<std::size_t>
  get_var_dims(const moab::Interface& mb, 
	       moab::EntityHandle& file_set,
	       const std::string& var_name) 
  {
    std::vector<size_t> dims;
    moab::Tag tag = 0;
    std::string tag_name;
    int sz = 0;
    tag_name = "__"+var_name+"_DIMS";
    moab::ErrorCode rval = mb.tag_get_handle(tag_name.c_str(), 0, moab::MB_TYPE_OPAQUE, tag, moab::MB_TAG_ANY);
    if (rval != moab::MB_SUCCESS)
      throw pargal_except("Error: " + mb.get_error_string(rval),
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    
    rval = mb.tag_get_length(tag, sz);
    dims.resize(sz, 0);
    const void* ptr = NULL;
    mb.tag_get_by_ptr(tag, &file_set, 1, &ptr);
    const moab::Tag * p = static_cast<const moab::Tag*>(ptr);
    for (int i = 0; i != sz; ++i) 
      {
	std::string dim_name;
	rval = mb.tag_get_name(p[i], dim_name);
	int dim_sz;
	moab::Tag dim_tag;
	rval = mb.tag_get_handle(dim_name.c_str(), 0, moab::MB_TYPE_OPAQUE, dim_tag, moab::MB_TAG_ANY);
	if (rval != moab::MB_SUCCESS)
	  throw pargal_except("Error: " + mb.get_error_string(rval),
			      __FILE__, __LINE__, __PRETTY_FUNCTION__);
	const void* dim_ptr = NULL;
	mb.tag_get_by_ptr(dim_tag, &file_set, 1, &dim_ptr, &dim_sz);
	dims[i] = dim_sz;
      }
    return dims;
  } 

  void 
  remove_gather_set(const moab::Interface& constmb, moab::EntityHandle& file_set)
  {
    moab::Interface& mb = const_cast<moab::Interface&>(constmb);
    moab::EntityHandle gset=0;
    moab::Tag gathersettag;
    moab::ErrorCode rval = mb.tag_get_handle("GATHER_SET", 1, moab::MB_TYPE_INTEGER, gathersettag, moab::MB_TAG_SPARSE);
    if (rval != moab::MB_SUCCESS) return;
      
    moab::Range gsets;
    int gatherval = 1;
    void *dim[] = {&gatherval};
    rval = mb.get_entities_by_type_and_tag(0, moab::MBENTITYSET, &gathersettag, dim, 1, gsets);
    if (rval != moab::MB_SUCCESS)     
      throw pargal_except("Error: Couldn't get gather set.",
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);     
    assert((int)gsets.size()==1);
    gset = gsets[0];    
    
    moab::Range gents;
    mb.get_entities_by_handle(gset, gents);
    rval = mb.remove_entities(file_set, gents);
    
    if (rval != moab::MB_SUCCESS)     
      throw pargal_except("Error: Couldn't remove gather ents.",
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);     
    rval = mb.remove_entities(file_set, &gset, 1);
    if (rval != moab::MB_SUCCESS)     
      throw pargal_except("Error: Couldn't remove gather set.",
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);       
  }
 
} // namespace pargal
