/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <pargal/pcatt.hpp>
#include <pargal/exception.hpp>
#include <iostream>

namespace pargal 
{  
  pcatt::pcatt(const std::string& name, 
	       const std::string& val,
	       const std::vector<int>& att_len) :
    m_var_name(name), m_val(val)
  {
    // for variables without any attribute, simply return
    if ( (val.size() == 0) && 
	 (att_len.size() == 0) && 
	 (att_len[0] == 0))
      return;
    else 
      {
	std::size_t start = 0;
	std::size_t att_counter = 0;
	for (std::size_t i = 0; i != m_val.size(); ++i)
	  {
	    if (m_val[i] == '\0') 
	      {
		std::string att_name(&m_val[start], i-start); 
		start=i+1;
		while (m_val[i] != ';') ++i;
		std::string data_type(&m_val[start], i-start);
		++i;
		start=i;
		i = att_len[att_counter];
		if (m_val[i] != ';')
		  throw pargal_except("Error: Paring Attributes",
				      __FILE__, __LINE__, __PRETTY_FUNCTION__);

		std::string data_val(&m_val[start], i-start);
		start=i+1;
		m_atts[att_name] = att_data(data_type, data_val);
		++att_counter;
	      }
	  }
      }
  }

  const std::string&
  pcatt::get_var_name() const
  {
    return m_var_name;
  }

  std::size_t
  pcatt::size() const
  {
    return m_atts.size();
  }

  std::set<std::string>
  pcatt::get_att_names() const
  {
    std::set<std::string> s;
    std::map<std::string, att_data>::const_iterator it_att = m_atts.begin();
    for (;it_att != m_atts.end(); ++it_att)
      s.insert(it_att->first);
    return s;
  }
  
  const std::string&
  pcatt::get_data_type(const std::string& att_name) const
  {
	std::map<std::string, att_data>::const_iterator iter = m_atts.find(att_name);
	if(iter != m_atts.end()){
		return (m_atts.find(att_name))->second.get_data_type();
	}
	else{
		throw pargal_except("Error: Unable to get data type for attribute",
				__FILE__, __LINE__, __PRETTY_FUNCTION__);
	}
  }
  
  void 
  pcatt::get_vals(const std::string& att_name, std::string& vals) const
  {
	std::map<std::string, att_data>::const_iterator iter = m_atts.find(att_name);
	if(iter == m_atts.end()){
		throw pargal_except("Error: Unable to get data type for attribute",
				__FILE__, __LINE__, __PRETTY_FUNCTION__);
	}

    if (iter->second.get_data_type() != "char")
      throw pargal_except("Error: Data type is not char",
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    vals = iter->second.get_val();
  }
  
  void 
  pcatt::get_vals(const std::string& att_name, std::vector<double>& vals) const
  {
	std::map<std::string, att_data>::const_iterator iter = m_atts.find(att_name);
	if(iter == m_atts.end()){
		throw pargal_except("Error: Unable to get data type for attribute",
				__FILE__, __LINE__, __PRETTY_FUNCTION__);
	}

    const void* tmp = iter->second.get_val().c_str();
    double* val_arr = (double *)tmp;
    int sz = iter->second.get_val().size() / sizeof(double);
    vals.resize(sz);
    for (std::size_t counter = 0; counter != static_cast<std::size_t>(sz); ++counter)
      vals[counter] = val_arr[counter];
  }

  void 
  pcatt::get_vals(const std::string& att_name, std::vector<float>& vals) const
  {
	std::map<std::string, att_data>::const_iterator iter = m_atts.find(att_name);
	if(iter == m_atts.end()){
		throw pargal_except("Error: Unable to get data type for attribute",
				__FILE__, __LINE__, __PRETTY_FUNCTION__);
	}

    const void* tmp = iter->second.get_val().c_str();
    float* val_arr = (float *)tmp;
    int sz = iter->second.get_val().size() / sizeof(float);
    vals.resize(sz);
    for (std::size_t counter = 0; counter != static_cast<std::size_t>(sz); ++counter)
      vals[counter] = val_arr[counter];
  }
    
  void 
  pcatt::get_vals(const std::string& att_name, std::vector<int>& vals) const 
  {
	std::map<std::string, att_data>::const_iterator iter = m_atts.find(att_name);
	if(iter == m_atts.end()){
		throw pargal_except("Error: Unable to get data type for attribute",
				__FILE__, __LINE__, __PRETTY_FUNCTION__);
	}

    const void* tmp = iter->second.get_val().c_str();
    int* val_arr = (int *)tmp;
    int sz = iter->second.get_val().size() / sizeof(int);
    vals.resize(sz);
    for (std::size_t counter = 0; counter != static_cast<std::size_t>(sz); ++counter)
      vals[counter] = val_arr[counter];
  }

  void 
  pcatt::get_vals(const std::string& att_name, std::vector<short>& vals) const
  {
	std::map<std::string, att_data>::const_iterator iter = m_atts.find(att_name);
	if(iter == m_atts.end()){
		throw pargal_except("Error: Unable to get data type for attribute",
				__FILE__, __LINE__, __PRETTY_FUNCTION__);
	}

    const void* tmp = iter->second.get_val().c_str();
    short* val_arr = (short *)tmp;
    int sz = iter->second.get_val().size() / sizeof(short);
    vals.resize(sz);
    for (std::size_t counter = 0; counter != static_cast<std::size_t>(sz); ++counter)
      vals[counter] = val_arr[counter];
  }    
} // namespace pargal
