/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <pargal/pcvar.hpp>
#include <moab/ParallelComm.hpp>

namespace pargal 
{
  //! return the size of a variable (owned) for a single time step
  std::size_t 
  pcvar::tstep_size()
  {         
    moab::ErrorCode rval = moab::MB_SUCCESS;
    moab::Interface& mb = m_finfo.get_moab();
    moab::EntityHandle& file_set = m_finfo.get_file_set();

    if(tstep_sz_ != 0){
        return tstep_sz_;
    }
    if (m_loc == "QUAD") {
      moab::Range quads_owned = get_local_entities_owned<moab::MBQUAD>(mb, file_set);
      int nquads = quads_owned.size();
      tstep_sz_ = static_cast<std::size_t>(nquads*m_num_lev);
    }
    else if (m_loc == "POLYGON") {
      moab::Range polygons_owned = get_local_entities_owned<moab::MBPOLYGON>(mb, file_set);
      int npolygons = polygons_owned.size();
      tstep_sz_ = static_cast<std::size_t>(npolygons*m_num_lev);
    }
    else if (m_loc == "VERTEX") {
      moab::Range verts_owned = get_local_entities_owned<moab::MBVERTEX>(mb, file_set);
      int nverts = verts_owned.size();
      tstep_sz_ = static_cast<std::size_t>(nverts*m_num_lev);
    }
    else if (m_loc == "EDGE") {
      moab::Range edges_owned = get_local_entities_owned<moab::MBEDGE>(mb, file_set);
      int nedges = edges_owned.size();
      tstep_sz_ = static_cast<std::size_t>(nedges*m_num_lev);
    }
    else if (m_loc == "SET") {
        moab::Tag tmp_tag;
        const void *tagp = NULL;
        int tag_len = 0;

        rval = mb.tag_get_handle(m_name.c_str(), 0, moab::MB_TYPE_DOUBLE, tmp_tag, moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN);
        if(rval != moab::MB_SUCCESS){
            throw pargal_except("Error: Unable to get tag handle" + mb.get_error_string(rval), __FILE__, __LINE__, __PRETTY_FUNCTION__);
        }

        moab::EntityHandle mb_file_set = m_finfo.get_file_set();
        rval = const_cast<moab::Interface&>(mb).tag_get_by_ptr(tmp_tag, &mb_file_set, 1, &tagp, &tag_len);
        if(rval != moab::MB_SUCCESS){
            throw pargal_except("Error: Unable to get tag length" + mb.get_error_string(rval), __FILE__, __LINE__, __PRETTY_FUNCTION__);
        }
        tstep_sz_ = tag_len;
    }
    else
      throw pargal_except("Error: pcvar::tstep_size for locations other than QUAD/VERTEX/POLYGON/EDGE are not implemented yet!",
              __FILE__, __LINE__, __PRETTY_FUNCTION__);
    return tstep_sz_;
  }
  
  //! return the size of a variable (owned) including all time steps
  std::size_t 
  pcvar::size()
  {         
    return m_tsteps.size() * tstep_size();
  }
  
	//! return the local dimensions of the variable
	void pcvar::get_local_dims(std::vector<std::size_t> &ldims)
	{
		std::vector<std::size_t> mesh_conform_dims;
		// Get global dims
		get_dims(ldims);
		if(m_loc == "SET"){
			return;
		}
		const std::string& grid_type = m_finfo.get_grid_type();
		if (grid_type == "CAM_UNKNOWN" || grid_type == "NOT_CAM"){
			throw pargal_except("Error: Unknown grid type, cannot compute local dims", __FILE__, __LINE__, "Unknown");
		}
		else if(grid_type == "CAM_SE"){
			std::size_t tstep_sz = tstep_size();
			std::size_t ndims = num_dims();
			int rdim = static_cast<int>(ndims - 1);
			std::size_t rdim_sz = get_dim(rdim).size();
			
			assert(get_dim(rdim).name() == "ncol");

			std::size_t tmp_sz = 1;
			for(int dim=rdim-1; dim >= 0; dim--){
				if(get_dim(dim).name() == "time"){
					break;
				}
				tmp_sz *= get_dim(dim).size();
			}

			mesh_conform_dims.push_back(tstep_sz/tmp_sz);
		}
		else if(grid_type == "MPAS"){
			std::size_t tstep_sz = tstep_size();
			std::size_t ndims = num_dims();
			int rdim = static_cast<int>(ndims - 1);
			std::size_t conform_dim_idx = -1;

			std::size_t tmp_sz = 1;
			for(int dim=rdim, i=0; dim>=0; dim--, i++){
				std::string dim_name = get_dim(dim).name();
				if(dim_name == "Time"){
					break;
				}
				else if((dim_name == "nCells") ||
						(dim_name == "nEdges") ||
						(dim_name == "nVertices")){
					conform_dim_idx = i;
					// Will be modified outside the loop after we calculate
					// the total size, tmp_sz, excluding conform dim
					mesh_conform_dims.push_back(0);
				}
				else{
					size_t dim_sz = get_dim(dim).size();
					tmp_sz *= dim_sz;
					// Don't change the dim size
					mesh_conform_dims.push_back(dim_sz);	
				}
			}

			assert(conform_dim_idx != -1);
			mesh_conform_dims[conform_dim_idx] = tstep_sz/tmp_sz;
		}
		else{
			if(m_finfo.get_part_method() != "SQIJ"){
				throw pargal_except("Only SQIJ partition is currently supported", __FILE__, __LINE__, "Unknown");
			}
			// We assume that variables that are not on "SET"
			// are atleast two dimensional
			// FIXME: What about "slon"/"slat" ? - we need to have
			// different cases for "lon", "slon", "lat", "slat"
			// FIXME: We don't support sparse tags yet
			std::vector<int> mesh_ij_minmax = get_dim_loc_minmax_owned();
			std::size_t ni = static_cast<std::size_t> (mesh_ij_minmax[3] - mesh_ij_minmax[2] + 1);
			std::size_t nj = static_cast<std::size_t> (mesh_ij_minmax[1] - mesh_ij_minmax[0] + 1);
			mesh_conform_dims.push_back(nj);
			mesh_conform_dims.push_back(ni);
		}

		std::vector<std::size_t>::iterator citer = mesh_conform_dims.begin();
		for(std::vector<std::size_t>::reverse_iterator liter = ldims.rbegin();
			(liter != ldims.rend()) && (citer != mesh_conform_dims.end());
			++liter, ++citer){
			*liter = *citer;	
		}
	}
  
  //! return the pointer pointing to the owned data at time step indexed by tindex  
  double* 
  pcvar::get_storage(int tindex)
  {        
	moab::ErrorCode rval = moab::MB_SUCCESS;

    if (m_loc != "QUAD" && m_loc != "POLYGON" && m_loc != "VERTEX" &&
		m_loc != "EDGE" && m_loc != "SET")
      throw pargal_except("Error: pcvar::get_storage for locations other than QUAD/VERTEX/SET/POLYGON/EDGE are not implemented yet!", 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    
	if(mesh_chunks_cache_[tindex].second > 0){
		return mesh_chunks_cache_[tindex].first;
	}
    
    moab::Interface& mb = m_finfo.get_moab();
    moab::EntityHandle& file_set = m_finfo.get_file_set();
    
	moab::Range ent_range;
	int nents = 0;
    
    if (m_loc == "QUAD") {
		ent_range = get_local_entities_owned<moab::MBQUAD>(mb, file_set);
		nents = ent_range.size();
    }
    else if (m_loc == "POLYGON") {
		ent_range = get_local_entities_owned<moab::MBPOLYGON>(mb, file_set);
		nents = ent_range.size();
    }
    else if (m_loc == "VERTEX") {
		ent_range = get_local_entities_owned<moab::MBVERTEX>(mb, file_set);
		nents = ent_range.size();
    }
    else if (m_loc == "EDGE") {
		ent_range = get_local_entities_owned<moab::MBEDGE>(mb, file_set);
		nents = ent_range.size();
    }
	else if (m_loc == "SET"){
		nents = 1;
		for(std::size_t i=0; i < m_dims.size(); i++){
			nents *= m_dims[i].size();	
		}
	}
    
    // allocate space for usr created variable
	unsigned int mb_tag_flags = 0x0;
	if(m_loc == "SET"){
		mb_tag_flags = moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN; 
	}
	else{
		mb_tag_flags = moab::MB_TAG_DENSE;
	}

	std::ostringstream tname;
	// get the tag values 
	moab::Tag usr_tag;
	int ent_sz = (m_loc == "SET") ? nents : m_num_lev;
    if (m_usr_created){
		mb_tag_flags |= moab::MB_TAG_CREAT;
		if(m_loc == "SET"){
			tname << m_name;
		}
		else{
			tname << m_name << m_tsteps[tindex];
		}
		//const double default_val = 0.0;
		rval = mb.tag_get_handle(tname.str().c_str(), ent_sz, moab::MB_TYPE_DOUBLE, usr_tag, mb_tag_flags);
		if (rval != moab::MB_SUCCESS) 
		  throw pargal_except("Error: " + mb.get_error_string(rval), 
					  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    }
    else{
		std::stringstream opts;
		opts << m_opts.str();
		if(m_loc != "SET"){
			opts << "TIMESTEP=";
			if (m_finfo.file_names().size() == 1) 
			  opts << m_tsteps[tindex];
			else
			  opts << 0;
			opts << ";";
		}

		// load variable from file
		std::string file_name = m_finfo.tstep_lookup(m_tsteps[tindex]);
		moab::ErrorCode rval = mb.load_file(file_name.c_str(),
							&file_set,
							opts.str().c_str());
		if (rval != moab::MB_SUCCESS) 
		  throw pargal_except("Error: " + mb.get_error_string(rval), 
					  __FILE__, __LINE__, __PRETTY_FUNCTION__);
			
		if(m_loc == "SET"){
			tname << m_name;
		}
		else{
			if (m_finfo.file_names().size() == 1) 
			  tname << m_name << m_tsteps[tindex];
			else
			  tname << m_name << 0;
		}
		rval = mb.tag_get_handle(tname.str().c_str(), ent_sz, moab::MB_TYPE_DOUBLE, usr_tag, mb_tag_flags);
		if (rval != moab::MB_SUCCESS) 
		  throw pargal_except("Error: " + mb.get_error_string(rval), 
					  __FILE__, __LINE__, __PRETTY_FUNCTION__);
	}    

	void *usr_tag_ptr;
	int count, buf_len;
	if(m_loc == "SET"){
		const void *tagp = NULL;
		moab::EntityHandle mb_file_set = m_finfo.get_file_set();
		rval = const_cast<moab::Interface&>(mb).tag_get_by_ptr(usr_tag, &mb_file_set, 1, &tagp, &count);
		// FIXME: This is ugly. Is there another way ?
		usr_tag_ptr = const_cast<void *>(tagp);
		buf_len = count;
	}
	else{
		rval = mb.tag_iterate(usr_tag, ent_range.begin(), ent_range.end(), count, usr_tag_ptr);
		buf_len = count * m_num_lev;
	}
	if (rval != moab::MB_SUCCESS || count != (int)nents) 
		  throw pargal_except("Error: " + mb.get_error_string(rval), 
					  __FILE__, __LINE__, __PRETTY_FUNCTION__);

	mesh_chunks_cache_[tindex] = std::make_pair<double *, std::size_t>(
									static_cast<double *>(usr_tag_ptr), buf_len);
	return static_cast<double *>(usr_tag_ptr);
  }
  
bool
pcvar::cache_read(int chunk_idx, double *&buf, std::size_t &buf_len, bool cont, cache_access_state &rd_state)
{        
	moab::ErrorCode rval = moab::MB_SUCCESS;
		
	if(m_loc != "QUAD" && m_loc != "POLYGON" && m_loc != "VERTEX" &&
		m_loc != "EDGE" && m_loc != "SET"){
		throw pargal_except("Error: pcvar::get_storage for locations other than QUAD/VERTEX/SET/POLYGON/EDGE are not implemented yet!", __FILE__, __LINE__, __PRETTY_FUNCTION__);
	}

	moab::Interface& mb = m_finfo.get_moab();
	moab::EntityHandle& file_set = m_finfo.get_file_set();

	moab::Range &ent_range = rd_state.ent_range;
	moab::Range::iterator mb_iter_begin, mb_iter_end;
	int nents = 0;
	// Total chunk size
	std::size_t chunk_sz = 0;

	if(m_loc == "QUAD"){
		if(cont == false){
			ent_range = get_local_entities_owned<moab::MBQUAD>(mb, file_set);
		}
		nents = ent_range.size();
		chunk_sz = nents * m_num_lev;
	}
	else if(m_loc == "POLYGON"){
		if(cont == false){
			ent_range = get_local_entities_owned<moab::MBPOLYGON>(mb, file_set);
		}
		nents = ent_range.size();
		chunk_sz = nents * m_num_lev;
	}
	else if(m_loc == "VERTEX"){
		if(cont == false){
			ent_range = get_local_entities_owned<moab::MBVERTEX>(mb, file_set);
		}
		nents = ent_range.size();
		chunk_sz = nents * m_num_lev;
	}
	else if(m_loc == "EDGE"){
		if(cont == false){
			ent_range = get_local_entities_owned<moab::MBEDGE>(mb, file_set);
		}
		nents = ent_range.size();
		chunk_sz = nents * m_num_lev;
	}
	else if(m_loc == "SET"){
		nents = 1;
		for(std::size_t i=0; i < m_dims.size(); i++){
			nents *= m_dims[i].size();	
		}
		chunk_sz = nents;
	}

	if(cont == false){
		mb_iter_begin = ent_range.begin();
		mb_iter_end = ent_range.end();
	}
	else{
		mb_iter_begin = rd_state.begin;
		mb_iter_end = ent_range.end();
	}

	if((cont == false) && (mesh_chunks_cache_[chunk_idx].second > 0)){
		// First call aka "not cont" && a valid first chunk is cached
		buf = mesh_chunks_cache_[chunk_idx].first;
		buf_len = mesh_chunks_cache_[chunk_idx].second;
		if(buf_len < chunk_sz){
			// More sequences to read
			rd_state.begin = ent_range.begin() + buf_len/m_num_lev;
			return true;
		}
		else{
			return false;
		}
	}

	// allocate space for usr created variable
	unsigned int mb_tag_flags = 0x0;
	if(m_loc == "SET"){
		mb_tag_flags = moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN; 
	}
	else{
		mb_tag_flags = moab::MB_TAG_DENSE;
	}

	std::ostringstream tname;
	// get the tag values 
	moab::Tag usr_tag;
	int ent_sz = (m_loc == "SET") ? nents : m_num_lev;

	if(m_usr_created){
		if(cont == false){
			mb_tag_flags |= moab::MB_TAG_CREAT;
		}
		if(m_loc == "SET"){
			tname << m_name;
		}
		else{
			tname << m_name << m_tsteps[chunk_idx];
		}
		//const double default_val = 0.0;
		rval = mb.tag_get_handle(tname.str().c_str(), ent_sz, moab::MB_TYPE_DOUBLE, usr_tag, mb_tag_flags);
		if((rval != moab::MB_SUCCESS) &&
			// The user is allowed to allocate mem for tag - skip this error
			// for user created tags
			(rval != moab::MB_ALREADY_ALLOCATED)){
			throw pargal_except("Error: " + mb.get_error_string(rval), 
					__FILE__, __LINE__, __PRETTY_FUNCTION__);
		}
	}
	else{
		if(cont == false){
			// load variable from file
			std::stringstream opts;
			opts << m_opts.str();
			if(m_loc != "SET"){
				opts << "TIMESTEP=";
				if(m_finfo.file_names().size() == 1){ 
					opts << m_tsteps[chunk_idx];
				}
				else{
					opts << 0;
				}
				opts << ";";
			}

			std::string file_name = m_finfo.tstep_lookup(m_tsteps[chunk_idx]);
			moab::ErrorCode rval = mb.load_file(file_name.c_str(), &file_set,
												opts.str().c_str());
			if(rval != moab::MB_SUCCESS){ 
				throw pargal_except("Error: " + mb.get_error_string(rval), 
						__FILE__, __LINE__, __PRETTY_FUNCTION__);
			}
		}
	
		if(m_loc == "SET"){
			tname << m_name;
		}
		else{
			if(m_finfo.file_names().size() == 1){ 
				tname << m_name << m_tsteps[chunk_idx];
			}
			else{
				tname << m_name << 0;
			}
		}
		rval = mb.tag_get_handle(tname.str().c_str(), ent_sz, moab::MB_TYPE_DOUBLE,
									usr_tag, mb_tag_flags);
		if(rval != moab::MB_SUCCESS){ 
			throw pargal_except("Error: " + mb.get_error_string(rval), 
					__FILE__, __LINE__, __PRETTY_FUNCTION__);
		}
	}    

	void *usr_tag_ptr;
	int count;	
	if(m_loc == "SET"){
		const void *tagp = NULL;
		moab::EntityHandle mb_file_set = m_finfo.get_file_set();
		rval = const_cast<moab::Interface&>(mb).tag_get_by_ptr(usr_tag,
					&mb_file_set, 1, &tagp, &count);
		// FIXME: This is ugly. Is there another way ?
		usr_tag_ptr = const_cast<void *>(tagp);
		buf = static_cast<double *>(usr_tag_ptr);
		buf_len = count;
	}
	else{
		rval = mb.tag_iterate(usr_tag, mb_iter_begin, mb_iter_end,
				count, usr_tag_ptr);
		buf = static_cast<double *>(usr_tag_ptr);
		buf_len = count * m_num_lev;
		mb_iter_begin = mb_iter_begin + count;
	}
	if((rval != moab::MB_SUCCESS) || (count <= 0)){ 
		throw pargal_except("Error: " + mb.get_error_string(rval), 
				__FILE__, __LINE__, __PRETTY_FUNCTION__);
	}

	if(cont == false){
		// We only cache the first seq for now
		mesh_chunks_cache_[chunk_idx] = std::make_pair<double *, std::size_t>(
										buf, buf_len);
	}
	if(mb_iter_begin != mb_iter_end){
		// More data to read on this mesh chunk
		rd_state.begin = mb_iter_begin;
		return true;
	}
	else{
		return false;
	}
}

void
pcvar::cache_delete(int cache_idx)
{
	pcvar::delete_storage(cache_idx);
}

	//! Set data cache storage mode
	void
	pcvar::set_storage_mode(pcvar::cache_mode mode)
	{
		storage_mode_ = mode;
	}

	//! Get data cache storage mode
	// FIXME: Is a getter needed?
	pcvar::cache_mode
	pcvar::get_storage_mode(void ) const
	{
		return storage_mode_;
	}
  
  //! delete the data at time step indexed by tindex
  void 
  pcvar::delete_storage(int tindex)
  {
	if(mesh_chunks_cache_[tindex].second <= 0){
		return;
	}

	if(m_loc == "SET"){
		return;
	}

	if(storage_mode_ == OPT_COMP){
		// Don't delete cached var data
		return;
	}

    moab::Interface& mb = m_finfo.get_moab();
    std::ostringstream tname;
    tname << this->name() << m_tsteps[tindex];
    moab::Tag tagh;
    mb.tag_get_handle(tname.str().c_str(), 0, moab::MB_TYPE_OPAQUE, tagh, moab::MB_TAG_ANY);
    mb.tag_delete(tagh);
	mesh_chunks_cache_[tindex].first = NULL;
	mesh_chunks_cache_[tindex].second = 0;
  }
  
  //! return the time steps the variable is referring to
  const std::vector<int>&   
  pcvar::get_tsteps() const 
  {
    return m_tsteps;
  }

  //! return the index for a given time step
  int 
  pcvar::get_index(int tstep) const
  {
    for (std::size_t i = 0; i != m_tsteps.size(); ++i)
      if (m_tsteps[i] == tstep)
	return i;
    return -1;
  }
    
  //! return the moab tag at time step indexed by tindex
  moab::Tag 
  pcvar::get_moab_tag(int tindex)
  {
    moab::Interface& mb = m_finfo.get_moab();
    get_storage(m_tsteps[tindex]);
    std::stringstream ss_var;
     if (m_finfo.file_names().size() == 1) 
     ss_var << m_name << m_tsteps[tindex];
     else
     ss_var << m_name << 0;
    moab::Tag tmpTag;
    moab::ErrorCode rval=mb.tag_get_handle(ss_var.str().c_str(), m_num_lev, moab::MB_TYPE_DOUBLE, tmpTag);
    if (rval != moab::MB_SUCCESS)
      throw pargal_except("Error: " + mb.get_error_string(rval), 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    return tmpTag;        
  }

	//! return moab tag iterators
	pcvar::iterator
	pcvar::get_chunk_iterator(const size_t odim)
	{
		return util::moab_tag_chunk_iterator(*this, 0, odim);
	}

	pcvar::red_iterator
	pcvar::get_red_chunk_iterator(const size_t odim)
	{
		return util::moab_tag_reset_chunk_iterator(*this, 0, odim);
	}

	//! return the dim type
	util::moab_tag_dim_type pcvar::get_dim_type(const size_t dim)
	{
		if(m_loc == "SET"){
			return util::MOAB_TAG_SPARSE_DIM;
		}

		std::string part_method = m_finfo.get_part_method();
		if((part_method == "SQIJ") || (part_method == "TRIVIAL_PARTITION")){
			// [gen]...[gen][lev/ilev][lat/slat][lon/slon]
			// [gen]...[gen][ilev/lev][ncol]
			// [gen]...[gen][nEdges/nCells][nVertLevels]
			size_t rdim = num_dims() - 1 - dim;
			std::string dim_name = get_dim(dim).name();
			switch(rdim){
				case 0:
					if((dim_name == "lon") ||
						(dim_name == "slon")){
						return util::MOAB_TAG_SQIJ_LON_DIM;
					}
					if(dim_name == "nVertLevels"){
						return util::MOAB_TAG_LEV_DIM;
					}
					break;
				case 1:
					if((dim_name == "lat") ||
						(dim_name == "slat")){
						return util::MOAB_TAG_SQIJ_LAT_DIM;
					}
					// continue
				case 2:
					if((dim_name == "lev") ||
						(dim_name == "ilev")){
						return util::MOAB_TAG_LEV_DIM;
					}
					// continue
				default:
					return util::MOAB_TAG_GEN_DIM;
			}
		}
		return util::MOAB_TAG_INVALID_DIM;
	}

  //! return the local min/max of the dimension (owned only)
  //! assume sqij partition method
  //! returned value is in imin/imax/jmin/jmax order
  std::vector<int> 
  pcvar::get_dim_loc_minmax_owned() const
  {
    if (m_loc != "QUAD") {
      throw pargal_except("Error: Locations other than QUAD are not implemented yet!", 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    }
	// FIXME: What about lDims ?
    return lCDims;    
  }
  
  //! return the global min/max of the dimension
  //! returned value is in imin/imax/jmin/jmax order
  std::vector<int> 
  pcvar::get_dim_global_minmax() const
  {    
    if (m_loc != "QUAD") {
      throw pargal_except("Error: Locations other than QUAD are not implemented yet!", 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    }
    return gCDims;
  }

  //! return the processes' rank whose column is the same as the current process
  //! current process's rank is also included
  std::vector<int> 
  pcvar::get_column_ranks() const
  {
    const moab::ParallelComm& mb_pcomm = this->get_finfo().get_pcomm();
    std::size_t np = mb_pcomm.proc_config().proc_size();
    std::size_t nr = mb_pcomm.proc_config().proc_rank();
	// FIXME: This is wrong - 2 & 3 dims are assumed as lat, lon
    std::size_t lat = this->get_dim(2).size();
    std::size_t lon = this->get_dim(3).size();
    
    if (np == 1) 
      return std::vector<int>(1, 0);

    std::vector<double> pfactors, ppfactors;
    for (std::size_t i = 2; i <= np; ++i)
      {
	if (!(np%i))
	  {
	    pfactors.push_back(i);
	    ppfactors.push_back(((double)(i*i))/np);
	  }
      }
    double ijratio = ((double) lon)/((double) lat);
    std::size_t ind = std::lower_bound(ppfactors.begin(), ppfactors.end(), ijratio) - ppfactors.begin();
    if (ind && fabs(ppfactors[ind-1]-ijratio) < fabs(ppfactors[ind]-ijratio)) --ind;
    
    int pi = pfactors[ind];
    int nri = nr % pi;
    std::vector<int> res;
    for (int rank = 0; rank != static_cast<int>(np); ++rank)
      { 
	if (rank%pi == nri)
	  res.push_back(rank);
      }
    return res;
  }

  //! return the processes' rank whose row is the same as the current process
  //! current process's rank is also included
  std::vector<int> 
  pcvar::get_row_ranks() const
  {
    const moab::ParallelComm& mb_pcomm = this->get_finfo().get_pcomm();
    std::size_t np = mb_pcomm.proc_config().proc_size();
    std::size_t nr = mb_pcomm.proc_config().proc_rank();
	// FIXME: This is wrong - 2 & 3 dims are assumed as lat, lon
    std::size_t lat = this->get_dim(2).size();
    std::size_t lon = this->get_dim(3).size();
    
    if (np == 1) 
      return std::vector<int>(1, 0);

    std::vector<double> pfactors, ppfactors;
    for (std::size_t i = 2; i <= np; ++i)
      {
	if (!(np%i))
	  {
	    pfactors.push_back(i);
	    ppfactors.push_back(((double)(i*i))/np);
	  }
      }
    double ijratio = ((double) lon)/((double) lat);
    std::size_t ind = std::lower_bound(ppfactors.begin(), ppfactors.end(), ijratio) - ppfactors.begin();
    if (ind && fabs(ppfactors[ind-1]-ijratio) < fabs(ppfactors[ind]-ijratio)) --ind;
    
    int pi = pfactors[ind];
    int nrj = nr / pi;
    std::vector<int> res;
    for (int rank = 0; rank != static_cast<int>(np); ++rank)
      { 
	if (rank/pi == nrj)
	  res.push_back(rank);
      }
    return res;
  }
  
  void
  pcvar::init_dim_loc_minmax_owned()
  { 
    const std::string& grid_type = m_finfo.get_grid_type();
	if((grid_type != "CAM_EUL") &&
		(grid_type != "CAM_FV")){
		return;
	}

    const moab::Interface& mb = m_finfo.get_moab();
    std::vector<std::string> dim_name(4);
    dim_name[0] = "__slon";
    dim_name[1] = "__slat";
    dim_name[2] = "__lon";
    dim_name[3] = "__lat";
    std::vector<int> vec(8,0);
    for (std::size_t i = 0; i != dim_name.size(); ++ i)
      {
	std::stringstream ss_tag_name;
	ss_tag_name << dim_name[i] << "_LOC_MINMAX";
	moab::Tag tmpTag;
	const void* data = NULL;
	moab::ErrorCode rval=mb.tag_get_handle(ss_tag_name.str().c_str(), 0, moab::MB_TYPE_INTEGER, tmpTag);
	int sz = 0;
	rval = mb.tag_get_by_ptr(tmpTag, &m_finfo.get_file_set(), 1, &data, &sz);
	if (rval != moab::MB_SUCCESS)
	  throw pargal_except("Error: " + mb.get_error_string(rval), 
			      __FILE__, __LINE__, __PRETTY_FUNCTION__);
	const int* dataInt = static_cast<const int*>(data);
	vec[i*2] = *dataInt;
	vec[i*2+1] = *(dataInt+1);
      }
    
    lDims[0] = vec[0];
    lDims[1] = vec[1];
    lDims[2] = vec[2];
    lDims[3] = vec[3];
    
    lCDims[0] = vec[4];
    lCDims[1] = vec[5];
    lCDims[2] = vec[6];
    lCDims[3] = vec[7];
  }

  void
  pcvar::init_dim_global_minmax()
  {
    const std::string& grid_type = m_finfo.get_grid_type();
	if((grid_type != "CAM_EUL") &&
		(grid_type != "CAM_FV")){
		return;
	}

    const moab::Interface& mb = m_finfo.get_moab();
    std::vector<std::string> dim_name(4);
    dim_name[0] = "__slon";
    dim_name[1] = "__slat";
    dim_name[2] = "__lon";
    dim_name[3] = "__lat";
    std::vector<int> vec(8,0);
    for (std::size_t i = 0; i != dim_name.size(); ++ i)
      {
	std::stringstream ss_tag_name;
	ss_tag_name << dim_name[i] << "_GLOBAL_MINMAX";
	moab::Tag tmpTag;
	const void* data = NULL;
	moab::ErrorCode rval=mb.tag_get_handle(ss_tag_name.str().c_str(), 0, moab::MB_TYPE_INTEGER, tmpTag);
	int sz = 0;
	rval = mb.tag_get_by_ptr(tmpTag, &m_finfo.get_file_set(), 1, &data, &sz);
	if (rval != moab::MB_SUCCESS)
	  throw pargal_except("Error: " + mb.get_error_string(rval), 
			      __FILE__, __LINE__, __PRETTY_FUNCTION__);
	const int* dataInt = static_cast<const int*>(data);
	vec[i*2] = *dataInt;
	vec[i*2+1] = *(dataInt+1);
      }

    gDims[0] = vec[0];
    gDims[1] = vec[1];
    gDims[2] = vec[2];
    gDims[3] = vec[3];
    
    gCDims[0] = vec[4];
    gCDims[1] = vec[5];
    gCDims[2] = vec[6];
    gCDims[3] = vec[7];
  }

  
} // namespace pargal
