/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <numeric>
#include <functional>
#include <cstddef>
#include "pargal/moab_tag_counter.hpp"
#include "pargal/moab_tag_chunk.hpp"

// Moab tag chunk counter members
void pargal::util::moab_tag_chunk_counter::init_seq_counters(
		pargal::util::moab_tag_chunk &tc)
{
	if(is_seq_counters_init_) return;

	int *begin, *end;
	begin = &chunk_counter_[0];
	end = &chunk_counter_[chunk_counter_.size()];
	switch(odim_type_){
		case MOAB_TAG_GEN_DIM:
		case MOAB_TAG_LEV_DIM:
			// Go through tag sequences
			for(pargal::util::moab_tag_chunk::iterator iter = tc.begin();
					iter != tc.end(); ++iter){
				pargal::util::moab_tag_chunk_seq& chunk_seq = *iter;
				end = begin + (chunk_seq.size() / odim_tc_to_counter_ratio_);
				seq_counters_.push_back(moab_tag_seq_counter(begin, end));
				begin = end;
			}
			break;
		default:
			// We don't care about handling sequences for other dims
			seq_counters_.push_back(moab_tag_seq_counter(begin, end));
	}
	is_seq_counters_init_ = true;
}
// Var red counter members
pargal::util::var_red_counter::var_red_counter(
	pargal::pcvar &var, const std::size_t odim):
	begin_(pargal::util::var_red_counter_iterator(chunk_counters_.begin(),
		chunk_counters_.end())),
	end_(pargal::util::var_red_counter_iterator(chunk_counters_.begin(),
		chunk_counters_.end()))
{
    std::vector<std::size_t> var_ldims;
	var.get_local_dims(var_ldims);
	std::size_t odim_tc_to_counter_ratio = 1;

    // FIXME: Can we initialize counter size to 1 instead of 0?
    // FIXME: Abstract the logic of finding the size - used when creating a dim too
    std::size_t counter_sz = 1;
	std::size_t ncounters = 1;
	// FIXME: When assigning the size of a counter there is assumption
	// about the iterator being used - we need to incorporate iteration
	// style when initializing the counter obj
	moab_tag_dim_type odim_type = var.get_dim_type(odim);
    switch(odim_type){
        case MOAB_TAG_SPARSE_DIM:
				{
					counter_sz = 1;
				}
				break;
        case MOAB_TAG_GEN_DIM:
				{
					std::size_t var_lsz = 1;
					counter_sz = var.tstep_size();

                    std::vector<std::size_t>::iterator iter = var_ldims.begin();
                    std::advance(iter, odim + 1);
                    var_lsz = std::accumulate(iter, var_ldims.end(), var_lsz, std::multiplies<std::size_t>());
					ncounters = var_lsz/counter_sz;
					odim_tc_to_counter_ratio = 1;
				}
				break;
        case MOAB_TAG_LEV_DIM:
                {
					counter_sz = var.tstep_size()/var_ldims[odim];
					odim_tc_to_counter_ratio = var_ldims[odim];
                }
                break;
        case MOAB_TAG_SQIJ_LAT_DIM:
                counter_sz = var_ldims[var_ldims.size()-1] * var.num_levels();
                break;
        case MOAB_TAG_SQIJ_LON_DIM:
                counter_sz = var_ldims[var_ldims.size()-2] * var.num_levels();
                break;
        default:
                throw pargal_except("Error creating counter - invalid dim", __FILE__, __LINE__, "Unknown");
    }
	chunk_counters_.resize(ncounters, moab_tag_chunk_counter(counter_sz, 0,
			odim_tc_to_counter_ratio, odim_type));
	begin_ = pargal::util::var_red_counter_iterator(chunk_counters_.begin(), chunk_counters_.end());
	end_ =  pargal::util::var_red_counter_iterator(chunk_counters_.end(), chunk_counters_.end());
}


void pargal::util::var_red_counter::reset(void )
{
	for(std::vector<moab_tag_chunk_counter>::iterator iter = chunk_counters_.begin();
			iter != chunk_counters_.end(); ++iter){
		for(pargal::util::moab_tag_chunk_counter::iterator citer = iter->begin();
				citer != iter->end(); ++citer){
			for(moab_tag_seq_counter::iterator siter = citer->begin();
				siter != citer->end(); ++siter){
				*siter = 0;
			}
		}
	}
}

// iterator
