/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "pargal/moab_tag_dim.hpp"
#include <vector>
#include <algorithm>
#include <functional>
#include <numeric>
#include <cstddef>
#include <pargal/pcvar.hpp>

pargal::util::moab_tag_dim *pargal::util::create_moab_tag_dim(pargal::pcvar &var, const std::size_t dim)
{
	switch(var.get_dim_type(dim)){
		case pargal::util::MOAB_TAG_SPARSE_DIM:
				return (new pargal::util::moab_tag_sparse_dim(var, dim));
		case pargal::util::MOAB_TAG_GEN_DIM:
				return (new pargal::util::moab_tag_gen_dim(var, dim));
		case pargal::util::MOAB_TAG_LEV_DIM:
				return (new pargal::util::moab_tag_lev_dim(var, dim));
		case pargal::util::MOAB_TAG_SQIJ_LAT_DIM:
				return (new pargal::util::moab_tag_sqij_lat_dim(var, dim));
		case pargal::util::MOAB_TAG_SQIJ_LON_DIM:
				return (new pargal::util::moab_tag_sqij_lon_dim(var, dim));
		default:
				throw pargal::pargal_except("Error creating dim, Unknown dim type",
						__FILE__, __LINE__, "Unknown");
	}
}

pargal::util::moab_tag_sparse_dim::moab_tag_sparse_dim(pargal::pcvar &var,
	const std::size_t dim)
{
	std::vector<std::size_t> var_dims;
	skip_cnt_ = 1;
	
	var.get_dims(var_dims);
	std::vector<std::size_t>::iterator iter = var_dims.begin();
	std::vector<std::size_t>::iterator iter_end = var_dims.begin();
	std::advance(iter_end, dim+1);
	skip_cnt_ = std::accumulate(iter, iter_end, skip_cnt_, std::multiplies<std::size_t>());
	skip_dist_ = 1;
	skip_dist_ = std::accumulate(iter_end, var_dims.end(), skip_dist_, std::multiplies<std::size_t>());
	reset_cnt_ = skip_cnt_;
	reset_dist_ = 1;
	static_cnt_ = skip_cnt_;
}

// FIXME: See if there is a way to generalize the clone strategy
// default constructor for dims + set_iter_info
pargal::util::moab_tag_sparse_dim::moab_tag_sparse_dim(
	const std::size_t static_cnt,
	const std::size_t reset_cnt, const std::size_t reset_dist,
	const std::size_t skip_cnt, const std::size_t skip_dist):
	pargal::util::moab_tag_dim::moab_tag_dim(static_cnt,
	reset_cnt, reset_dist, skip_cnt, skip_dist)
{}

pargal::util::moab_tag_dim *pargal::util::moab_tag_sparse_dim::clone(void ) const
{
	return (new pargal::util::moab_tag_sparse_dim(static_cnt_, reset_cnt_,
			reset_dist_, skip_cnt_, skip_dist_));
}

pargal::util::moab_tag_gen_dim::moab_tag_gen_dim(pargal::pcvar &var,
	const std::size_t dim)
{
	static_cnt_ = 0;
	reset_cnt_ = var.tstep_size();
	reset_dist_ = 0;
	skip_cnt_ = 1;
	skip_dist_ = 1;
}

pargal::util::moab_tag_gen_dim::moab_tag_gen_dim(
	const std::size_t static_cnt,
	const std::size_t reset_cnt, const std::size_t reset_dist,
	const std::size_t skip_cnt, const std::size_t skip_dist):
	pargal::util::moab_tag_dim::moab_tag_dim(static_cnt,
	reset_cnt, reset_dist, skip_cnt, skip_dist)
{}

pargal::util::moab_tag_dim *pargal::util::moab_tag_gen_dim::clone(void ) const
{
	return (new pargal::util::moab_tag_gen_dim(static_cnt_, reset_cnt_,
			reset_dist_, skip_cnt_, skip_dist_));
}

pargal::util::moab_tag_lev_dim::moab_tag_lev_dim(pargal::pcvar &var,
	const std::size_t dim)
{
	std::size_t nlevs = var.num_levels();

	static_cnt_ = nlevs;
	skip_cnt_ = static_cnt_;
	skip_dist_ = nlevs;

	// FIXME: We should set cnts that we don't use to an invalid count
	reset_cnt_ = var.tstep_size();
}

pargal::util::moab_tag_lev_dim::moab_tag_lev_dim(
	const std::size_t static_cnt,
	const std::size_t reset_cnt, const std::size_t reset_dist,
	const std::size_t skip_cnt, const std::size_t skip_dist):
	pargal::util::moab_tag_dim::moab_tag_dim(static_cnt,
	reset_cnt, reset_dist, skip_cnt, skip_dist)
{}

pargal::util::moab_tag_dim *pargal::util::moab_tag_lev_dim::clone(void ) const
{
	return (new pargal::util::moab_tag_lev_dim(static_cnt_, reset_cnt_,
			reset_dist_, skip_cnt_, skip_dist_));
}

pargal::util::moab_tag_sqij_lat_dim::moab_tag_sqij_lat_dim(pargal::pcvar &var,
	const std::size_t dim)
{
	std::vector<std::size_t> var_ldims;
	var.get_local_dims(var_ldims);

	std::size_t var_ldims_sz = var_ldims.size();
	if(var_ldims.size() < 2){
		throw pargal::pargal_except("Attempting to use create lat dim for a variable with no lat dim", __FILE__, __LINE__, "Unknown");
	}

	size_t nlevs = var.num_levels();
	reset_cnt_ = var_ldims[var_ldims_sz - 1] * nlevs;
	reset_dist_ = 0;

	// FIXME: We should set cnts that we don't use to an invalid count
	static_cnt_ = 0;
	skip_cnt_ = var.tstep_size();
}

pargal::util::moab_tag_sqij_lat_dim::moab_tag_sqij_lat_dim(
	const std::size_t static_cnt,
	const std::size_t reset_cnt, const std::size_t reset_dist,
	const std::size_t skip_cnt, const std::size_t skip_dist):
	pargal::util::moab_tag_dim::moab_tag_dim(static_cnt,
	reset_cnt, reset_dist, skip_cnt, skip_dist)
{}

pargal::util::moab_tag_dim *pargal::util::moab_tag_sqij_lat_dim::clone(void ) const
{
	return (new pargal::util::moab_tag_sqij_lat_dim(static_cnt_, reset_cnt_,
			reset_dist_, skip_cnt_, skip_dist_));
}

pargal::util::moab_tag_sqij_lon_dim::moab_tag_sqij_lon_dim(pargal::pcvar &var,
	const std::size_t dim)
{
	std::vector<std::size_t> var_ldims;
	var.get_local_dims(var_ldims);

	std::size_t var_ldims_sz = var_ldims.size();
	if(var_ldims.size() < 1){
		throw pargal::pargal_except("Attempting to use create lon dim for a variable with no lon dim", __FILE__, __LINE__, "Unknown");
	}

	std::size_t nlevs = var.num_levels();
	skip_cnt_ = var_ldims[var_ldims_sz - 1] * nlevs;
	skip_dist_ = skip_cnt_;

	reset_cnt_ = nlevs;
	reset_dist_ = 0;
	
	// FIXME: We should set cnts that we don't use to an invalid count
	static_cnt_ = 0;
}

pargal::util::moab_tag_sqij_lon_dim::moab_tag_sqij_lon_dim(
	const std::size_t static_cnt,
	const std::size_t reset_cnt, const std::size_t reset_dist,
	const std::size_t skip_cnt, const std::size_t skip_dist):
	pargal::util::moab_tag_dim::moab_tag_dim(static_cnt,
	reset_cnt, reset_dist, skip_cnt, skip_dist)
{}

pargal::util::moab_tag_dim *pargal::util::moab_tag_sqij_lon_dim::clone(void ) const
{
	return (new pargal::util::moab_tag_sqij_lon_dim(static_cnt_, reset_cnt_,
			reset_dist_, skip_cnt_, skip_dist_));
}

