/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <functional>
#include <numeric>
#include <utility>
#include <vector>
#include <cstddef>
#include <pargal/moab_tag_iterator.hpp>
#include <pargal/pcvar.hpp>
#include <pargal/exception.hpp>
#include <pargal/moab_tag_chunk.hpp>

pargal::util::moab_tag_chunk_iterator::moab_tag_chunk_iterator(
	pargal::pcvar &var, const size_t start_tstep, const size_t odim):pcvar_(var),tstep_(start_tstep), step_offset_(0), odim_(odim), is_tstep_cached_(false), is_sparse_var_(false)
{
	const std::string part_method = var.get_finfo().get_part_method();
	if(((part_method != "SQIJ") && (part_method != "TRIVIAL_PARTITION")) &&
		(var.get_loc() != "SET")){
		throw pargal::pargal_except(std::string("Tag chunk iterator is currently not defined for ") + part_method + std::string(" partition method") , __FILE__, __LINE__, __PRETTY_FUNCTION__);
	}

	// FIXME: Make sure get_tsteps() returns the total number of tsteps in the var
	// i.e., accumulate( [0, time_dim] )
	nsteps_ = var.get_tsteps().size();
	is_chunk_init_.resize(nsteps_, false);
	step_sz_ = var.tstep_size();
	switch(var.get_dim_type(odim_)){
		case MOAB_TAG_SPARSE_DIM:
				{
					// Look into skip_iterator to see why nsteps_
					// is calculated this way
					std::vector<size_t> var_dims;
					var.get_dims(var_dims);
					is_sparse_var_ = true;
					nsteps_before_reduce_ = 1;
					
					// FIXME: Skip iterator logic used here
					nsteps_ = 1;
					std::vector<size_t>::iterator iter = var_dims.begin();
					std::vector<size_t>::iterator iter_end = var_dims.begin();
					std::advance(iter_end, odim);
					
					nsteps_ = std::accumulate(iter, iter_end, nsteps_, std::multiplies<std::size_t>());
					std::advance(iter_end, 1);
					iter = iter_end;
					nsteps_ = std::accumulate(iter, var_dims.end(), nsteps_, std::multiplies<std::size_t>());
				}
				break;
		case MOAB_TAG_GEN_DIM:
				nsteps_before_reduce_ = nsteps_ - 1;
				break;
		case MOAB_TAG_LEV_DIM:
				//continue
		case MOAB_TAG_SQIJ_LAT_DIM:
				//continue
		case MOAB_TAG_SQIJ_LON_DIM:
				// Every step needs a reduction op
				nsteps_before_reduce_ = 0;
				break;
	}
	nsteps_before_reduce_counter_ = nsteps_before_reduce_;
}

pargal::util::moab_tag_chunk_iterator::~moab_tag_chunk_iterator()
{
	if(is_tstep_cached_){
		pcvar_.cache_delete(tstep_);
		is_tstep_cached_ = false;
	}
}

void pargal::util::moab_tag_chunk_iterator::chbegin(void )
{
	// Anything needed here?
}

bool pargal::util::moab_tag_chunk_iterator::chend(void )
{
	return (tstep_ >= nsteps_);
}

pargal::util::moab_tag_chunk pargal::util::moab_tag_chunk_iterator::operator*()
{
	std::list<std::pair<double *, double *> > chunk_seqs;
	double *begin, *end;
	std::size_t seq_sz = 0;
	pargal::pcvar::cache_access_state rd_state;
	bool cont_reading_chunk = false;

	do{
		cont_reading_chunk = pcvar_.cache_read(tstep_, begin, seq_sz, cont_reading_chunk, rd_state);
		end = begin + seq_sz;
		begin += step_offset_;
		chunk_seqs.push_back(std::make_pair(begin, end));
	}while(cont_reading_chunk == true);

	bool needs_reduce = false;
	is_tstep_cached_ = true;
	if(nsteps_before_reduce_counter_ <= 0){
		//reset counter - we need a reduce on this tag chunk
		nsteps_before_reduce_counter_ = nsteps_before_reduce_;
		needs_reduce = true;
	}
	return pargal::util::moab_tag_chunk(chunk_seqs, odim_, pcvar_, needs_reduce, is_chunk_init_[tstep_]);
}

void pargal::util::moab_tag_chunk_iterator::operator++()
{
	// We assume that once the chunk is accessed and we move on
	// it is inited 
	is_chunk_init_[tstep_] = true;
	if(!is_sparse_var_){
		if(is_tstep_cached_){
			pcvar_.cache_delete(tstep_);
			is_tstep_cached_ = false;
		}
		// tstep_ increments till we reach (num of tsteps)
		tstep_++;
	}
	else{
		step_offset_++;
		// ntsteps_ decrements till nsteps_ - 1 == 0 ; tstep always == 0
		nsteps_--;
	}
	nsteps_before_reduce_counter_--;
}

pargal::util::moab_tag_reset_chunk_iterator::moab_tag_reset_chunk_iterator(
	pargal::pcvar &pcvar, const size_t tstep, const size_t odim):
	pargal::util::moab_tag_chunk_iterator(pcvar, tstep, odim)
{}

pargal::util::moab_tag_reset_chunk_iterator::~moab_tag_reset_chunk_iterator()
{
	// Don't delete the reset chunks
	is_tstep_cached_ = false;
}

void pargal::util::moab_tag_reset_chunk_iterator::operator++()
{
	// Note that we only care about when to reduce for reset chunks.
	// Infinite traversal loop
	is_chunk_init_[tstep_] = true;
	tstep_++;
	// Reset iterator if needed
	if(tstep_ >= nsteps_) { tstep_ = 0; }
	nsteps_before_reduce_counter_--;
}
