/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/** \mainpage The Parallel Gridded Analysis Library (ParGAL)
 *
 * The large volume of data currently produced by climate models is
 * overwhelming the current, decades-old visualization workflow. The
 * traditional methods for visualizing climate output also have not
 * kept pace with changes in the types of grids used, the number of
 * variables involved, and the number of different simulations
 * performed with a climate model or the feature-richness of
 * high-resolution simulations.
 *
 * The bottleneck to producing climate model images is not always the
 * drawing of the image on the screen but rather the calculations that
 * must be performed on the climate model output before an image can
 * be made. ParGAL will speed up this post-processing, or analysis,
 * phase of climate visualization and vastly improve the speed of
 * climate data analysis compared to the current serial tools. ParGAL
 * will build on currently available software technology that will
 * also permit calculations to be performed in parallel on many
 * different numerical grids. We will interface ParGAL with popular
 * climate modeling tool the NCAR Command Language.
 *
 * There are four partition methods provided by MOAB for different
 * application requirements, including ALLJORKORI, ALLJKBAL, SQIJ,
 * SQJK. The algorithms implemented inside ParGAL assumes SQIJ
 * partition method is used.
 *
 * Questions and comments should be sent to parvis-dev 
 * _at_ mcs.anl.gov.
 */

#ifndef PARGAL_ALGORITHM_HPP_INCLUDED
#define PARGAL_ALGORITHM_HPP_INCLUDED

#include <pargal/algo_impl.hpp>

namespace pargal 
{
  //! computes the sum of a variable's given dimension at all other dimensions
  /**
     \param var     The variable on which to calculate the sum
     \param var_sum The variable that contains the sum 
     \param dims    The dimensions of the variable on which to calculate the sum. 
                    Must be consecutive and monotonically increasing     
     \param keep    The default value is false, which means to delete var if it is a 
                    file variable. This can be turned off by setting it to true.

     Example: \code
     std::vector<std::size_t> dims(1, 0);
     dim_sum_n(var, var_sum, dims); ... \endcode

     Except the sum over time dimension where var_sum is the size
     of the var divided by number of time steps, var_sum has the same
     size as var when computing sum over other dimensions. The
     result is stored at index 0 for the corresponding dimension.
  */
  void dim_sum_n(pcvar& var, pcvar& var_sum, const std::vector<std::size_t>& dims, bool keep=false);

  //! computes the average of a variable's given dimension at all other dimensions
  /**
     \param var     The variable on which to calculate the average
     \param var_avg The variable which contains the average 
     \param dims    The dimensions of the variable on which to calculate the average. 
                    Must be consecutive and monotonically increasing     
     \param keep    The default value is false, which means to delete var if it is a 
                    file variable. This can be turned off by setting it to true.

     Example: \code
     std::vector<std::size_t> dims(1, 0);
     dim_avg_n(var, var_avg, dims); ... \endcode

     Except the average over time dimension where var_avg is the size
     of the var divided by number of time steps, var_avg has the same
     size as var when computing average over other dimensions. The
     result is stored at index 0 for the corresponding dimension.
  */
  void dim_avg_n(pcvar& var, pcvar& var_avg, const std::vector<std::size_t>& dims, bool keep=false);

  //! computes the product of a variable's given dimension at all other dimensions
  /**
     \param var     The variable on which to calculate the product
     \param var_prod The variable which contains the product 
     \param dims    The dimensions of the variable on which to calculate the product. 
                    Must be consecutive and monotonically increasing     
     \param keep    The default value is false, which means to delete var if it is a 
                    file variable. This can be turned off by setting it to true.

     Example: \code
     std::vector<std::size_t> dims(1, 0);
     dim_product_n(var, var_prod, dims); ... \endcode

     Except the prodcut over time dimension where var_avg is the size
     of the var divided by number of time steps, var_prod has the same
     size as var when computing product over other dimensions. The
     result is stored at index 0 for the corresponding dimension.
  */
  void dim_product_n(pcvar& var, pcvar& var_prod, const std::vector<std::size_t>& dims, bool keep=false);

  //! computes the maximum of a variable's given dimension at all other dimensions
  /**
     \param var     The variable on which to calculate the max
     \param var_max The variable which contains the max 
     \param dims    The dimensions of the variable on which to calculate the maximum. 
                    Must be consecutive and monotonically increasing     
     \param keep    The default value is false, which means to delete var if it is a 
                    file variable. This can be turned off by setting it to true.

     Example: \code
     std::vector<std::size_t> dims(1, 0);
     dim_max_n(var, var_max, dims); ... \endcode
     
     Except the max over time dimension where var_max is the size of
     the var divided by number of time steps, var_max has the same
     size as var when computing max over other dimensions. The result
     is stored at index 0 for the corresponding dimension.
   */
  void dim_max_n(pcvar& var, pcvar& var_max, const std::vector<std::size_t>& dims, bool keep=false);

  //! computes the minimum of a variable's given dimension at all other dimensions
  /**
     \param var     The variable on which to calculate the min
     \param var_min The variable which contains the min 
     \param dims    The dimensions of the variable on which to calculate the minimum. 
                    Must be consecutive and monotonically increasing     
     \param keep    The default value is false, which means to delete var if it is a 
                    file variable. This can be turned off by setting it to true.
     
     Example: \code
     std::vector<std::size_t> dims(1, 0);
     dim_min_n(var, var_min, dims); ... \endcode
     
     Except the min over time dimension where var_min is the size of
     the var divided by number of time steps, var_min has the same
     size as var when computing min over other dimensions. The result
     is stored at index 0 for the corresponding dimension.
   */
  void dim_min_n(pcvar& var, pcvar& var_min, const std::vector<std::size_t>& dims, bool keep=false);

  //! return the maximum element of a variable  
  //! \param keep   The default value is false, which means to delete var if it is a 
  //!               file variable. This can be turned off by setting it to true.
  double max(pcvar& var, bool keep=false);

  //! return the minimum element of a variable  
  //! \param keep   The default value is false, which means to delete var if it is a 
  //!               file variable. This can be turned off by setting it to true.
  double min(pcvar& var, bool keep=false);

  /*
    \brief  
            This algorithm calculates vorticity from a velocity field on a
            rectilinear lat/lon grid.  MOAB is used to read a netcdf file 
            containing the grid and nodal velocity data and Intrepid is
            used to calculate the partial derivatives assuming a bilinear
            approximation of velocity on a grid cell.

    \verbatim

      Gradient of scalar quantity u:
       
               grad u(x) = \sum_i u_i grad \phi_i(x)
   
      where u_i are the values of u at the nodes and \phi_i are the basis
      functions for each node i.

      For a velocity field (u,v) on a spherical grid (lambda,phi), vorticity 
      is calculated from elements of (grad u) and (grad v) as:

               vorticity = 1/(R cos(phi)) dv/dlambda - 1/R du/dphi + u/R tan(phi)
       
    \endverbatim

  */
  void vorticity(pcvar& var_u, pcvar& var_v, pcvar& var_vort);

  /*    
    \brief  
            This algorithm calculates divergence from a velocity field on a
            rectilinear lat/lon grid.  MOAB is used to read a netcdf file 
            containing the grid and nodal velocity data and Intrepid is
            used to calculate the partial derivatives assuming a bilinear
            approximation of velocity on a grid cell.

    \verbatim

      Gradient of scalar quantity u:
       
               grad u(x) = \sum_i u_i grad \phi_i(x)
   
      where u_i are the values of u at the nodes and \phi_i are the basis
      functions for each node i.

      For a velocity field (u,v) on a spherical grid (lambda,phi), divergence
      is calculated from elements of (grad u) and (grad v) as:

               divergence = 1/(R cos(phi)) du/dlambda + 1/R dv/dphi - v/R tan(phi)
       
    \endverbatim

  */
  void divergence(pcvar& var_u, pcvar& var_v, pcvar& var_div);  

  //! gather the value of a variable to root 0
  /**
     \param var      The variable on which to gather
     \param var_data The storage of the result
     \param keep    The default value is false, which means to delete var if it is a 
                    file variable. This can be turned off by setting it to true.
     
     Example: \code
     dim_avg_n(var, var_avg, 0); 
     std::vector<double> var_data;
     gather(var_avg, var_data); ... \endcode

     1. The vector will be resized to the size of each time step times
           the number of time steps to be gathered of the variable.
           3. Only Datatype supported for gather and pcvar is double
           for now.  4. For the pcvar created by the user, the user
           has to make sure the time steps to be gathered exsit in the
           mesh (which means data is in memory). For the pcvar created
           based on single or multiple files, the data will be loaded
           into memory if it is not loaded already. The data for a
           pcvar is gathered in a time step fashion, and each time
           step will be deleted immediately in order to save memeory.
  */
  template <typename T>
  void gather(const pcvar& var, std::vector<T>& var_data, bool keep=false);  
  
  //! gather the value of a variable for selected time steps to root 0
  /**
     \param var      The variable on which to gather
     \param tsteps   The time steps to be gathered
     \param var_data The storage of the result
     \param keep    The default value is false, which means to delete var if it is a 
                    file variable. This can be turned off by setting it to true.
     
     Example: \code
     dim_avg_n(var, var_avg, 0); 
     std::vector<double> var_data;
     std::vector<int> avg_tsteps(1, 0); 
     gather(var_avg, avg_tsteps, var_data); ... \endcode

     The note for gathering the entire pcvar also applies here.
  */
  template <typename T>
  void gather(const pcvar& var, const std::vector<int>& tsteps, std::vector<T>& var_data, bool keep=false);  

} // namespace pargal

#endif
