/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_FILE_INFO_HPP_INCLUDED
#define PARGAL_FILE_INFO_HPP_INCLUDED

#include <moab/Interface.hpp>
#include <moab/ParallelComm.hpp>
#include <pargal/pcdim.hpp>
#include <pargal/pcatt.hpp>
#include <pargal/fvar.hpp>
#include <map>
#include <set>
#include <string>
#include <vector>

namespace pargal 
{  
  class fileinfo 
  {    
  public:
    ///////////////////////////////////////////
    // constructor destructor
    ///////////////////////////////////////////
    
    //! constructor
    explicit fileinfo(const std::vector<std::string>& file_names,
		      moab::Interface& mb, 
		      const moab::ParallelComm& mb_pcomm,
		      moab::EntityHandle& file_set,
		      const std::string& part_method);
    
    //! destructor
    ~fileinfo();

    ///////////////////////////////////////////
    // general queries
    ///////////////////////////////////////////
    
    //! return the number of dimensions in the netcdf file
    std::size_t num_dims() const;
    
    //! return the number of variables in the netcdf file
    std::size_t num_vars() const;
    
    //! return the number of global attributes in the netcdf file    
    std::size_t num_atts() const;
    
    //! return the dimension names in the netcdf file
    std::set<std::string> dim_names() const;
    
    //! return the variable names in the netcdf file
    std::set<std::string> var_names() const;
    
    //! return the global attribute names in the netcdf file
    std::set<std::string> att_names() const;
    
	//! Returns true if a dimension by name dim_name exists
	//! false otherwise
	bool has_dim(const std::string& dim_name) const;
    
    //! return a dimension by name
    //! Note: if it is a time dimension, the size returned is the
    //!       size of time dimension within a single file times the 
    //!       number of files read, i.e., assume each file has the
    //!       same size of time dimension.
    const pcdim& get_dim(const std::string& dim_name) const;
    
    //! return the global attribute
    const pcatt& get_att() const;
    
    //! return the file names
    const std::vector<std::string>& file_names() const;
    
    ///////////////////////////////////////////
    // variable related queries
    ///////////////////////////////////////////
    
    //! return the type of the variable
    moab::DataType get_var_type(const std::string& name) const;
    
    //! return the number of dimensions of the variable
    std::size_t get_var_num_dims(const std::string& name) const;
    
    //! return a reference to the index-th dimension (starting at 0)
    const pcdim& get_var_dim(const std::string& name, const std::size_t index) const;
    
    //! return a vector of dimensions
    std::vector<pcdim> get_var_dims(const std::string& name) const;
    
    //! return the number of attributes of the variable
    std::size_t get_var_num_atts(const std::string& name) const;
    
    //! return the attribute by name
    const pcatt& get_var_att(const std::string& varname) const;

    //! return the location information of a variable
    const std::string& get_var_location(const std::string& name) const;

    ///////////////////////////////////////////
    // moab queries
    ///////////////////////////////////////////

    //! return the moab instance
    moab::Interface& get_moab();

    //! return the moab ParallelComm instance
    const moab::ParallelComm& get_pcomm() const;

    //! return the file_set
    moab::EntityHandle& get_file_set();

    //! return the partition method
    const std::string& get_part_method() const;

    ///////////////////////////////////////////
    // other queries
    ///////////////////////////////////////////
    //
	//! return MPI communicator for ops associated with the files
	//! represented by this file info
	MPI_Comm get_mpi_comm(void ) const;
    
    //! return the file name containing the specified timestep 
    //! Note: current implementation assmues each file contain same
    //! number of time steps
    const std::string& tstep_lookup(int tstep) const;

    //! return the grid_type
    //! /CAM_EUL/CAM_FV/CAM_SE/CAM_UNKNOWN/NOT_CAM
    const std::string& get_grid_type() const;

  private:    
    ///////////////////////////////////////////
    // modifiers
    ///////////////////////////////////////////
    
    //! insert a variable
    void insert(const std::string& var_name, const fvar& var);
    
    //! insert a dimension
    void insert(const std::string& dim_name, const pcdim& dim);

    //! insert a attribute
    void insert(const std::string& att_name, const pcatt& att);
    
    ///////////////////////////////////////////
    // initialization
    ///////////////////////////////////////////
    
    //! init 
    void init();
    
    //! init dimensions
    void init_dims();

    //! init variables
    void init_vars();

    //! init attributes
    void init_atts();

    //! init grid type
    void init_grid_type();

  private:    
    ///////////////////////////////////////////
    // data members
    ///////////////////////////////////////////
    
    //! variables
    std::map<std::string, const fvar*> m_vars;
    
    //! dimensions
    std::map<std::string, const pcdim*> m_dims;
    
    //! global and variable attributes
    std::map<std::string, const pcatt*> m_atts;
    
    //! files 
    std::vector<std::string> m_file_names;
    
    //! moab instance
    moab::Interface& m_mb;

    //! moab ParallelComm instance
    const moab::ParallelComm& m_mb_pcomm;

	//! Dup of MOAB mpi comm for ParGAL MPI ops
	MPI_Comm mpi_comm_;

    //! moab file set
    moab::EntityHandle& m_file_set;
      
    //! partition method
    std::string m_part_method;

    //! grid type
    std::string m_grid_type;
  };    
} // namespace pargal

#endif
