/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_FILE_VAR_HPP_INCLUDED
#define PARGAL_FILE_VAR_HPP_INCLUDED

#include <pargal/pcdim.hpp>
#include <pargal/pcatt.hpp>
#include <moab/Types.hpp>
#include <string>
#include <vector>
#include <map>

namespace pargal 
{
  class fvar
  {
  public:
    ///////////////////////////////////////////
    // queries
    ///////////////////////////////////////////
    
    //! return the name of the variable
    inline const std::string& name() const;
    
    //! return the type of the variable
    inline moab::DataType type() const;
    
    //! return the number of dimensions of the variable
    inline std::size_t num_dims() const;
    
    //! return a reference to the index-th dimension (starting at 0)
    inline const pcdim& get_dim(const std::size_t index) const;
    
    //! return a vector of dimension pointers
    inline const std::vector<const pcdim*>& get_dims() const;
    
    //! return the number of attributes of the variable
    inline std::size_t num_atts() const;
    
    //! return the attribute names of the variable
    inline std::vector<std::string> att_names() const;
    
    //! return the attribute by name
    inline const pcatt& get_att(const std::string& name) const;

    inline const std::string& location() const;
    
  private:
    ///////////////////////////////////////////
    // constructor destructor
    ///////////////////////////////////////////
    
    //constructor  
    inline explicit fvar(const std::string& name, 
			 moab::DataType type, 
			 const std::vector<const pcdim*>& dims,
			 const std::string& location
			 );			    
    
    //! destructor
    inline ~fvar();
    
    ///////////////////////////////////////////
    // data member
    ///////////////////////////////////////////
    
    std::string m_name;
    moab::DataType m_type;
    std::vector<const pcdim*> m_dims;
    std::map<std::string, pcatt*> m_atts;
    std::string m_location;
    friend class fileinfo;
  };
  
  ///////////////////////////////////////////
  // implementation detail
  ///////////////////////////////////////////
        
  //constructor  
  fvar::fvar(const std::string& name, 
	     moab::DataType type, 
	     const std::vector<const pcdim*>& dims,
	     const std::string& location
	     ) :			    
    m_name(name),
    m_type(type),
    m_dims(dims),
    m_location(location)
  {}
  
  fvar::~fvar()
  {}
  
  //! return the name of the variable  
  const std::string& 
  fvar::name() const
  {
    return m_name;
  }
    
  //! return the type of the variable  
  moab::DataType 
  fvar::type() const
  {
    return m_type;
  }
  
  //! return the number of dimensions of the variable  
  std::size_t 
  fvar::num_dims() const
  {
    return m_dims.size();
  }
  
  //! return a reference to the index-th dimension (starting at 0)   
  const pcdim& 
  fvar::get_dim(const std::size_t index) const
  {
    return *(m_dims[index]);
  }

  //! return a vector of dimension pointers
  const std::vector<const pcdim*>& 
  fvar::get_dims() const
  {
    return m_dims;
  }
  
  //! return the number of attributes of the variable  
  std::size_t 
  fvar::num_atts() const
  {
    return 0;
  }
  
  //! return the attribute names of the variable  
  std::vector<std::string> 
  fvar::att_names() const
  {
    return std::vector<std::string>();
  }
  
  //! return the attribute by name
  const pcatt& 
  fvar::get_att(const std::string& name) const
  {
    return *(const_cast<fvar*>(this)->m_atts[name]);
  }        

  //! return the location of the variable  
  const std::string& 
  fvar::location() const
  {
    return m_location;
  }

} // namespace pargal

#endif
