/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_GATHER_IMPL_HPP_INCLUDED
#define PARGAL_GATHER_IMPL_HPP_INCLUDED

#include <pargal/mesh.hpp>
#include <pargal/pcvar.hpp>
#include <numeric>

namespace pargal 
{    
  //! gather algorithm for alljorkori partition
  template<typename T>
  void 
  gather_partition_alljorkori(const std::size_t& lon, 
			      const std::size_t& lat, 
			      const std::size_t& lev, 
			      const std::size_t& num_procs,
			      const std::vector<int>& recvcnts, 
			      std::vector<T>& var_data,
			      const std::size_t& pos,
			      const std::vector<int>& displs, 
			      const std::vector<T>& var_data_tmp);
  
  //! gather the local value of a variable
  template <typename T>
  void 
  get_local_varname_val(const moab::Interface& mb,
			const moab::ParallelComm& mb_pcomm, 
			moab::EntityHandle& file_set,
			const std::string& var_name, 
			std::vector<T>& var_data,
			const std::size_t& pos,
			const std::size_t& lev,
			const std::size_t& loc_lat,
			const std::size_t& loc_lon);

  ///////////////////////////////////////////
  // template implementation detail
  ///////////////////////////////////////////
  
  template <typename T>
  void 
  gather(const pcvar& var, std::vector<T>& var_data, bool keep=false)
  {
    moab::Interface& mb = const_cast<pcvar&>(var).get_finfo().get_moab();
    moab::EntityHandle& file_set = const_cast<pcvar&>(var).get_finfo().get_file_set();    
    const moab::ParallelComm& mb_pcomm = var.get_finfo().get_pcomm();
    bool is_file_var = !var.is_usr_created();
    int nvals = 0;
    if(var.get_loc() == "QUAD"){
      nvals = get_num_quads(mb, file_set);
    }
    else if(var.get_loc() == "POLYGON"){
      nvals = get_num_polygons(mb, file_set);
    }
    else if(var.get_loc() == "VERTEX"){
      nvals = get_num_vertices(mb, file_set);
	}
    else if(var.get_loc() == "EDGE"){
      nvals = get_num_edges(mb, file_set);
	}
    else{
      throw pargal_except("Error: gather for other location not implemented yet" + var.name(),
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
	}
    int recvcnts_step = nvals * var.get_dim(1).size();
    const std::vector<int>& tsteps = var.get_tsteps();
    int recvcnts_total = recvcnts_step * tsteps.size();
    int rank = mb_pcomm.rank();
    if (rank == 0)
      var_data.resize(recvcnts_total);
    for (std::size_t i = 0; i != tsteps.size(); ++i)
      {	    
	const_cast<pcvar&>(var).get_storage(i);
	if ((var.get_finfo().file_names().size() != 1) && (!var.is_usr_created()))
	  gather_tstep(var, 0, var_data, recvcnts_step*i);
	else
	  gather_tstep(var, tsteps[i], var_data, recvcnts_step*i);
	if (!keep && is_file_var) const_cast<pcvar&>(var).delete_storage(i);
      }
  }

  template <typename T>
  void 
  gather(const pcvar& var, 
	 const std::vector<int>& tsteps,
	 std::vector<T>& var_data, bool keep=false)
  {
    moab::Interface& mb = const_cast<pcvar&>(var).get_finfo().get_moab();
    moab::EntityHandle& file_set = const_cast<pcvar&>(var).get_finfo().get_file_set();    
    const moab::ParallelComm& mb_pcomm = var.get_finfo().get_pcomm();
    bool is_file_var = !var.is_usr_created();
    int nvals = 0;
    if (var.get_loc() == "QUAD"){
      nvals = get_num_quads(mb, file_set);
    }
    else if(var.get_loc() == "POLYGON"){
      nvals = get_num_polygons(mb, file_set);
    }
    else if(var.get_loc() == "VERTEX"){
      nvals = get_num_vertices(mb, file_set);
	}
    else if(var.get_loc() == "EDGE"){
      nvals = get_num_edges(mb, file_set);
	}
    else{
      throw pargal_except("Error: gather for other location not implemented yet" + var.name(),
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
	}
    int recvcnts_step = nvals * var.num_levels();
    int recvcnts_total = recvcnts_step * tsteps.size();
    int rank = mb_pcomm.rank();
    if (rank == 0)
      var_data.resize(recvcnts_total);
    for (std::size_t i = 0; i != tsteps.size(); ++i)
      {	    
	int idx = var.get_index(tsteps[i]);
	if (idx == -1)
	  throw pargal_except("Error: gather non-existent time steps from pcvar " + var.name(),
			      __FILE__, __LINE__, __PRETTY_FUNCTION__);

	const_cast<pcvar&>(var).get_storage(idx);
	if ((var.get_finfo().file_names().size() != 1) && (!var.is_usr_created()))
	  gather_tstep(var, 0, var_data, recvcnts_step*i);
	else
	  gather_tstep(var, tsteps[i], var_data, recvcnts_step*i);
	if (!keep && is_file_var) const_cast<pcvar&>(var).delete_storage(idx);
      }
  }

  template <typename T>
  void 
  gather_tstep(const pcvar& var, 
	       const int& tstep,
	       std::vector<T>& var_data,
	       const std::size_t& pos)
  {    
	std::string grid_type = var.get_finfo().get_grid_type();
	if((grid_type == "CAM_SE") ||
		(grid_type == "MPAS")){
		gather_tstep_ucd(var, tstep, var_data, pos);
	}
	else{
		gather_tstep_scd(var, tstep, var_data, pos);
	}
  }

// Returns true if data needs to be converted from MOAB 2.5D to
// 3D
// Moab does 3D-->2.5D conversion when the level dimension
// is not the fastest changing dimension
// FIXME: This info could be added in the pcvar constructor
inline bool needs_conv_from_mb_rep(const pcvar &var)
{
	const std::vector<pcdim>& var_dims = var.get_dims();
	int lev_idx = -1, cur_idx=0;
	for(std::vector<pcdim>::const_reverse_iterator riter = var_dims.rbegin();
		riter != var_dims.rend();
		++riter, cur_idx++){
		std::string dim_name = riter->name();
		if((dim_name == "lev") ||
			(dim_name == "ilev") ||
			(dim_name == "nVertLevels")){
			lev_idx = cur_idx;
		}
	}

	if((lev_idx != -1) && (lev_idx != 0)){
		return true;
	}
	else{
		return false;
	}
}
  
  template <typename T>
  void 
  gather_tstep_ucd(const pcvar& var, 
		   const int& tstep,
		   std::vector<T>& var_data,
		   const std::size_t& pos)
  {    
    moab::Interface& mb = const_cast<pcvar&>(var).get_finfo().get_moab();
    moab::EntityHandle& file_set = const_cast<pcvar&>(var).get_finfo().get_file_set();    
    const moab::ParallelComm& mb_pcomm = var.get_finfo().get_pcomm();
    moab::ErrorCode rval;
    moab::EntityHandle gset=0;
    if (mb_pcomm.rank() == 0) {
      moab::Tag gathersettag;
      rval = mb.tag_get_handle("GATHER_SET", 1, moab::MB_TYPE_INTEGER, gathersettag, moab::MB_TAG_SPARSE);
      if (rval != moab::MB_SUCCESS)     
	throw pargal_except("Error: Couldn't get gather set tag.",
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);     
      moab::Range gsets;
      int gatherval = 1;
      void *dim[] = {&gatherval};
      rval = mb.get_entities_by_type_and_tag(0, moab::MBENTITYSET, &gathersettag, dim, 1, gsets);
      if (rval != moab::MB_SUCCESS)     
	throw pargal_except("Error: Couldn't get gather set.",
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);     
      assert((int)gsets.size()==1);
      gset = gsets[0];    
    }

    moab::Range ents_owned;
	if(var.get_loc() == "VERTEX"){
		ents_owned = get_local_entities_owned<moab::MBVERTEX>(mb, file_set);    
	}
	else if (var.get_loc() == "POLYGON"){
		ents_owned = get_local_entities_owned<moab::MBPOLYGON>(mb, file_set);    
	}
	else if (var.get_loc() == "EDGE"){
		ents_owned = get_local_entities_owned<moab::MBEDGE>(mb, file_set);    
	}

    std::stringstream var_name;
    var_name << var.name();
    var_name << tstep;    
    moab::Tag tagh;
    int num_lev = var.num_levels();
    rval = mb.tag_get_handle(var_name.str().c_str(), num_lev, moab::MB_TYPE_DOUBLE, tagh, moab::MB_TAG_DENSE);
    moab::Tag idTag;
    rval = mb.tag_get_handle(GLOBAL_ID_TAG_NAME, 1, 
			     moab::MB_TYPE_INTEGER, idTag, moab::MB_TAG_DENSE);    
    const_cast<moab::ParallelComm&>(mb_pcomm).gather_data(ents_owned, tagh, idTag, gset);    
  
    if (mb_pcomm.rank() == 0) {
		moab::Range ents;
		rval = moab::MB_FAILURE;
		if(var.get_loc() == "VERTEX"){
			rval = mb.get_entities_by_type(gset, MBVERTEX, ents);
		}
		else if(var.get_loc() == "POLYGON"){
			rval = mb.get_entities_by_type(gset, MBPOLYGON, ents);
		}
		else if(var.get_loc() == "EDGE"){
			rval = mb.get_entities_by_type(gset, MBEDGE, ents);
		}
		if(rval != moab::MB_SUCCESS){
			throw pargal_except("Error: Unable to get entities for the gather set",
					__FILE__, __LINE__, __PRETTY_FUNCTION__);
		}

		void *data;
		int count = 0;
		moab::Range::iterator mb_iter = ents.begin();
		moab::Range::iterator mb_end = ents.end();
		// Collecting mb_seqs is a better than gathering data in each sequence
		// since this approach does not make assumptions about the var dim
		// as long as the rep of the gathered data is 2.5D
		// FIXME: Check out the performance of the other method - copying
		// each sequence one by one, instead of gathering and then copying
		std::vector<std::pair<T *, int> > mb_seqs;
		do{
			rval = mb.tag_iterate(tagh, mb_iter, mb_end, count, data);
			if(rval != moab::MB_SUCCESS){
				throw pargal_except("Error: Unable to iterate over gathered data",
						__FILE__, __LINE__, __PRETTY_FUNCTION__);
			}
			mb_iter += count;
			T *db_data = static_cast<T *>(data);
			mb_seqs.push_back(std::pair<T *, int>(db_data, count));
		}while(mb_iter != mb_end);

		std::size_t idx = const_cast<std::size_t&>(pos);
		bool rearrange_data = needs_conv_from_mb_rep(var);
		if(rearrange_data){
			for(int l=0; l<num_lev; l++){
				for(typename std::vector<std::pair<T *, int> >::iterator iter 
						= mb_seqs.begin();
						iter != mb_seqs.end();
						++iter){
					T *db_data = (*iter).first;
					int cnt = (*iter).second;
					assert(cnt <= var_data.size() - idx);
					for(int i=0; i<cnt; i++){
						var_data[idx++] = db_data[i*num_lev+l];
					}
				}
			}
		}
		else{
			for(typename std::vector<std::pair<T *, int> >::iterator iter
					= mb_seqs.begin();
					iter != mb_seqs.end();
					++iter){
				T *db_data = (*iter).first;
				int cnt = (*iter).second;
				assert((cnt * num_lev) <= (var_data.size() - idx));
				memcpy(&var_data[idx], db_data, cnt * num_lev * sizeof(T));	
				idx += (cnt * num_lev);
			}
		}
    }
  }

  //! gather the value of a variable at specific time step 
  template <typename T>
  void 
  gather_tstep_scd(const pcvar& var, 
		   const int& tstep,
		   std::vector<T>& var_data,
		   const std::size_t& pos)
  {
    if (var.get_loc() != "QUAD")
      throw pargal_except("Error: gather for locations other than QUAD are not implemented yet!", 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);    
    moab::Interface& mb = const_cast<pcvar&>(var).get_finfo().get_moab();
    moab::EntityHandle& file_set = const_cast<pcvar&>(var).get_finfo().get_file_set();    
    const moab::ParallelComm& mb_pcomm = var.get_finfo().get_pcomm();
    std::size_t num_procs = mb_pcomm.size();
    std::stringstream var_name;
    var_name << var.name();
    var_name << tstep;    
    std::size_t lev = var.num_levels();
    const std::vector<int> global_vertex_minmax = var.get_dim_vertex_global_minmax();
    std::size_t lat = global_vertex_minmax[3] - global_vertex_minmax[2] + 1;
    std::size_t lon = global_vertex_minmax[1] - global_vertex_minmax[0] + 1;
    const std::vector<int> global_dim_minmax = var.get_dim_global_minmax();    
    std::size_t clat = global_dim_minmax[3] - global_dim_minmax[2] + 1;
    std::size_t clon = global_dim_minmax[1] - global_dim_minmax[0] + 1;    
    if (num_procs == 1) {
      get_local_varname_val(mb, mb_pcomm, file_set, var_name.str(), var_data, pos, lev, clat, clon);      
    }
    else
      {
	std::vector<int> loc_minmax = var.get_dim_loc_minmax_owned();
	std::size_t loc_lat = static_cast<std::size_t>(loc_minmax[3] - loc_minmax[2]+1);
	std::size_t loc_lon = static_cast<std::size_t>(loc_minmax[1] - loc_minmax[0]+1);
	int sendcnt = get_local_num_quads(mb, file_set) * lev;
	int mpi_errno;
	MPI_Comm mpicomm = mb_pcomm.comm();
	int rank = mb_pcomm.rank();
	std::vector<int> displs;
	std::vector<int> recvcnts;
	if (rank == 0) 
	  {
	    displs.resize(num_procs, 0);
	    recvcnts.resize(num_procs, 0);
	  }
	mpi_errno = MPI_Gather(&sendcnt, 1, MPI_INT, 
			       &displs[0], 1, MPI_INT, 0, mpicomm);
	if (mpi_errno != MPI_SUCCESS) {
	  throw pargal_except("Error: Failed to gather num local elements per step",
			      __FILE__, __LINE__, __PRETTY_FUNCTION__); 
	}
	std::vector<T> var_data_tmp;
	int recvcnts_total = get_num_quads(mb, file_set) * lev;
	if (rank == 0) 
	  {
	    std::copy(displs.begin(), displs.end(), recvcnts.begin());
	    std::partial_sum(displs.begin(), displs.end(), displs.begin());
	    if (recvcnts_total != displs[num_procs-1])
	      throw pargal_except("Error: Incorrect receive counts per step",
				  __FILE__, __LINE__, __PRETTY_FUNCTION__);
	    std::copy_backward(displs.begin(), --displs.end(), displs.end());
	    displs[0] = 0;
	    var_data_tmp.resize(recvcnts_total);
	  }
	std::vector<T> sendbuf(sendcnt);
	get_local_varname_val(mb, mb_pcomm, file_set, var_name.str(), sendbuf, 0, lev, loc_lat, loc_lon);
	mpi_errno = MPI_Gatherv(&sendbuf[0], sendcnt, MPI_DOUBLE,
				&var_data_tmp[0], &recvcnts[0], &displs[0], MPI_DOUBLE,
				0, mpicomm);
	if (mpi_errno != MPI_SUCCESS) 
	  throw pargal_except("Error: Failed to gather elements",
			      __FILE__, __LINE__, __PRETTY_FUNCTION__); 
	if (rank == 0) 
	  {    
	    const std::string& part_method(const_cast<pcvar&>(var).get_finfo().get_part_method());
	    if (part_method == "SQIJ")
	      gather_partition_sqij(lon, lat, clon, clat, lev, num_procs, var_data, pos, displs, var_data_tmp);
	    else if (part_method == "ALLJORKORI")
	      gather_partition_alljorkori(lon, lat, lev, num_procs, recvcnts, var_data, pos, displs, var_data_tmp);
	    else if (part_method == "ALLJKBAL")
	      throw pargal_except("Error: Gather method for ALLJKBAL not implemented yet",
				  __FILE__, __LINE__, __PRETTY_FUNCTION__); 	
	    else if (part_method == "SQJK")
	      throw pargal_except("Error: Gather method for SQJK not implemented yet",
				  __FILE__, __LINE__, __PRETTY_FUNCTION__); 	
	    else
	      throw pargal_except("Error: Gather method for unrecoganized partition method",
				  __FILE__, __LINE__, __PRETTY_FUNCTION__); 	
	  }	
      } 
  }

  //! gather algorithm for alljorkori partition
  template <typename T>
  void 
  gather_partition_alljorkori(const std::size_t& lon, 
			      const std::size_t& lat, 
			      const std::size_t& lev, 
			      const std::size_t& num_procs,
			      const std::vector<int>& recvcnts, 
			      std::vector<T>& var_data,
			      const std::size_t& pos,
			      const std::vector<int>& displs,
			      const std::vector<T>& var_data_tmp
			      )
  {
    if (lat > num_procs) 
      {
	for (std::size_t i = 0; i != num_procs; ++i) 
	  {
	    int recvcnts_step = recvcnts[i]/lev; 
	    for (std::size_t j = 0; j != lev; ++j) 
	      {
		memcpy(&var_data[pos]+j*lat*lon+displs[i]/lev, 
		       &var_data_tmp[0]+displs[i]+j*recvcnts_step, 
		       sizeof(T)*recvcnts_step);
	      }
	  }
      }
    else 
      {
	throw pargal_except("Error: Gather alljorkori for cases other than lat > num procs haven't implemented yet",
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);
      }    
  }
  
  //! gather algorithm for alljorkori partition
  template <typename T>
  void 
  gather_partition_sqij(const std::size_t& lon, 
			const std::size_t& lat, 
			const std::size_t& clon, 
			const std::size_t& clat, 
			const std::size_t& lev, 
			const std::size_t& np,
			//const std::vector<int>& recvcnts, 
			std::vector<T>& var_data,
			const std::size_t& pos,
			const std::vector<int>& displs,
			const std::vector<T>& var_data_tmp
			)
  {
    std::vector<double> pfactors, ppfactors;
    for (std::size_t i = 2; i <= np; ++i)
      {
	if (!(np%i))
	  {
	    pfactors.push_back(i);
	    ppfactors.push_back(((double)(i*i))/np);
	  }
      }
    double ijratio = ((double) lon)/((double) lat);
    std::size_t ind = std::lower_bound(ppfactors.begin(), ppfactors.end(), ijratio) - ppfactors.begin();
    if (ind && fabs(ppfactors[ind-1]-ijratio) < fabs(ppfactors[ind]-ijratio)) --ind;
    
    int pi = pfactors[ind];
    int pj = np / pi;

    int I = lon-1, J = lat-1;
    int iextra = I%pi, jextra = J%pj, i = I/pi, j = J/pj;
    std::vector<int> vnri(np, 0);
    for (std::size_t nr = 0; nr != np; ++nr)
      vnri[nr] = nr % pi;    
    std::vector<int> vnrj(np, 0);
    for (std::size_t nr = 0; nr != np; ++nr)
      vnrj[nr] = nr / pi;
    for (std::size_t rank = 0; rank != np; ++rank)
      {
	std::size_t li_min = i*vnri[rank] + std::min(iextra, vnri[rank]);
	std::size_t li_max = li_min + i + (vnri[rank] < iextra ? 1 : 0);
	if (vnri[rank] == pi-1) ++li_max;
	std::size_t lj_min = j*vnrj[rank] + std::min(jextra, vnrj[rank]);
	std::size_t lj_max = lj_min + j + (vnrj[rank] < jextra ? 1 : 0);
	std::size_t lk_min = 0;
	std::size_t lk_max = lev;
	for (std::size_t l = lk_min; l != lk_max; ++l)
	  for (std::size_t k = lj_min; k != lj_max; ++k)
	    memcpy(&var_data[pos] + l*clat*clon + k*clon + li_min,
		   &var_data_tmp[0]+displs[rank]+l*(li_max-li_min)*(lj_max-lj_min)+(k-lj_min)*(li_max-li_min),
		   sizeof(T)*(li_max-li_min));
      } // for all proc
  } //gather partition sqij

  template <typename T>
  void 
  get_local_varname_val(const moab::Interface& mb,
			const moab::ParallelComm& mb_pcomm, 
			moab::EntityHandle& file_set,
			const std::string& var_name, 
			std::vector<T>& var_data,
			const std::size_t& pos,
			const std::size_t& lev,
			const std::size_t& loc_lat,
			const std::size_t& loc_lon)
  {
    moab::Range quads;
    moab::Range quads_owned;
    int num_errs = 0;
    num_errs += mb.get_entities_by_type(file_set, moab::MBQUAD, quads);
    num_errs += const_cast<moab::ParallelComm&>(mb_pcomm).filter_pstatus(quads, PSTATUS_NOT_OWNED, PSTATUS_NOT, -1,
					&quads_owned);
    moab::Tag var_name_tag;
    num_errs += mb.tag_get_handle(var_name.c_str(), 0, moab::MB_TYPE_OPAQUE, var_name_tag, moab::MB_TAG_ANY);      
    std::vector<T> tmp_data(lev*loc_lat*loc_lon);
    num_errs += mb.tag_get_data(var_name_tag, quads_owned, &tmp_data[0]);      
    if (num_errs != 0)
      throw pargal_except("Error: Failed to get local varname val",
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    std::size_t tmp_pos = 0;
    for (std::size_t j = 0; j != loc_lat; ++j)     
      for (std::size_t i = 0; i != loc_lon; ++i) 
	for (std::size_t k = 0; k != lev; ++k) 
	  var_data[pos+k*loc_lat*loc_lon+j*loc_lon + i] = tmp_data[tmp_pos++];
  }  
} // namespace pargal

#endif
