/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_MESH_FUNC_HPP_INCLUDED
#define PARGAL_MESH_FUNC_HPP_INCLUDED

#include <moab/Interface.hpp>
#include <moab/ParallelComm.hpp>
#include "MBTagConventions.hpp"
#include "MBEntityType.h"
#include <pargal/exception.hpp>

namespace pargal 
{   
  ///////////////////////////////////////////
  // mesh info
  ///////////////////////////////////////////
  
  //! generic function return the local number of entities
  template <moab::EntityType T>
  std::size_t get_local_num_entities_owned(const moab::Interface& mb,
				     moab::EntityHandle& file_set);
  
  //! generic function return the local range of entities
  template <moab::EntityType T>
  moab::Range get_local_entities_owned(const moab::Interface& mb,
				       moab::EntityHandle& file_set);

  //! generic function return the total number of entities at rank 0
  template<moab::EntityType T>
  std::size_t get_num_entities(const moab::Interface& mb,
			       moab::EntityHandle& file_set);
  
  //! return the number of local vertices    
  std::size_t get_local_num_vertices(const moab::Interface& mb,
				     moab::EntityHandle& file_set);
  
  //! return the total number of vertices at rank 0
  std::size_t get_num_vertices(const moab::Interface& mb,
			       moab::EntityHandle& file_set);
  
  //! return the total number of quads at rank 0
  std::size_t get_num_quads(const moab::Interface& mb,
			    moab::EntityHandle& file_set);
  
  //! return the number of local quads    
  std::size_t get_local_num_quads(const moab::Interface& mb,
				  moab::EntityHandle& file_set);
  
  //! return the total number of polygons at rank 0
  std::size_t get_num_polygons(const moab::Interface& mb,
			    moab::EntityHandle& file_set);
  
  //! return the number of local polygons    
  std::size_t get_local_num_polygons(const moab::Interface& mb,
				  moab::EntityHandle& file_set);
  
  //! return the total number of edges at rank 0
  std::size_t get_num_edges(const moab::Interface& mb,
			    moab::EntityHandle& file_set);
  
  //! return the number of local edges    
  std::size_t get_local_num_edges(const moab::Interface& mb,
				  moab::EntityHandle& file_set);
  
  //! return the dimensions of a variable loaded in the mesh
  std::vector<std::size_t> get_var_dims(const moab::Interface& mb, 
					moab::EntityHandle& file_set,
					const std::string& var_name); 

  //! remove gather set from file set
  void remove_gather_set(const moab::Interface& mb,
				  moab::EntityHandle& file_set);
  
  ///////////////////////////////////////////
  // template implementation detail
  ///////////////////////////////////////////
  
  //! generic function return the local range of entities
  template <moab::EntityType T>
  moab::Range get_local_entities_owned(const moab::Interface& mb,
				       moab::EntityHandle& file_set) 
  {
    moab::Range ents;
    moab::Range ents_owned;    
    mb.get_entities_by_type(file_set, T, ents);
    moab::ParallelComm mb_pcomm(const_cast<moab::Interface*>(&mb), MPI_COMM_WORLD);
    moab::ErrorCode rval = mb_pcomm.filter_pstatus(ents, PSTATUS_NOT_OWNED, PSTATUS_NOT, -1,
						   &ents_owned);
    if (rval != moab::MB_SUCCESS)
      throw pargal_except("Error: " + mb.get_error_string(rval), 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__);
    return ents_owned;
  }

  template <moab::EntityType T>
  std::size_t
  get_local_num_entities_owned(const moab::Interface& mb, 
			 moab::EntityHandle& file_set)
  {
    return get_local_entities_owned<T>(mb, file_set).size();
  }

  //! generic function return the total number of entities at rank 0
  template<moab::EntityType T>
  std::size_t 
  get_num_entities(const moab::Interface& mb,
		   moab::EntityHandle& file_set)
  {
    std::size_t local_num_entities = get_local_num_entities_owned<T>(mb, file_set);
    std::size_t num_entities = 0;
    moab::ParallelComm mb_pcomm(const_cast<moab::Interface*>(&mb), MPI_COMM_WORLD);
    MPI_Comm mpicomm = mb_pcomm.comm();
    MPI_Reduce(&local_num_entities, &num_entities, 1, MPI_UNSIGNED, MPI_SUM, 0, 
	       mpicomm);
    return num_entities;
  }  
} // namespace pargal

#endif
