/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_MOAB_TAG_CHUNK_HPP_INCLUDED
#define PARGAL_MOAB_TAG_CHUNK_HPP_INCLUDED

#include <cstddef>
#include <cassert>
#include <memory>
#include <utility>
#include <list>
#include "moab_tag_iterator.hpp"
#include "moab_tag_dim.hpp"

namespace pargal{
namespace util{

// We assume that MOAB tag values are doubles - if we change this assumption in future
// we can convert the chunk classes into template classes
class moab_tag_chunk_seq : public iterable{
	public:
		typedef double * iterator;
		typedef const double * const_iterator;
		iterator const begin(void ) const;
		const_iterator end(void ) const;
		std::size_t size(void ) const;
		bool has_missing_val(void ) const;
		double get_missing_val(void ) const;
		// FIXME: These counts are currently pertaining to the tag chunks
		// not the tag sequences. Either we need to update these or move
		// static,reset,skip counts/dists to tag chunk class
		std::size_t get_static_count(void ) const;
		std::size_t get_reset_count(void ) const;
		std::size_t get_reset_dist(void ) const;
		std::size_t get_skip_count(void ) const;
		std::size_t get_skip_dist(void ) const;

		moab_tag_chunk_seq(const moab_tag_chunk_seq &other);
		moab_tag_chunk_seq(double *begin, double *end, std::auto_ptr<moab_tag_dim> mb_tag_dim, bool has_missing_val, double missing_val);
	protected:
		iterator const begin_;
		const_iterator end_;
		bool has_missing_val_;
		double missing_val_;
		std::auto_ptr<moab_tag_dim> mb_tag_dim_;
};

class moab_tag_chunk{
	public:
		typedef std::list<moab_tag_chunk_seq>::iterator iterator;
		typedef std::list<moab_tag_chunk_seq>::const_iterator const_iterator;
		iterator begin(void );
		const_iterator end(void ) const;
		std::size_t size(void ) const;
		typedef double * key_type;
		key_type get_key(void ) const;
		bool needs_reduce(void ) const;
		bool is_init(void ) const;
		bool has_missing_val(void ) const;
		double get_missing_val(void ) const;
		class key_compare_functor{
			public:
				bool operator()(const key_type k1, const key_type k2) const;
		};

		moab_tag_chunk(std::list<std::pair<double *, double *> >& chunk_seqs, const std::size_t odim, pargal::pcvar &var, bool needs_reduce, bool is_init);
	private:
		std::size_t sz_;
		std::size_t odim_;
		std::list<moab_tag_chunk_seq> chunk_seqs_;
		bool has_missing_val_;
		double missing_val_;
		bool needs_reduce_;
		bool is_init_;
};

// Moab tag chunk seq utility funcs
inline std::size_t distance(const moab_tag_chunk_seq::iterator begin,
    const moab_tag_chunk_seq::const_iterator end)
{
    return (end - begin);
}

// Moab tag chunk seq member funcs
inline moab_tag_chunk_seq::moab_tag_chunk_seq(double *begin, double *end, std::auto_ptr<moab_tag_dim> mb_tag_dim, bool has_missing_val, double missing_val):begin_(begin), end_(end), mb_tag_dim_(mb_tag_dim), has_missing_val_(has_missing_val), missing_val_(missing_val)
{
}

inline moab_tag_chunk_seq::iterator const moab_tag_chunk_seq::begin(void ) const
{
	return begin_;
}

inline moab_tag_chunk_seq::const_iterator moab_tag_chunk_seq::end(void ) const
{
	return end_;
}

inline std::size_t moab_tag_chunk_seq::size(void ) const
{
	// FIXME: Use difference?
	return (end_ - begin_);
}

inline bool moab_tag_chunk_seq::has_missing_val(void ) const
{
	return has_missing_val_;
}

inline double moab_tag_chunk_seq::get_missing_val(void ) const
{
	return missing_val_;
}

inline std::size_t moab_tag_chunk_seq::get_static_count(void ) const
{
	return mb_tag_dim_->get_static_count();
}

inline std::size_t moab_tag_chunk_seq::get_reset_count(void ) const
{
	return mb_tag_dim_->get_reset_count();
}

inline std::size_t moab_tag_chunk_seq::get_reset_dist(void ) const
{
	return mb_tag_dim_->get_reset_dist();
}

inline std::size_t moab_tag_chunk_seq::get_skip_count(void ) const
{
	return mb_tag_dim_->get_skip_count();
}

inline std::size_t moab_tag_chunk_seq::get_skip_dist(void ) const
{
	return mb_tag_dim_->get_skip_dist();
}

// Moab tag chunk member funcs

inline moab_tag_chunk::iterator moab_tag_chunk::begin(void )
{
	return chunk_seqs_.begin();
}

inline moab_tag_chunk::const_iterator moab_tag_chunk::end(void ) const
{
	return chunk_seqs_.end();
}

inline std::size_t moab_tag_chunk::size(void ) const
{
	return sz_; 
}

inline bool moab_tag_chunk::has_missing_val(void ) const
{
	return has_missing_val_;
}

inline double moab_tag_chunk::get_missing_val(void ) const
{
	return missing_val_;
}

inline bool moab_tag_chunk::needs_reduce(void ) const
{
	return needs_reduce_;
}

inline bool moab_tag_chunk::is_init(void ) const
{
	return is_init_;
}

inline moab_tag_chunk::key_type moab_tag_chunk::get_key(void ) const
{
	// FIXME:  We assume,
	// As long as we begin on the same pos chunks are the same
	assert(!chunk_seqs_.empty());
	return chunk_seqs_.front().begin();
}

inline bool moab_tag_chunk::key_compare_functor::operator()(
	const moab_tag_chunk::key_type k1,
	const moab_tag_chunk::key_type k2) const
{
	return (k1 < k2);
}
} // namespace util
} // namespace pargal

#endif /* PARGAL_MOAB_TAG_CHUNK_HPP_INCLUDED */
