/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_MOAB_TAG_COUNTER_HPP_INCLUDED
#define PARGAL_MOAB_TAG_COUNTER_HPP_INCLUDED

#include<cstddef>
#include<vector>
#include<memory>
#include<functional>
#include<algorithm>
#include "pargal/pcvar.hpp"
#include "pargal/moab_tag_iterator.hpp"

namespace pargal{
namespace util{

// FIXME: Move the different classes to different files?

// FIXME: Ideally these should be size_ts instead of ints - interaction with
// MPI is easier with ints though - no MPI_SIZE_T
class moab_tag_seq_counter{
	public:
		typedef int * iterator;
		typedef const int * const_iterator;

		moab_tag_seq_counter(int *begin, int *end);

		iterator begin(void ) const;
		const_iterator end(void ) const;
		std::size_t size(void ) const;

	private:
		iterator begin_, end_;
};

class moab_tag_chunk_counter{
	public:
		typedef std::vector<moab_tag_seq_counter>::iterator iterator;
		typedef std::vector<moab_tag_seq_counter>::const_iterator const_iterator;

		moab_tag_chunk_counter(std::size_t sz, int init_val,
			std::size_t odim_tc_to_counter_ratio,
			moab_tag_dim_type odim_type);
		void init_seq_counters(pargal::util::moab_tag_chunk &tc);

		iterator begin(void );
		const_iterator end(void ) const;
		int &operator[](const size_t index);
		std::size_t size(void ) const;

	private:
		bool is_seq_counters_init_;
		std::vector<moab_tag_seq_counter> seq_counters_;
		std::vector<int> chunk_counter_;
		// FIXME: Is there a better way than storing this ratio & odim here?
		// tc_sz/counter_sz for the operating dim
		// Used to create seq counters for multi-seq tag chunks
		std::size_t odim_tc_to_counter_ratio_;
		moab_tag_dim_type odim_type_;
};


class var_red_counter_iterator{
	public:
		var_red_counter_iterator(std::vector<moab_tag_chunk_counter>::iterator begin, std::vector<moab_tag_chunk_counter>::iterator end);
		void operator++();
		moab_tag_chunk_counter &operator*() const;
		bool operator!=(const var_red_counter_iterator &other_iter) const;
	private:
		std::vector<moab_tag_chunk_counter>::iterator begin_, end_, citer_;
};

class var_red_counter{
	public:
		typedef var_red_counter_iterator iterator;
		var_red_counter(pargal::pcvar &var, const std::size_t odim);
		iterator begin(void ) const;
		const iterator end(void ) const;
		moab_tag_chunk_counter &operator[](const size_t index);
		std::size_t size(void ) const;
		void reset(void );
	private:
		std::vector<moab_tag_chunk_counter> chunk_counters_;
		iterator begin_, end_;
};

// Moab tag seq counter inlines
inline moab_tag_seq_counter::moab_tag_seq_counter(int *begin, int *end):
			begin_(begin), end_(end)
{}

inline moab_tag_seq_counter::iterator moab_tag_seq_counter::begin(void ) const
{
	return begin_;
}

inline moab_tag_seq_counter::const_iterator moab_tag_seq_counter::end(void ) const
{
	return end_;
}

inline std::size_t moab_tag_seq_counter::size(void ) const
{
	return (end_ - begin_);
}

// Moab tag chunk counter inlines
inline moab_tag_chunk_counter::moab_tag_chunk_counter(std::size_t sz, int init_val,
		std::size_t odim_tc_to_counter_ratio,
		moab_tag_dim_type odim_type):
			is_seq_counters_init_(false), chunk_counter_(sz, init_val),
			odim_tc_to_counter_ratio_(odim_tc_to_counter_ratio),
			odim_type_(odim_type)
{}

inline moab_tag_chunk_counter::iterator moab_tag_chunk_counter::begin(void )
{
	return seq_counters_.begin();
}

inline moab_tag_chunk_counter::const_iterator moab_tag_chunk_counter::end(void ) const
{
	return seq_counters_.end();
}

inline int& moab_tag_chunk_counter::operator[](const size_t index)
{
	return chunk_counter_[index];
}

inline std::size_t moab_tag_chunk_counter::size(void ) const
{
	return chunk_counter_.size();
}

// Var red counter inlines
inline pargal::util::var_red_counter::iterator
pargal::util::var_red_counter::begin(void) const
{
	return begin_;
}

inline const pargal::util::var_red_counter::iterator
pargal::util::var_red_counter::end(void ) const
{
	return end_;
}

inline pargal::util::moab_tag_chunk_counter &pargal::util::var_red_counter::operator[](const size_t index)
{
	return chunk_counters_[index];
}

inline std::size_t var_red_counter::size(void ) const
{
	return chunk_counters_.size();
}

// var red counter iterator inlines
inline pargal::util::var_red_counter_iterator::var_red_counter_iterator(
	std::vector<moab_tag_chunk_counter>::iterator begin,
	std::vector<moab_tag_chunk_counter>::iterator end
):begin_(begin),citer_(begin),end_(end)
{
}

inline void
pargal::util::var_red_counter_iterator::operator++()
{
	++citer_;
	if(citer_ == end_){
		citer_ = begin_;
	}
}

inline pargal::util::moab_tag_chunk_counter &
pargal::util::var_red_counter_iterator::operator*() const
{
	return *citer_;
}

inline bool
pargal::util::var_red_counter_iterator::operator!=(
	const pargal::util::var_red_counter_iterator &other_iter) const
{
	// reduction counters never stop - infinitely loops
	return true;
}

} // namespace util
} // namespace pargal

#endif /* PARGAL_MOAB_TAG_COUNTER_HPP_INCLUDED */
