/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_MOAB_TAG_DIM_HPP_INCLUDED
#define PARGAL_MOAB_TAG_DIM_HPP_INCLUDED

#include "moab_tag_iterator.hpp"

namespace pargal{
namespace util{

enum moab_tag_dim_type{
	// Dim belongs in a sparse tag
	MOAB_TAG_SPARSE_DIM=0,
	// Dim belongs in a dense tag and is ordinary
	MOAB_TAG_GEN_DIM,
	// Corresponds to "lev"/"ilev" in a dense tag
	MOAB_TAG_LEV_DIM,
	// Corresponds to "lat"/"slat" in a dense tag - SQIJ partition method
	MOAB_TAG_SQIJ_LAT_DIM,
	// Corresponds to "lon"/"slon" in a dense tag - SQIJ partition method
	MOAB_TAG_SQIJ_LON_DIM,
	// Invalid dim
	MOAB_TAG_INVALID_DIM
};

template<typename dim>
struct dim_traits{
	typedef typename dim::dim_iterator_category dim_iterator_category;
	typedef typename dim::dim_red_iterator_category dim_red_iterator_category;
};

class moab_tag_dim : public iterable{
	public:
		std::size_t get_static_count(void ) const;
		std::size_t get_reset_count(void ) const;
		std::size_t get_reset_dist(void ) const;
		std::size_t get_skip_count(void ) const;
		std::size_t get_skip_dist(void ) const;
		// virtual constructor
		virtual moab_tag_dim *clone(void ) const = 0;
		virtual ~moab_tag_dim();
	protected:
		moab_tag_dim(std::size_t static_cnt, std::size_t reset_cnt,
			std::size_t reset_dist, std::size_t skip_cnt, std::size_t skip_dist);
		moab_tag_dim();
		std::size_t static_cnt_, reset_cnt_, skip_cnt_;
		std::size_t reset_dist_, skip_dist_;
};

moab_tag_dim *create_moab_tag_dim(pargal::pcvar &var, const std::size_t dim);

class moab_tag_sparse_dim : public moab_tag_dim{
	public:
		typedef moab_tag_sparse_skip_dim_iterator dim_iterator_category;
		typedef moab_tag_sparse_static_dim_iterator dim_red_iterator_category;

		moab_tag_sparse_dim(pargal::pcvar &var, const std::size_t dim);
		moab_tag_dim *clone(void ) const;
	private:
		moab_tag_sparse_dim(std::size_t static_cnt, std::size_t reset_cnt,
			std::size_t reset_dist, std::size_t skip_cnt, std::size_t skip_dist);
};

class moab_tag_gen_dim : public moab_tag_dim{
	public:
		typedef moab_tag_gen_dim_iterator dim_iterator_category;
		typedef moab_tag_reset_dim_iterator dim_red_iterator_category;

		moab_tag_gen_dim(pargal::pcvar &var, const std::size_t dim);
		moab_tag_dim *clone(void ) const;
	private:
		moab_tag_gen_dim(std::size_t static_cnt, std::size_t reset_cnt,
			std::size_t reset_dist, std::size_t skip_cnt, std::size_t skip_dist);
};

class moab_tag_lev_dim : public moab_tag_dim{
	public:
		typedef moab_tag_gen_dim_iterator dim_iterator_category;
		typedef moab_tag_static_skip_dim_iterator dim_red_iterator_category;

		moab_tag_lev_dim(pargal::pcvar &var, const std::size_t dim);
		moab_tag_dim *clone(void ) const;
	private:
		moab_tag_lev_dim(std::size_t static_cnt, std::size_t reset_cnt,
			std::size_t reset_dist, std::size_t skip_cnt, std::size_t skip_dist);
};

class moab_tag_sqij_lat_dim : public moab_tag_dim{
	public:
		typedef moab_tag_gen_dim_iterator dim_iterator_category;
		typedef moab_tag_reset_dim_iterator dim_red_iterator_category;

		moab_tag_sqij_lat_dim(pargal::pcvar &var, const std::size_t dim);
		moab_tag_dim *clone(void ) const;
	private:
		moab_tag_sqij_lat_dim(std::size_t static_cnt, std::size_t reset_cnt,
			std::size_t reset_dist, std::size_t skip_cnt, std::size_t skip_dist);
};

class moab_tag_sqij_lon_dim : public moab_tag_dim{
	public:
		typedef moab_tag_gen_dim_iterator dim_iterator_category;
		typedef moab_tag_reset_skip_dim_iterator dim_red_iterator_category;

		moab_tag_sqij_lon_dim(pargal::pcvar &var, const std::size_t dim);
		moab_tag_dim *clone(void ) const;
	private:
		moab_tag_sqij_lon_dim(std::size_t static_cnt, std::size_t reset_cnt,
			std::size_t reset_dist, std::size_t skip_cnt, std::size_t skip_dist);
};

// The inlines
inline pargal::util::moab_tag_dim::moab_tag_dim(std::size_t static_cnt,
	std::size_t reset_cnt, std::size_t reset_dist, std::size_t skip_cnt,
	std::size_t skip_dist):static_cnt_(static_cnt), reset_cnt_(reset_cnt),
	reset_dist_(reset_dist), skip_cnt_(skip_cnt), skip_dist_(skip_dist)
{}

inline pargal::util::moab_tag_dim::moab_tag_dim(): static_cnt_(0), reset_cnt_(0),
	reset_dist_(0), skip_cnt_(0), skip_dist_(0)
{}

inline std::size_t pargal::util::moab_tag_dim::get_static_count(void ) const
{
	return static_cnt_;
}

inline std::size_t pargal::util::moab_tag_dim::get_reset_count(void ) const
{
	return reset_cnt_;
}

inline std::size_t pargal::util::moab_tag_dim::get_reset_dist(void ) const
{
	return reset_dist_;
}

inline std::size_t pargal::util::moab_tag_dim::get_skip_count(void ) const
{
	return skip_cnt_;
}

inline std::size_t pargal::util::moab_tag_dim::get_skip_dist(void ) const
{
	return skip_dist_;
}

// FIXME: Is this required?
inline pargal::util::moab_tag_dim::~moab_tag_dim()
{
}

// FIXME: How can we avoid using these macros?
#define PARGAL_VAR_DIM_VOID_FUNC_DISPATCH(dim_type_, fname_, ...) do{		\
	switch((dim_type_)){													\
		case pargal::util::MOAB_TAG_SPARSE_DIM:								\
			fname_<util::moab_tag_sparse_dim>(__VA_ARGS__);					\
			break;															\
		case pargal::util::MOAB_TAG_GEN_DIM:								\
			fname_<util::moab_tag_gen_dim>(__VA_ARGS__);					\
			break;															\
		case pargal::util::MOAB_TAG_LEV_DIM:								\
			fname_<util::moab_tag_lev_dim>(__VA_ARGS__);					\
			break;															\
		case pargal::util::MOAB_TAG_SQIJ_LAT_DIM:							\
			fname_<util::moab_tag_sqij_lat_dim>(__VA_ARGS__);				\
			break;															\
		case pargal::util::MOAB_TAG_SQIJ_LON_DIM:							\
			fname_<util::moab_tag_sqij_lon_dim>(__VA_ARGS__);				\
			break;															\
		default:															\
			throw pargal::pargal_except("Unknown dim type", __FILE__, __LINE__,	\
					"Unknown");												\
	}																		\
}while(0)

#define PARGAL_VAR_DIM_NON_VOID_FUNC_DISPATCH(ret_, dim_type_, fname_, ...) do{		\
	switch((dim_type_)){													\
		case pargal::util::MOAB_TAG_SPARSE_DIM:								\
			(ret_) = fname_<util::moab_tag_sparse_dim>(__VA_ARGS__);		\
			break;															\
		case pargal::util::MOAB_TAG_GEN_DIM:								\
			(ret_) = fname_<util::moab_tag_gen_dim>(__VA_ARGS__);			\
			break;															\
		case pargal::util::MOAB_TAG_LEV_DIM:								\
			(ret_) = fname_<util::moab_tag_lev_dim>(__VA_ARGS__);			\
			break;															\
		case pargal::util::MOAB_TAG_SQIJ_LAT_DIM:							\
			(ret_) = fname_<util::moab_tag_sqij_lat_dim>(__VA_ARGS__);		\
			break;															\
		case pargal::util::MOAB_TAG_SQIJ_LON_DIM:							\
			(ret_) = fname_<util::moab_tag_sqij_lon_dim>(__VA_ARGS__);		\
			break;															\
		default:															\
			throw pargal::pargal_except("Unknown dim type", __FILE__, __LINE__,	\
					"Unknown");												\
	}																		\
}while(0)

} // namespace util
} // namespace pargal

#endif /* PARGAL_MOAB_TAG_DIM_HPP_INCLUDED */
