/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_ALGO_HPP_INCLUDED
#define PARGAL_ALGO_HPP_INCLUDED

#include <pargal/pcvar.hpp>
#include <pargal/exception.hpp>
#include <moab/ParallelComm.hpp>
#include "MBEntityType.h"
#include "MBTagConventions.hpp"
#include <sstream>

// Intrepid includes
#include <external/intrepid/Intrepid_Utils.hpp>
#include "Intrepid_FunctionSpaceTools.hpp"
#include "Intrepid_CellTools.hpp"
#include "Intrepid_HGRAD_QUAD_C1_FEM.hpp"
#include "Intrepid_DefaultCubatureFactory.hpp"

// Teuchos includes
#include "Teuchos_GlobalMPISession.hpp"

// Shards includes
#include "Shards_CellTopology.hpp"

// Epetra includes
#include "Epetra_Map.h"
#include "Epetra_MpiComm.h"
#include "Epetra_FECrsMatrix.h"
#include "Epetra_FEVector.h"

// AztecOO includes
#include "AztecOO.h"

// ML includes
#include <external/intrepid/ml_MultiLevelPreconditioner.hpp>

namespace pargal 
{  
  class pcalgo 
  {    
  public:        
    // implementation of vorticity
    void vorticity(pcvar& var_u, pcvar& var_v, pcvar& var_vort);
    
    // implementation of divergence
    void divergence(pcvar& var_u, pcvar& var_v, pcvar& var_div);
    
  private:
    void vort_div_cell_center(pcvar& var_u, pcvar& var_v, pcvar& var_vort_div, 
		              bool is_algo_vort);

    void vort_div_nodes_spectral(pcvar& var_u, pcvar& var_v, pcvar& var_vort_div, 
		                 bool is_algo_vort);


    /** \brief Computes cubed-sphere face number for element defined by Cartesian
               3-d coordinates.

        \param  elemFaceNum       [out] - Face number associated with element (1 - 6)
        \param  elemCoordsSphere   [in] - Quadrilateral element nodes given in 
                                          Cartesian 3-d coordinates on unit sphere
    */
    template<class ArrayIn, class ArrayOut>
    void getElementFaceNo(ArrayOut &elemFaceNum, ArrayIn &elemCoordsSphere);

    /** \brief Computes element coordinates in cubed-sphere frame given the element face
               number and quadrilateral element nodal coordinates in 3-d Cartesian
               representation on unit sphere.

        \param  elemCoordsCube    [out] - Coordinates of quadrilateral element nodes
                                          on cubed sphere face. Equiangular projection 
                                          assumes coordinates are in [pi/4,pi/4].
        \param  elemFaceNum        [in] - Face number associated with element (1 - 6)
        \param  elemCoordsSphere   [in] - Quadrilateral element nodes given in 
                                          Cartesian 3-d coordinates on unit sphere
    */
    template<class ArrayIn1, class ArrayIn2, class ArrayOut>
    void getCoordsCube(ArrayOut &elemCoordsCube, ArrayIn1 &elemFaceNum, ArrayIn2 &elemCoordsSphere);

    /** \brief Computes matrix needed to convert zonal and meridional velocity components
               to covariant or contravariant components.

        \param  D                [out] - matrix for converting velocity components
        \param  elemFaceNum       [in] - Face number associated with element (1 - 6)
        \param  pointsInElem      [in] - Points in element to evaluate jacobG
    */
    template<class ArrayIn1, class ArrayIn2, class ArrayOut>
    void getDMatrix(ArrayOut &D, ArrayIn1 &pointsInElem, ArrayIn2 &elemFaceNum);

    /** \brief Computes Jacobian determinant of metric tensor for cubed sphere
               grid.

        \param  jacobG           [out] - Jacobian determinant of metric tensor
        \param  pointsInElem      [in] - Points in element to evaluate jacobG
    */
    template<class ArrayIn, class ArrayOut>
    void getJacobianG(ArrayOut &jacobG, ArrayIn &pointsInElem);
    
    static void init_bilinear_quad();

    //static void init_spectral_quad(int degree);
    
    ///////////////////////////////////////////
    // data member
    ///////////////////////////////////////////
    static int numCubPoints;        

    static double radius; // radius of the Earth
    
    // cell topology 
    static shards::CellTopology cellType;
    
    // integration points and weights
    static Intrepid::FieldContainer<double> cubPoints; 
    static Intrepid::FieldContainer<double> cubWeights;
    
    // basis function values and gradients
    static Intrepid::FieldContainer<double> basisValues;
    static Intrepid::FieldContainer<double> basisGrads;

    static int numFields;
  };
} // namespace pargal

#endif
