/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PARGAL_VAR_HPP_INCLUDED
#define PARGAL_VAR_HPP_INCLUDED

#include <moab/Interface.hpp>
#include <pargal/pcdim.hpp>
#include <pargal/pcatt.hpp>
#include <pargal/fileinfo.hpp>
#include <pargal/exception.hpp>
#include <moab/Types.hpp>
#include <pargal/mesh.hpp>
#include <pargal/moab_tag_iterator.hpp>
#include <pargal/moab_tag_dim.hpp>
#include <string>
#include <vector>
#include <sstream>
#include <memory>
#include <utility>

namespace pargal 
{  
  class pcvar 
  {    
  public:
	// Available data cache management policies
	enum cache_mode{
		// Optimize memory usage
		OPT_MEM,
		// Optimize computation time
		OPT_COMP
	};
	class cache_access_state{
		public:
			moab::Range ent_range;
			moab::Range::iterator begin;
	};
	// Available chunk iterators - used internally by algorithms
	typedef util::moab_tag_chunk_iterator iterator;
	typedef const util::moab_tag_chunk_iterator const_iterator;
	typedef util::moab_tag_reset_chunk_iterator red_iterator;
	typedef const util::moab_tag_reset_chunk_iterator const_red_iterator;

    ///////////////////////////////////////////
    // constructor destructor
    ///////////////////////////////////////////
    
    //constructor  
    inline explicit pcvar(const std::string& name, 
			  moab::DataType type, 
			  const std::vector<pcdim>& dims,
			  const fileinfo& finfo,
			  const std::vector<int>& tsteps,
			  bool usr_created = false,
			  const std::string& loc = std::string("QUAD"),
			  bool usr_allocate_memory = false,
			  double missing_val = 0.0
			  );			    
    
    //! destructor
    inline ~pcvar();
    
    ///////////////////////////////////////////
    // queries
    ///////////////////////////////////////////
    
    //! return the name of the variable
    inline const std::string& name() const;
    
    //! return the type of the variable
    inline moab::DataType type() const;
    
    //! return the number of dimensions of the variable
    inline std::size_t num_dims() const;
    
    //! return a reference to the index-th dimension (starting at 0)
    inline pcdim get_dim(const std::size_t index) const;

	//! return the dim type
	util::moab_tag_dim_type get_dim_type(const size_t dim);

    //! return the local min/max of the dimension (including shared part)
    //! assume sqij partition method
    //! returned value is in imin/imax/jmin/jmax order
    //std::vector<std::size_t> get_dim_loc_minmax() const;
    
    //! return the local min/max of the dimension (owned only)
    //! assume sqij partition method
    //! returned value is in imin/imax/jmin/jmax order
    std::vector<int> get_dim_loc_minmax_owned() const;

    //! return the global min/max of the dimension
    //! returned value is in imin/imax/jmin/jmax order
    std::vector<int> get_dim_global_minmax() const;
      
    //! return the processes' rank whose column is the same as the current process
    //! current process's rank is also included
    std::vector<int> get_column_ranks() const;

    //! return the processes' rank whose row is the same as the current process 
    //! current process's rank is also included
    std::vector<int> get_row_ranks() const;

    //! return the dimensions of the variable
    inline const std::vector<pcdim>& get_dims() const;

	//! return the dimension sizes of the variable
	void get_dims(std::vector<std::size_t> &dims) const;

	//! return the local dimensions of a variable
	void get_local_dims(std::vector<size_t> &ldims);

    //! initialize the variable
    void init(double init_val);

	// FIXME: We need better names than get_storage() and delete_storage()
	// We also need a way to disable deletion of data read from files

    //! return the pointer pointing to the owned data at time step indexed by tindex
    double* get_storage(int tindex = 0);

	//! Read data at mesh data chunk indexed by cache_idx to pcvar cache
	//! and get a pointer to it. Data chunks usually corresponds to
	//! pcvar timesteps.
	//! Returns true if there is more data to read, false otherwise
	//! If there is more data to read the function updates the begin
	//! iterator to the begin of the next read
	//! If there is more data to read, read() needs to be called again
	//! with cont=true and the returned moab iterator, begin
	/**
		Returns true if there is more data to be read, false otherwise
		\param chunk_idx    The mesh data chunk index, usually the timestep, to read
		\param buf          A pointer to the mesh data chunk
		\param buf_len      The length of the mesh data chunk returned
		\param cont         Indicates if sub chunks in the mesh data chunk are continued to be read. cont == false the first time a mesh data chunk is read, if read() returns true subsequent calls to read need to set cont == true
		\param begin        Moab iterator to the beginning of the data chunk, is valid only if cont == true. This iterator is initialized and updated by the function if there is more data to read 
		Example: \code
				int tstep = 0; //NUM_OF_TSTEPS
				bool cont = false;
				do{
					pcvar::cache_access_state rd_state;
					do{
						cont = var.cache_read(tstep, buf, buf_len, cont, rd_state);
						// Operate on buf
					}while(cont == true);
					var.cache_delete(tstep);
					tstep++;
				}while(tstep < NUM_OF_TSTEPS);
				\endcode
	*/
	bool cache_read(int chunk_idx, double *&buf, std::size_t &buf_len, bool cont, cache_access_state &rd_state);

	//! Delete data cached for data chunk indexed by chunk_idx
	void cache_delete(int chunk_idx);
    
	//! set storage cache mode
	void set_storage_mode(cache_mode mode);

	//! get storage cache mode
	cache_mode get_storage_mode(void ) const;
    
    //! delete the data at time step indexed by tindex
    void delete_storage(int tindex);
    
    //! return the index for a given time step; return -1 if not found
    int get_index(int tstep) const;

    //! return the moab tag at time step indexed by tindex
    moab::Tag get_moab_tag(int tindex = 0);

	//! return moab tag chunk iterators
	//! Note: These iterators are internally used by algorithms
	//! to iterator through tag chunks
	iterator get_chunk_iterator(const size_t odim);
	red_iterator get_red_chunk_iterator(const size_t odim);

    //! return the size of a variable (owned) including all time steps
    //! Note: current implementaion assumes it is the number of
    //! vertices of the mesh at one time step times the number of 
    //! time steps
    std::size_t size();

    //! return the size of a variable (owned) for a single time step
    std::size_t tstep_size();

    //! return the time steps the variable is referring to
    const std::vector<int>& get_tsteps() const;

    //! return the fileinfo
    inline fileinfo& get_finfo(); 
    inline const fileinfo& get_finfo() const; 

    //! return whether the variable is user created
    inline bool is_usr_created() const;

    //! return the variable location information
    inline const std::string& get_loc() const;

    //! return the number of levels of the variable
    inline std::size_t num_levels() const;

    //! return the global min/max of the vertex dimension
    //! returned value is in imin/imax/jmin/jmax order
    inline std::vector<int> get_dim_vertex_global_minmax() const;

    //! return the missing val if it has one
    //! otherwise return mininum double defined by the stl library
    inline double get_missing_val() const;
    
    //! return whether missing val has been defined
    inline bool has_missing_val() const;

  private:
    ///////////////////////////////////////////
    // helper functions
    ///////////////////////////////////////////

    //! init local dimesions
    void init_dim_loc_minmax_owned();

    //! init global dimesions
    void init_dim_global_minmax();

  private:
    ///////////////////////////////////////////
    // data member
    ///////////////////////////////////////////
    
    std::string m_name;
    moab::DataType m_type;
    std::vector<pcdim> m_dims;
    fileinfo& m_finfo;
    std::vector<int> m_tsteps;
    std::stringstream m_opts;
    bool m_usr_created;
    std::string m_loc;    
	// This chunk cache only caches pointer to the first seq of a chunk
    std::vector<std::pair<double*, std::size_t> > mesh_chunks_cache_;
    std::size_t tstep_sz_;
    std::size_t m_num_lev;
    std::vector<int> lDims;
    std::vector<int> lCDims;
    std::vector<int> gDims;
    std::vector<int> gCDims;
    bool m_has_missing_val;
    double m_missing_val;
	cache_mode storage_mode_;
  };
  
  ///////////////////////////////////////////
  // implementation detail
  ///////////////////////////////////////////
  
  //constructor  
  pcvar::pcvar(const std::string& name, 
	       moab::DataType type, 
	       const std::vector<pcdim>& dims,
	       const fileinfo& finfo,
	       const std::vector<int>& tsteps,
	       bool usr_created,
	       const std::string& loc,
	       bool usr_allocate_mem,
	       double missing_val
	       ) :			    
    m_name(name),
    m_type(type),
    m_dims(dims),
    m_finfo(const_cast<fileinfo&>(finfo)),
    m_tsteps(tsteps),
    m_usr_created(usr_created),
    m_loc(loc),
    tstep_sz_(0),
    m_num_lev(1),
    m_has_missing_val(false),
    m_missing_val(missing_val),
	storage_mode_(OPT_MEM)
  {
    m_opts << "PARALLEL=READ_PART;PARTITION;PARTITION_METHOD=";
    m_opts << finfo.get_part_method();
    m_opts << ";VARIABLE=";
    m_opts << m_name;  
    m_opts << ";NOMESH;";          

    if (m_missing_val != 0)
      m_has_missing_val = true;

    if (!usr_created)
      m_loc = m_finfo.get_var_location(m_name);
    
	size_t ent_sz = 1;
	if(m_loc == "SET"){
		for (std::size_t i = 0; i != m_dims.size(); ++i) {
			ent_sz *= m_dims[i].size();
		}
	}
	else{
		for (std::size_t i = 0; i != m_dims.size(); ++i) {
			if ((m_dims[i].name() == "lev") ||
				(m_dims[i].name() == "ilev") ||
				(m_dims[i].name() == "nVertLevels")){
				ent_sz = m_num_lev =  m_dims[i].size();
			}
		}
	}

    mesh_chunks_cache_.resize(m_tsteps.size());    
    if (usr_allocate_mem) {      
      moab::Interface& mb = m_finfo.get_moab();
      moab::EntityHandle& file_set = m_finfo.get_file_set();      
      for (std::size_t idx=0; idx != m_tsteps.size(); ++idx) {
		std::ostringstream tname;
		if(m_loc == "SET"){
			tname << m_name;
		}
		else{
			tname << m_name << m_tsteps[idx];
		}
		void *tag_ptr = NULL;
		moab::Tag tmp_tag;
		unsigned int tag_flags = 0x0;
		if(m_loc == "SET"){
			tag_flags = moab::MB_TAG_SPARSE | moab::MB_TAG_VARLEN;
		}
		else{
			tag_flags = moab::MB_TAG_DENSE;
		}
		moab::ErrorCode rval = mb.tag_get_handle(tname.str().c_str(), ent_sz, moab::MB_TYPE_DOUBLE, tmp_tag, tag_flags);
	if (rval != moab::MB_SUCCESS)
	  throw pargal_except("Error: Tags are supposed to created already!", 
			      __FILE__, __LINE__, __PRETTY_FUNCTION__);
	else {
	  int count = -1;
	  if (m_loc == "VERTEX") {
	    moab::Range verts_owned = get_local_entities_owned<moab::MBVERTEX>(mb, file_set);
	    rval = const_cast<moab::Interface&>(mb).tag_iterate(tmp_tag, verts_owned.begin(), verts_owned.end(), count, tag_ptr);
	    int nverts = verts_owned.size();
	    if (rval != moab::MB_SUCCESS || count <= 0) 
	      throw pargal_except("Error: " + mb.get_error_string(rval), 
				  __FILE__, __LINE__, __PRETTY_FUNCTION__);	
	  }
		else if(m_loc == "POLYGON"){
			moab::Range polygons_owned = get_local_entities_owned<moab::MBPOLYGON>(
											mb, file_set);
			rval = const_cast<moab::Interface&>(mb).tag_iterate(tmp_tag,
						polygons_owned.begin(), polygons_owned.end(), count, tag_ptr);
			int npolygons = polygons_owned.size();
			if((rval != moab::MB_SUCCESS) || (count <= 0)){ 
				throw pargal_except("Error: " + mb.get_error_string(rval), 
					  __FILE__, __LINE__, __PRETTY_FUNCTION__);	
			}
		}
		else if(m_loc == "EDGE"){
			moab::Range edges_owned = get_local_entities_owned<moab::MBEDGE>(
											mb, file_set);
			rval = const_cast<moab::Interface&>(mb).tag_iterate(tmp_tag,
						edges_owned.begin(), edges_owned.end(), count, tag_ptr);
			int nedges = edges_owned.size();
			if((rval != moab::MB_SUCCESS) || (count <= 0)){ 
				throw pargal_except("Error: " + mb.get_error_string(rval), 
					  __FILE__, __LINE__, __PRETTY_FUNCTION__);	
			}
		}
	  else if (m_loc == "QUAD") {
	    moab::Range quads_owned = get_local_entities_owned<moab::MBQUAD>(mb, file_set);
	    rval = const_cast<moab::Interface&>(mb).tag_iterate(tmp_tag, quads_owned.begin(), quads_owned.end(), count, tag_ptr);
	    int nquads = quads_owned.size();
	    if (rval != moab::MB_SUCCESS || count <= 0) 
	      throw pargal_except("Error: " + mb.get_error_string(rval), 
				  __FILE__, __LINE__, __PRETTY_FUNCTION__);	
	  }
	  else if (m_loc == "SET") {
		const void *tagp = NULL;
		int tag_len;
		rval = const_cast<moab::Interface&>(mb).tag_get_by_ptr(tmp_tag, &file_set, 1, &tagp, &count);
	    if (rval != moab::MB_SUCCESS || count <= 0){ 
	      throw pargal_except("Error: " + mb.get_error_string(rval), 
				  __FILE__, __LINE__, __PRETTY_FUNCTION__);	
		}
		// FIXME: This is ugly...
		tag_ptr = const_cast<void *>(tagp);
	  }
	  else 
	    throw pargal_except("Error: locations other than QUAD/VERTEX/SET/POLYGON/EDGE are not implemented yet!", 
				__FILE__, __LINE__, __PRETTY_FUNCTION__);      
		mesh_chunks_cache_[idx] = std::make_pair<double *, std::size_t>(
									static_cast<double *>(tag_ptr), count);
	}
      }      
    }
    
    lDims.resize(4, -1);
    lCDims.resize(4, -1);
    gDims.resize(4, -1);
    gCDims.resize(4, -1);
    init_dim_loc_minmax_owned();
    init_dim_global_minmax();
    if (usr_created) return;
    const pcatt& var_att = m_finfo.get_var_att(m_name);
    std::set<std::string> var_att_names = var_att.get_att_names();
    std::string missing_val_att_name;
    if (std::find(var_att_names.begin(), var_att_names.end(), 
		  "missing_value") != var_att_names.end())
      missing_val_att_name="missing_value";
    else if (std::find(var_att_names.begin(), var_att_names.end(), 
		       "_FillValue") != var_att_names.end())
      missing_val_att_name="_FillValue";
    if (missing_val_att_name.empty()) return;
    m_has_missing_val=true;
    std::string att_data_type = var_att.get_data_type(missing_val_att_name);
    if (att_data_type == "float") {
      std::vector<float> vals;
      var_att.get_vals(missing_val_att_name, vals);
      m_missing_val=static_cast<double>(vals[0]);
    } else if (att_data_type== "double") {
      std::vector<double> vals;
      var_att.get_vals(missing_val_att_name, vals);
      m_missing_val=vals[0];
    }
    else {
      throw pargal_except("Error: data type other than float and double for missing values are not supported yet", 
			  __FILE__, __LINE__, __PRETTY_FUNCTION__); 
    }
  }
  
  //! destructor
  pcvar::~pcvar()
  {
    for (std::size_t idx = 0; idx != m_tsteps.size(); ++idx)
      this->delete_storage(idx);
  }
  
  //! return the name of the variable  
  const std::string& 
  pcvar::name() const
  {
    return m_name;
  }
    
  //! return the type of the variable  
  moab::DataType 
  pcvar::type() const
  {
    return m_type;
  }
  
  //! return the number of dimensions of the variable  
  std::size_t 
  pcvar::num_dims() const
  {
    return m_dims.size();
  }
  
  //! return a reference to the index-th dimension (starting at 0)   
  pcdim
  pcvar::get_dim(const std::size_t index) const
  {
    return m_dims[index];
  }
  
  const std::vector<pcdim>&
  pcvar::get_dims() const
  {
    return m_dims;
  }

	inline void pcvar::get_dims(std::vector<std::size_t> &dims) const
	{
		size_t ndims = m_dims.size();
		dims.resize(ndims);
		for(size_t i=0; i<ndims; i++){
			dims[i] = m_dims[i].size();	
		}
	}

	inline void pcvar::init(double init_val)
	{
		std::size_t tstep_sz = this->tstep_size();
		for(std::size_t tstep = 0; tstep < m_tsteps.size(); tstep++){
			cache_access_state rd_state;
			bool cont_reading_chunk = false;
			std::size_t seq_sz = 0;

			do{
				double *buf;
				cont_reading_chunk = this->cache_read(tstep, buf, seq_sz,
										cont_reading_chunk, rd_state);
				for(std::size_t idx=0; idx<seq_sz; idx++){
					buf[idx] = init_val;
				}	
			}while(cont_reading_chunk == true);
			// Don't delete storage
		}
	}

  //! return the fileinfo
  fileinfo& 
  pcvar::get_finfo()
  {
    return m_finfo;
  }  
  
  //! return the fileinfo
  const fileinfo& 
  pcvar::get_finfo() const
  {
    return m_finfo;
  }  

  //! return whether the variable is user created
  bool 
  pcvar::is_usr_created() const
  {
    return m_usr_created;
  }
  //! return the location information of the variable
  const std::string& 
  pcvar::get_loc() const
  {
    return m_loc;
  }

  //! return the number of levels of the variable
  std::size_t 
  pcvar::num_levels() const
  {
    return m_num_lev;
  }
  
  //! return the global min/max of the vertex dimension
  std::vector<int>
  pcvar::get_dim_vertex_global_minmax() const
  {
    return gDims;
  }

  //! return the missing val if it has one
  //! otherwise return mininum double defined by the stl library
  double 
  pcvar::get_missing_val() const
  {
    return m_missing_val;
  }

  //! return whether missing val has been defined
  bool 
  pcvar::has_missing_val() const
  {
    return m_has_missing_val;    
  }
  
} // namespace pargal

#endif
