/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define BOOST_TEST_MODULE vorticity_homme_test
#include <external/boost/test_included_unit_test.hpp>
#include <boost/test/output_test_stream.hpp>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <moab/Core.hpp>
#include <pargal.hpp>
#include "mpi.h"

using namespace pargal;
using namespace moab;

#define STRINGIFY_(X) #X
#define STRINGIFY(X) STRINGIFY_(X)

std::string var_name0("U"); 
std::string var_name1("V");
std::string var_name_vort("UVvort");

BOOST_AUTO_TEST_CASE (vorticity_homme_test)
{
  int rank = 0;
  int& argc = boost::unit_test::framework::master_test_suite().argc;
  char**& argv = boost::unit_test::framework::master_test_suite().argv;
  
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  std::string file_name(argv[1]);
  std::string vort_output(argv[2]);

  try
    {
  
      // set up MOAB interface and parallel communication
      moab::Core moab;
      moab::Interface& mb = moab;
      moab::ParallelComm mb_pcomm(&mb, MPI_COMM_WORLD);

      // create meshset
      EntityHandle file_set;
      ErrorCode rval = mb.create_meshset(MESHSET_SET, file_set);
      if (rval != MB_SUCCESS) 
        throw pargal_except("Error: " + mb.get_error_string(rval),
        			__FILE__, __LINE__, __PRETTY_FUNCTION__);
  
      // load file metadata
      std::vector<std::string> file_names;
      file_names.push_back(file_name);  
      std::string part_method("TRIVIAL_PARTITION");
      fileinfo finfo(file_names, mb, mb_pcomm, file_set, part_method);
  
      // load mesh
      std::stringstream opts;  
      opts << "PARALLEL=READ_PART;PARTITION;PARALLEL_RESOLVE_SHARED_ENTS;GATHER_SET=0;PARTITION_METHOD=";
      opts << part_method;
      opts << ";VARIABLE=";
      rval = mb.load_file(file_name.c_str(), &file_set, opts.str().c_str());
      if (rval != moab::MB_SUCCESS) 
        throw pargal_except("Error: " + mb.get_error_string(rval),
			__FILE__, __LINE__, __PRETTY_FUNCTION__);
  
      std::size_t num_tsteps = finfo.get_dim("time").size();
      std::vector<int> tsteps(num_tsteps);
      for (std::size_t i = 0; i != num_tsteps; ++i)
        tsteps[i] = i;  
  
      pcvar var0(var_name0, moab::MB_TYPE_DOUBLE, finfo.get_var_dims(var_name0),
             finfo, tsteps, false, finfo.get_var_location(var_name0), false, 0.0);

      pcvar var1(var_name1, moab::MB_TYPE_DOUBLE, finfo.get_var_dims(var_name1),
             finfo, tsteps, false, finfo.get_var_location(var_name1), false, 0.0);

      pcvar var_vort(var_name_vort, moab::MB_TYPE_DOUBLE, finfo.get_var_dims(var_name1),
                 finfo, tsteps, true, finfo.get_var_location(var_name1), false, 0.0);

  
      //compute vorticity
      vorticity(var0, var1, var_vort);
  
      //gather vorticity
      std::vector<double> vort_result;
      gather(var_vort, tsteps, vort_result);
  
      if (rank == 0) 
        {
          boost::test_tools::output_test_stream output(vort_output, true);
          //output.precision(3);

          for (size_t i = 0; i != vort_result.size(); ++i)
        	output << vort_result[i] << std::endl;
      
          BOOST_CHECK(output.match_pattern());
        }
    }
  catch (pargal::pargal_except& e)
    {
      e.get_info();
    }
  catch (std::exception& e)
    {
      std::cout << e.what() << std::endl;
    }
  catch (...)
    {
      std::cerr << "Unrecoganized Exception!" << std::endl;
    }

  MPI_Finalize();
}
