/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define BOOST_TEST_MODULE avg_lon_singlef_test
#include <external/boost/test_included_unit_test.hpp>
#include <boost/test/output_test_stream.hpp>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <moab/Core.hpp>
#include <pargal.hpp>
#include "mpi.h"

using namespace pargal;
using namespace moab;

#define STRINGIFY_(X) #X
#define STRINGIFY(X) STRINGIFY_(X)

/*
#ifdef MESHDIR
std::string file_name(STRINGIFY(MESHDIR) "/CAM/EulSpec/datasets/eul26x48x96.t.3.nc");
std::string ncl_avg_output(STRINGIFY(MESHDIR) "/CAM/EulSpec/native/dim_avg_n/eul26x48x96.T.LON.avg.dat");
#else
std::string file_name("../../../GridFiles/CAM/EulSpec/datasets/eul26x48x96.t.3.nc");
std::string ncl_avg_output("../../../GridFiles/CAM/EulSpec/native/dim_avg_n/eul26x48x96.T.LON.avg.dat");
#endif
*/

std::string var_avg_name("Tavg");

BOOST_AUTO_TEST_CASE (avg_lon_singlef_test)
{
  int rank = 0;
  int& argc = boost::unit_test::framework::master_test_suite().argc;
  char**& argv = boost::unit_test::framework::master_test_suite().argv;
  
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  std::string file_name(argv[1]);
  std::string ncl_avg_output(argv[2]);
  std::string var_name(argv[3]);

  try 
    { 
      // create moab instance   
      Core moab;
      Interface& mb = moab;
      moab::ParallelComm mb_pcomm(&mb, MPI_COMM_WORLD);

      // create meshset
      EntityHandle file_set;
      ErrorCode rval = mb.create_meshset(MESHSET_SET, file_set);
      if (rval != MB_SUCCESS) 
	throw pargal_except("Error: " + mb.get_error_string(rval),
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);

      // load file metadata
      std::vector<std::string> file_names;
      file_names.push_back(file_name);  
      std::string part_method("SQIJ");
      fileinfo finfo(file_names, mb, mb_pcomm, file_set, part_method);
  
      std::stringstream opts;  
      opts << "PARALLEL=READ_PART;PARTITION;PARALLEL_RESOLVE_SHARED_ENTS;PARALLEL_GHOSTS=2.0.1;PARTITION_METHOD=";
      opts << part_method;
      opts << ";VARIABLE=";
      rval = mb.load_file(file_name.c_str(), &file_set, opts.str().c_str());
      if (rval != moab::MB_SUCCESS) 
	throw pargal_except("Error: " + mb.get_error_string(rval),
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);
  
      std::size_t num_tsteps = finfo.get_dim("time").size();
      std::vector<int> tsteps(num_tsteps);
      for (std::size_t i = 0; i != num_tsteps; ++i)
	tsteps[i] = i;
  
      pcvar var(var_name, moab::MB_TYPE_DOUBLE, finfo.get_var_dims(var_name),
		finfo, tsteps, false);
      
      const std::vector<pcdim>& var_avg_dims = finfo.get_var_dims(var_name);
      pcvar var_avg(var_avg_name, moab::MB_TYPE_DOUBLE, 
		    var_avg_dims, finfo, tsteps, true,
			var.get_loc(), false, var.get_missing_val());
      
      //comupte average over lon
      std::vector<std::size_t> dims(1,3);
      dim_avg_n(var, var_avg, dims);
      
      //gather average
      std::vector<double> avg_result;
      gather(var_avg, tsteps, avg_result);
      
      if (rank == 0) 
	{
	  boost::test_tools::output_test_stream output(ncl_avg_output, true);
	  std::size_t time = var_avg.get_dim(0).size();
	  std::size_t lev =  var_avg.get_dim(1).size();
	  std::size_t lat =  var_avg.get_dim(2).size();
	  std::size_t lon =  var_avg.get_dim(3).size();
	  output.precision(16);
	  for (std::size_t t=0; t!=time; ++t)
	    for (std::size_t k=0; k!=lev; ++k)
	      for (std::size_t j=0; j!=lat; ++j)
		for (std::size_t i=0; i!=lon; ++i)
		  if (i == 0)
		    output << "(" << t << "," << k << "," << j << ")\t" 
			   << avg_result[t*lev*lat*lon+k*lat*lon+j*lon+i]
			   << std::endl;
	  BOOST_CHECK(output.match_pattern());
	}
    }
  catch (pargal::pargal_except& e) 
    {
      e.get_info();
    }
  catch (std::exception& e) 
    {
      std::cout << e.what() << std::endl;
    }
  catch (...) 
    {
      std::cerr << "Unrecoganized Exception!" << std::endl;
    }
  
  MPI_Finalize();
}
