/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define BOOST_TEST_MODULE attributes_test
#include <external/boost/test_included_unit_test.hpp>
#include <iostream>
#include <string>
#include <vector>
#include <moab/Core.hpp>
#include <pargal.hpp>
#include "mpi.h"

using namespace pargal;
using namespace moab;

#define STRINGIFY_(X) #X
#define STRINGIFY(X) STRINGIFY_(X)

#ifdef MESHDIR
std::string file_name(STRINGIFY(MESHDIR) "/CAM/EulSpec/datasets/eul26x48x96.t.3.nc");
#else
std::string file_name("../../../GridFiles/CAM/EulSpec/datasets/eul26x48x96.t.3.nc");
#endif

BOOST_AUTO_TEST_CASE (attributes_test)
{
  int& argc = boost::unit_test::framework::master_test_suite().argc;
  char**& argv = boost::unit_test::framework::master_test_suite().argv;
  
  MPI_Init(&argc, &argv);
  try 
    { 
      // create moab instance 
      Core moab;
      Interface& mb = moab;
      moab::ParallelComm mb_pcomm(&mb, MPI_COMM_WORLD);
      
      // create meshset
      EntityHandle file_set;
      ErrorCode rval = mb.create_meshset(MESHSET_SET, file_set);
      if (rval != MB_SUCCESS) 
	throw pargal_except("Error: " + mb.get_error_string(rval),
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);
      
      // load file metadata  
      std::vector<std::string> file_names;
      file_names.push_back(file_name);
      std::string part_method("SQIJ");
      fileinfo fi(file_names, mb, mb_pcomm, file_set, part_method);

      std::size_t num_global_atts = fi.num_atts();
      BOOST_CHECK(num_global_atts == 10);
      
      const pcatt& gl_att = fi.get_att();
      BOOST_CHECK(gl_att.size() == 10);
      
      std::set<std::string> global_att_names = gl_att.get_att_names();
      std::set<std::string> file_global_att_names;
      file_global_att_names.insert("Conventions");
      file_global_att_names.insert("source");
      file_global_att_names.insert("case");
      file_global_att_names.insert("title");
      file_global_att_names.insert("logname");
      file_global_att_names.insert("host");
      file_global_att_names.insert("Version");
      file_global_att_names.insert("revision_Id");
      file_global_att_names.insert("history");
      file_global_att_names.insert("nco_openmp_thread_number");
      BOOST_CHECK(global_att_names == file_global_att_names);

      const std::string source_data_type = gl_att.get_data_type("source");
      BOOST_CHECK(source_data_type == "char");
      
      std::string source_vals;
      gl_att.get_vals("source", source_vals);
      BOOST_CHECK(source_vals == "CAM");

      const std::string nco_openmp_data_type = gl_att.get_data_type("nco_openmp_thread_number");
      BOOST_CHECK(nco_openmp_data_type == "int");

      std::vector<int> nco_openmp_vals;
      gl_att.get_vals("nco_openmp_thread_number", nco_openmp_vals);
      BOOST_CHECK(nco_openmp_vals[0] == 1);

      const pcatt& var_att = fi.get_var_att("T");
      BOOST_CHECK(var_att.size() == 3);
      
      std::set<std::string> var_att_names = var_att.get_att_names();
      std::set<std::string> file_var_att_names;
      file_var_att_names.insert("long_name");
      file_var_att_names.insert("units");
      file_var_att_names.insert("cell_method");
      BOOST_CHECK(var_att_names == file_var_att_names);

      const std::string units_data_type = var_att.get_data_type("units");
      BOOST_CHECK(units_data_type == "char");
      
      std::string units_vals;
      var_att.get_vals("units", units_vals);
      BOOST_CHECK(units_vals == "K");
    }
  catch (pargal::pargal_except& e) 
    {
      e.get_info();
      BOOST_CHECK(false);
    }
  catch (std::exception& e) 
    {
      std::cout << e.what() << std::endl;
      BOOST_CHECK(false);
    }
  catch (...) 
    {
      std::cerr << "Unrecoganized Exception!" << std::endl;
      BOOST_CHECK(false);
    }
  
  MPI_Finalize();
}
