/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define BOOST_TEST_MODULE read_file_test
#include <external/boost/test_included_unit_test.hpp>
#include <iostream>
#include <string>
#include <vector>
#include <moab/Core.hpp>
#include <pargal.hpp>
#include "mpi.h"

using namespace pargal;
using namespace moab;

#define STRINGIFY_(X) #X
#define STRINGIFY(X) STRINGIFY_(X)

#ifdef MESHDIR
std::string file_name(STRINGIFY(MESHDIR) "/CAM/EulSpec/datasets/eul26x48x96.t.3.nc");
#else
std::string file_name("../../../GridFiles/CAM/EulSpec/datasets/eul26x48x96.t.3.nc");
#endif

BOOST_AUTO_TEST_CASE (read_file_test)
{
  int& argc = boost::unit_test::framework::master_test_suite().argc;
  char**& argv = boost::unit_test::framework::master_test_suite().argv;
  
  MPI_Init(&argc, &argv);
  try 
    { 
      // create moab instance 
      Core moab;
      Interface& mb = moab;
      moab::ParallelComm mb_pcomm(&mb, MPI_COMM_WORLD);
      
      // create meshset
      EntityHandle file_set;
      ErrorCode rval = mb.create_meshset(MESHSET_SET, file_set);
      if (rval != MB_SUCCESS) 
	throw pargal_except("Error: " + mb.get_error_string(rval),
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);
      
      // load file metadata  
      std::vector<std::string> file_names;
      file_names.push_back(file_name);
      std::string part_method("SQIJ");
      fileinfo fi(file_names, mb, mb_pcomm, file_set, part_method);
      
      std::size_t num_dims = fi.num_dims();
      BOOST_CHECK(num_dims == 4);
  
      std::size_t num_vars = fi.num_vars();
      BOOST_CHECK(num_vars == 8);
      
      std::set<std::string> dim_names = fi.dim_names();
      std::set<std::string> file_dim_names;
      file_dim_names.insert("time");
      file_dim_names.insert("lev");
      file_dim_names.insert("lat");
      file_dim_names.insert("lon");
      BOOST_CHECK(dim_names == file_dim_names);

      const std::set<std::string>& var_names = fi.var_names();
      std::set<std::string> file_var_names;
      file_var_names.insert("T");
      file_var_names.insert("U");
      file_var_names.insert("V");
      file_var_names.insert("lat");
      file_var_names.insert("lev");
      file_var_names.insert("lon");
      file_var_names.insert("time");
      file_var_names.insert("gw");
      BOOST_CHECK(var_names == file_var_names);
      
      BOOST_CHECK(fi.get_var_dim("T", 0).name() == "time");
      BOOST_CHECK(fi.get_var_dim("T", 1).name() == "lev");
      BOOST_CHECK(fi.get_var_dim("T", 2).name() == "lat");
      BOOST_CHECK(fi.get_var_dim("T", 3).name() == "lon");
      
      BOOST_CHECK(fi.get_var_dim("T", 0).size() == 3);
      BOOST_CHECK(fi.get_var_dim("T", 1).size() == 26);
      BOOST_CHECK(fi.get_var_dim("T", 2).size() == 48);
      BOOST_CHECK(fi.get_var_dim("T", 3).size() == 96);

      BOOST_CHECK(fi.get_var_location("T") == "QUAD");
      BOOST_CHECK(fi.get_var_location("U") == "QUAD");
      BOOST_CHECK(fi.get_var_location("V") == "QUAD");
    }
  catch (pargal::pargal_except& e) 
    {
      e.get_info();
      BOOST_CHECK(false);
    }
  catch (std::exception& e) 
    {
      std::cout << e.what() << std::endl;
      BOOST_CHECK(false);
    }
  catch (...) 
    {
      std::cerr << "Unrecoganized Exception!" << std::endl;
      BOOST_CHECK(false);
    }
  
  MPI_Finalize();
}
