/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define BOOST_TEST_MODULE read_file_test
#include <external/boost/test_included_unit_test.hpp>
#include <iostream>
#include <string>
#include <vector>
#include <moab/Core.hpp>
#include <pargal.hpp>
#include "mpi.h"

using namespace pargal;
using namespace moab;

#define STRINGIFY_(X) #X
#define STRINGIFY(X) STRINGIFY_(X)

#ifdef MESHDIR
std::string file_name(STRINGIFY(MESHDIR) "/CAM/HOMME/datasets/homme26x3458.t.3.nc");
#else
std::string file_name("../../../GridFiles/CAM/HOMME/datasets/homme26x3458.t.3.nc");
#endif

BOOST_AUTO_TEST_CASE (read_file_test)
{
  int& argc = boost::unit_test::framework::master_test_suite().argc;
  char**& argv = boost::unit_test::framework::master_test_suite().argv;
  
  MPI_Init(&argc, &argv);
  try 
    { 
      // create moab instance 
      Core moab;
      Interface& mb = moab;
      moab::ParallelComm mb_pcomm(&mb, MPI_COMM_WORLD);
      
      // create meshset
      EntityHandle file_set;
      ErrorCode rval = mb.create_meshset(MESHSET_SET, file_set);
      if (rval != MB_SUCCESS) 
	throw pargal_except("Error: " + mb.get_error_string(rval),
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);
      
      // load file metadata  
      std::vector<std::string> file_names;
      file_names.push_back(file_name);
      std::string part_method("TRIVIAL_PARTITION");
      fileinfo fi(file_names, mb, mb_pcomm, file_set, part_method);
      
      BOOST_CHECK(fi.get_var_location("T") == "VERTEX");
      BOOST_CHECK(fi.get_var_location("U") == "VERTEX");
      BOOST_CHECK(fi.get_var_location("V") == "VERTEX");

      std::stringstream opts;  
      opts << "PARALLEL=READ_PART;PARTITION;PARALLEL_RESOLVE_SHARED_ENTS;GATHER_SET=0;PARTITION_METHOD=";
      opts << part_method;
      opts << ";VARIABLE=";
      rval = mb.load_file(file_name.c_str(), &file_set, opts.str().c_str());
      if (rval != moab::MB_SUCCESS) 
	throw pargal_except("Error: " + mb.get_error_string(rval),
			    __FILE__, __LINE__, __PRETTY_FUNCTION__);
      std::string var_name = "T";
      std::size_t num_tsteps = fi.get_dim("time").size();
      std::vector<int> tsteps(num_tsteps);
      for (std::size_t i = 0; i != num_tsteps; ++i)
	tsteps[i] = i;
      
      pcvar var(var_name, moab::MB_TYPE_DOUBLE, fi.get_var_dims(var_name),
		fi, tsteps, false);

      var.get_storage();

    }
  catch (pargal::pargal_except& e) 
    {
      e.get_info();
      BOOST_CHECK(false);
    }
  catch (std::exception& e) 
    {
      std::cout << e.what() << std::endl;
      BOOST_CHECK(false);
    }
  catch (...) 
    {
      std::cerr << "Unrecoganized Exception!" << std::endl;
      BOOST_CHECK(false);
    }
  
  MPI_Finalize();
}
