/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <boost/test/test_tools.hpp>
#include <boost/test/output_test_stream.hpp>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <limits>
#include <cstdlib>
#include <cstring>
#include <cmath>
#include <moab/Core.hpp>
#include <pargal.hpp>
#include "mpi.h"

namespace pargal_tutil{

static const char test_stream_dfl_delim[] = {'\t'};
class test_stream{
	public:
		test_stream(const std::string &res_file);
		void set_tolerance(double tolerance);
		void set_stream_delim(const std::string &delim);
		test_stream& operator<<(const std::string& str);
	private:
		bool cmp(char *str1, char *str2, const char *delim);
		static const double DFL_TOLERANCE = 1e-12;
		static const int MAX_FILE_LINE_LEN = 100;
		std::ifstream istr_;
		std::string delim_;
		double tolerance_;
};

test_stream::test_stream(const std::string &res_file):tolerance_(DFL_TOLERANCE)
{
	istr_.open(res_file.c_str());
	if(!istr_.is_open()){
		BOOST_FAIL("Unable to open the result file");
		// FIXME: Throw an exception here?
	}
	delim_ = std::string(test_stream_dfl_delim);
}

inline void test_stream::set_tolerance(double tolerance)
{
	tolerance_ = tolerance;
}

inline void test_stream::set_stream_delim(const std::string &delim)
{
	delim_ = delim;
}

test_stream& test_stream::operator<<(const std::string &str)
{
	char usr_str[MAX_FILE_LINE_LEN];
	char file_str[MAX_FILE_LINE_LEN];
	std::string line;

	if(!istr_.good()){
		BOOST_FAIL("Error while comparing output: Reached end of file");
		return *this;
	}

	std::strncpy(usr_str, str.c_str(), MAX_FILE_LINE_LEN);

	std::getline(istr_, line);
	std::strncpy(file_str, line.c_str(), MAX_FILE_LINE_LEN);

	if(cmp(usr_str, file_str, delim_.c_str()) == false){
		BOOST_FAIL("Error while comparing output : values differ");
	}	

	return *this;
}

bool test_stream::cmp(char *str1, char *str2, const char *delim)
{
	// Strings are of format (0,0,0) 239.5949401855469
	char *token1, *token2, *save_ptr1, *save_ptr2;
	char *array_idx;

	token1 = strtok_r(str1, delim, &save_ptr1);
	token2 = strtok_r(str2, delim, &save_ptr2);
	array_idx = token1;

	if((token1 == NULL) || (token2 == NULL)){
		BOOST_FAIL("Error while comparing output: Invalid file or usr string, missing index");
		return false;
	}

	if(strcmp(token1, token2) != 0){
		BOOST_FAIL("Error while comparing output: Comparing different indices, index:" << array_idx << " and " << token2);
		return false;
	}

	token1 = strtok_r(NULL, delim, &save_ptr1);
	token2 = strtok_r(NULL, delim, &save_ptr2);
	if((token1 == NULL) || (token2 == NULL)){
		BOOST_FAIL("Error while comparing output: Invalid file or usr string, missing val");
		return false;
	}

	double dbl1, dbl2;
	dbl1 = atof(token1);
	dbl2 = atof(token2);

	if(fabs((dbl1 - dbl2)/((dbl1 != 0) ? dbl1 : 1.0)) < tolerance_){
		return true;
	}
	else{
		BOOST_FAIL("Error while comparing output: Values are different at index:" << array_idx << std::setprecision(16) << ": Value1 = " << dbl1 << ", Value2 = " << dbl2 << ", difference = " << fabs(dbl1 - dbl2) );
		BOOST_MESSAGE("Try setting a different tolerance if the values are close enough");
		return false;
	}
}

/* This function checks values at all "dims" in "value" with values read from
 * file with name, "cvalue_fname"
 */
void var_dims_check_root(pargal::pcvar &pcvar, const std::vector<int> &dims_to_check, const std::vector<double> &value, test_stream &test_str)
{
	int rank;
	std::size_t ndims = pcvar.get_dims().size();
	std::vector<std::size_t> place_val_max(ndims, 1);
	std::vector<std::size_t> place_val_i(ndims, 0);
	std::vector<std::size_t> place_val_wgt(ndims, 1);

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	if(rank != 0){
		return;
	}

	BOOST_REQUIRE(ndims != 0);
	BOOST_REQUIRE(dims_to_check.size() != 0);

	for(std::vector<int>::const_iterator iter = dims_to_check.begin(); iter != dims_to_check.end(); ++iter){
		std::size_t dim_to_check = *iter;
		BOOST_REQUIRE(dim_to_check < ndims);
		place_val_max[dim_to_check] = pcvar.get_dim(dim_to_check).size();
	}
	for(int i=ndims-2; i>= 0; i--){
		place_val_wgt[i] = place_val_wgt[i+1] * pcvar.get_dim(i+1).size();
	}
	std::size_t idx_max = 0;
	for(int i=0; i<ndims; i++){
		idx_max += place_val_wgt[i] * (place_val_max[i] - 1);
	}
	
	std::ostringstream cvalue_ostr;
	cvalue_ostr.precision(16);

	std::vector<int>::const_iterator dims_to_check_iter;
	//std::cout << " idx_max = " << idx_max << std::endl << std::flush;
	for(size_t i=0; i <= idx_max;){
		//std::cout << place_val_i[0] << ", " << place_val_i[1] << ", " << place_val_i[2] << ", " << place_val_i[3] << " : " << i << std::endl << std::flush;
		cvalue_ostr << "(" ;
		for(dims_to_check_iter = dims_to_check.begin(); dims_to_check_iter != dims_to_check.end() - 1; ++dims_to_check_iter){
			cvalue_ostr << place_val_i[*dims_to_check_iter] << ",";
		}
		cvalue_ostr << place_val_i[*dims_to_check_iter];
		cvalue_ostr << ")\t" ;
		cvalue_ostr << value[i] << "\n";
		test_str << cvalue_ostr.str();
		cvalue_ostr.str("");
		
		place_val_i[ndims -1] += 1;
		i++;
		if(place_val_i[ndims - 1] >= place_val_max[ndims - 1]){
			place_val_i[ndims - 1] = 0;
			for(int j = ndims - 2; j>=0; j--){
				place_val_i[j] += 1;
				if(place_val_i[j] >= place_val_max[j]){
					place_val_i[j] = 0;
				}
				else{
					i = 0;
					for(int k=0; k<ndims; k++){
						i += place_val_wgt[k] * place_val_i[k];
					}
					break;
				}
			}
		}
	}
}

void var_dims_check_root(pargal::pcvar &pcvar, const std::vector<int> &dims_to_check, const std::vector<double> &value, const std::string &cvalue_fname)
{
	test_stream test_str(cvalue_fname);
	var_dims_check_root(pcvar, dims_to_check, value, test_str);
}

void var_dims_check_root(pargal::pcvar &pcvar, const std::vector<int> &dims_to_check, const std::vector<double> &value, const std::string &cvalue_fname, double tolerance)
{
	test_stream test_str(cvalue_fname);
	test_str.set_tolerance(tolerance);
	var_dims_check_root(pcvar, dims_to_check, value, test_str);
}

} /* namespace pargal_tutil */
